(* begin hide *)
(* 45678901234567890123456789012345678901234567890123456789012345678901234567 *)
Require Import FLAM_Algebra.
Require Import FiniteSets.
Require Import Equalities.
(* end hide *)

(* A delegation is a 4-tuple (p,q,ℓ,r), representing (p≽q,ℓ) hosted at r. *)
Module DelT <: MiniDecidableType.
  Inductive T :=
  | make : Prin -> Prin -> Prin -> name -> T.

  Definition t := T.
  Definition eq_dec : forall (d1 d2 : t), {d1 = d2} + {d1 <> d2}.
    decide equality ; try (apply name_dec) ; decide equality ; apply name_dec.
  Defined.

  (*
   * Expresses the assertion that a given delegation is hosted at a given node.
   *)
  Definition hostedAt (cn : name) (d : t) :=
    match d with
    | make _ _ _ rn => cn = rn
    end.

End DelT.

Module Del := Make_UDT DelT.

(*
 * A delegation set is a FiniteSet of Del.t.
 *
 * Mostly a backward-compatibility layer for clients of an older version of
 * DelSet.
 *)
Module DelSet := FiniteSet Del.