(* begin hide *)
(* 45678901234567890123456789012345678901234567890123456789012345678901234567 *)
Require Import StrongInduction.

Require Import FLAM_Logic.
(* end hide *)

Import DelUseSet.

Lemma robActImpliesConfVoiceAct:
  forall H c pc l p q D,
    [H;c;pc;l ⊩ p ≽ q | D]
      -> [H;c;pc;l ⊩ ∇(p→) ≽ ∇(q→) | D].
Proof.
  assert (forall n H c pc l p q D,
      (exists D', [H;c;pc;l ⊩ p ≽ q # D';n] /\ included D' D)
        -> [H;c;pc;l ⊩ ∇(p→) ≽ ∇(q→) | D]).
  {
    induction n as [|n IHn] using strong_induction ; intros ; do 2 destruct H0.
    { inversion H0 ; omega. }
    {
      inversion H0.
      {
        pose proof normForm p as Hp.
        pose proof normForm q as Hq.
        destruct Hp as [p_c [p_i [pfp_c [pfp_i Hp]]]].
        destruct Hq as [q_c [q_i [pfq_c [pfq_i Hq]]]].
        rewrite Hp, Hq in *.
        setoid_rewrite Hp.
        setoid_rewrite Hq.
        algMagic.
        pose proof voiceOfNormJConf p_c pfp_c as Hvoice_p_c.
        pose proof voiceOfNormJConf q_c pfq_c as Hvoice_q_c.
        rewrite Hvoice_p_c, Hvoice_q_c.

        eapply robProj_integ.
        eapply robStatic.
        eapply normJConf ; auto.
        setoid_replace (p_c→) with ((p_c→ ∧ p_i←)→) by algMagic.
        setoid_replace (q_c→) with ((q_c→ ∧ q_i←)→) by algMagic.
        eapply AF_Proj_conf ; eautoflam.
      }
      {
        eexists.
        split.
          eautoflam.
          apply inclusionTrans with (y := x) ; auto.
          rewrite H6.
          autoflam.
      }
      {
        setoid_rewrite H4.
        algMagic.
        eautoflam.
      }
      {
        setoid_rewrite H3.
        algMagic.
        eapply robConjR.
          eapply IHn with (j := n1) ; auto with zarith.
          eexists.
          split ; eauto.
          apply inclusionTrans with (y := x) ; auto.
          rewrite H6.
          autoflam.

        eapply IHn with (j := n2) ; auto with zarith.
          eexists.
          split ; eauto.
          apply inclusionTrans with (y := x) ; auto.
          rewrite H6.
          autoflam.
      }
      {
        assert ([H;c;pc;l ⊩ ∇(p1→) ≽ ∇(q→) | D1]) as IH1.
        { eapply IHn with (j := n1) ; eautoflamz. }
        assert ([H;c;pc;l ⊩ ∇(p2→) ≽ ∇(q→) | D2]) as IH2.
        { eapply IHn with (j := n2) ; eautoflamz. }

        pose proof normForm p1 as [p1c [p1i [pfp1c [pfp1i Hp1]]]].
        pose proof normForm p2 as [p2c [p2i [pfp2c [pfp2i Hp2]]]].
        rewrite Hp1, Hp2 in *.
        setoid_rewrite Hp1 in IH1.
        setoid_rewrite Hp2 in IH2.

        setoid_replace ((p1c→ ∧ p1i←)→) with (p1c→) in IH1 by algMagic.
        setoid_replace ((p2c→ ∧ p2i←)→) with (p2c→) in IH2 by algMagic.
        pose proof voiceOfNormJConf _ pfp1c as Hvoice_p1c.
        pose proof voiceOfNormJConf _ pfp2c as Hvoice_p2c.
        rewrite Hvoice_p1c, Hvoice_p2c in *.

        setoid_rewrite H3.
        algMagic.
        setoid_rewrite <- disjConf.
        pose proof normJorJ _ pfp1c _ pfp2c
            as [p1cORp2c [pfp1cORp2c Hp1cORp2c]].
        assert (∇(p1cORp2c→) == p1cORp2c←) as Hvoice.
        {
          pose proof voiceOfNormJConf _ pfp1cORp2c as Hrewrite.
          rewrite Hrewrite.
          eapply refl.
        }
        assert (p1c ∨ p2c == p1cORp2c) as Hrewrite.
        {
          assert (p1cORp2c = projT1 (norm (p1c ∨ p2c))) as Hrewrite
              by (rewrite Hp1cORp2c ; auto).
          rewrite Hrewrite.
          apply norm_correct.
        }
        rewrite <- Hrewrite in Hvoice.
        setoid_rewrite Hvoice.

        algMagic.
        eapply robDisjL.
        {
          eapply robWeakenD with (D' := D1) ; auto.
          apply inclusionTrans with (y := x) ; auto.
          rewrite H6. autoflam.
        }
        {
          eapply robWeakenD with (D' := D2) ; auto.
          apply inclusionTrans with (y := x) ; auto.
          rewrite H6. autoflam.
        }
      }
      {
        assert ([H3;c;pc';l' ⊩ ∇(p→) ≽ ∇(q→) | D1]).
        { eapply IHn with (j := n1) ; eautoflamz. }
        do 3 destruct H19.

        eapply robWeakenH with (H1 := H3) (H2 := H4) ; auto.
        eapply robWeaken with (pc0 := pc') (l0 := l').
        {
          unfold robWeakensTo.
          split.
          {
            eapply robWeakenD with (D' := D2).
            {
              eexists.
              split.
                eauto.
                autoflam.
            }
            {
              apply inclusionTrans with (y := x) ; auto.
              rewrite H9.
              autoflam.
            }
          }
          {
            eapply robWeakenD with (D' := D3).
            {
              eexists.
              split.
                eauto.
                autoflam.
            }
            {
              apply inclusionTrans with (y := x) ; auto.
              rewrite H9.
              autoflam.
            }
          }
        }
        {
          eexists.
          split ; eauto.
          apply inclusionTrans with (y := x) ; auto.
          rewrite H9.
          apply inclusionTrans with (y := D1) ; autoflam.
        }
      }
      {
        eapply robTrans with (q := ∇(q0→)).
        {
          eapply IHn with (j := n1) ; autoflamz.
          eexists.
          split ; eauto.
          apply inclusionTrans with (y := x) ; auto.
          rewrite H6.
          autoflam.
        }
        {
          eapply IHn with (j := n2) ; autoflamz.
          eexists.
          split ; eauto.
          apply inclusionTrans with (y := x) ; auto.
          rewrite H6.
          autoflam.
        }
      }
      {
        eapply robFwd with (rn := rn) (D1 := D1) (D2 := D2).
        eauto. eauto. auto.
        {
          algMagic.
          eapply robTrans with (q := (pc→ ∧ l) ∧ ∇(q)).
          { eautoflam. }
          {
            eapply robConjR;
                [eapply robConjL1|eapply robConjL2];
                eapply robStatic;
                eautoflam.
          }
        }
        { eapply IHn with (j := n2) ; eautoflamz. }
        {
          apply inclusionTrans with (y := x) ; auto.
          rewrite H7.
          autoflam.
        }
      }
    }
  }

  intros.
  do 3 destruct H1.
  eauto.
Qed.

Lemma robActImpliesVoiceAct:
  forall H c pc l p q D,
    [H;c;pc;l ⊩ p ≽ q | D]
      -> [H;c;pc;l ⊩ ∇(p) ≽ ∇(q) | D].
Proof.
  intros.
  setoid_rewrite <- (confIntegBasis p).
  setoid_rewrite <- (confIntegBasis q).
  repeat setoid_rewrite voiceOfConj.
  eapply robConjR.
  {
    eapply robConjL1.
    eapply robActImpliesConfVoiceAct.
    auto.
  }
  {
    eapply robConjL2.
    algMagic.
    eapply robProj_integ.
    auto.
  }
Qed.

Lemma emptyPathIsLocal:
  forall n H cn pc l p q D du p' q' l' r rn,
       r = native rn
    -> [H;cn;pc;l ⊢ p ≽ q # D;n] \/ [H;cn;pc;l ⊩ p ≽ q # D;n]
    -> In du D
    -> DelUse.getDel du = DelT.make p' q' l' rn
    -> DelUse.getPath du = nil
    -> rn = cn.
Proof.
  induction n as [|n IHn] using strong_induction;
    intros H cn pc l p q D du p' q' l' r rn Hrn Hderiv HinD Hdel Hpath.
  {
    destruct Hderiv as [Hderiv | Hderiv].
    {
      inversion Hderiv;
        try (omega);
        try (
          setoid_replace D with empty in HinD by auto;
          contradict HinD;
          apply emptyIsEmpty;
          fail).
      {
        rewrite H6 in HinD.
        pose proof (initInv _ _ HinD) as Hdu.
        subst du.
        simpl in Hdel.
        rewrite H4 in Hdel.
        inversion Hdel.
        auto.
      }
    }
    { inversion Hderiv ; omega. }
  }
  {
    destruct Hderiv as [Hderiv | Hderiv].
    {
      inversion Hderiv;
        try (eapply IHn with (j := n1) ; eauto with zarith ; fail);
        try (
          try (rewrite H3 in HinD);
          try (rewrite H4 in HinD);
          try (rewrite H5 in HinD);
          destruct (unionInv _ _ _ HinD);
          [eapply IHn with (j := n1) | eapply IHn with (j := n2)];
          eauto with zarith;
          fail).
      {
        rewrite H5 in HinD.
        destruct (unionInv _ _ _ HinD).
        { eapply IHn with (j := n1) ; eauto with zarith. }
        {
          destruct (transferCorrect2 _ _ _ H15) as [du' [? ?]].
          contradict Hpath.
          subst du.
          destruct du'. simpl.
          intro Hfalse. symmetry in Hfalse. contradict Hfalse.
          apply List.nil_cons.
        }
      }
      {
        rewrite H7 in HinD.
        destruct (unionInv3 _ _ _ _ HinD) as [?|[?|?]];
          [ eapply IHn with (j := n1)
           | eapply IHn with (j := n2)
           | eapply IHn with (j := n3)];
          eauto with zarith.
      }
    }
    {
      inversion Hderiv;
        try (
          rewrite H4 in HinD;
          destruct (unionInv _ _ _ HinD);
          [eapply IHn with (j := n1) | eapply IHn with (j := n2)];
          eauto with zarith;
          fail);
        try (
          try (rewrite H4 in HinD);
          try (rewrite H7 in HinD);
          destruct (unionInv3 _ _ _ _ HinD) as [?|[?|?]];
          [  eapply IHn with (j := n1)
           | eapply IHn with (j := n2)
           | eapply IHn with (j := n3)];
          eauto with zarith;
          fail).
      {
        rewrite H2 in HinD.
        contradict HinD.
        apply emptyIsEmpty.
      }
      { eapply IHn with (j := n1) ; eauto with zarith. }
      {
        rewrite H5 in HinD.
        destruct (unionInv _ _ _ HinD).
        { eapply IHn with (j := n1) ; eauto with zarith. }
        {
          destruct (transferCorrect2 _ _ _ H15) as [du' [? ?]].
          contradict Hpath.
          subst du.
          destruct du'. simpl.
          intro Hfalse. symmetry in Hfalse. contradict Hfalse.
          apply List.nil_cons.
        }
      }
    }
  }
Qed.

Lemma delHostOnPath:
  forall n H cn pc l p q D du p' q' l' r rn,
       r = native rn
    -> [H;cn;pc;l ⊢ p ≽ q # D;n] \/ [H;cn;pc;l ⊩ p ≽ q # D;n]
    -> In du D
    -> DelUse.getDel du = DelT.make p' q' l' rn
    -> DelUse.getPath du <> nil
    -> [cn ⊢ r ≽ (DelUse.disjPath du)].
Proof.
  induction n as [|n IHn] using strong_induction;
    intros H cn pc l p q D du p' q' l' r rn Hrn Hderiv HinD Hdel Hpath.
  {
    destruct Hderiv as [Hderiv | Hderiv].
    {
      inversion Hderiv;
        try (omega);
        try (
          setoid_replace D with empty in HinD by auto;
          contradict HinD;
          apply emptyIsEmpty;
          fail).
      {
        rewrite H6 in HinD.
        pose proof (initInv _ _ HinD) as Hdu.
        unfold DelUse.init in Hdu.
        destruct du as [d' path].
        inversion Hdu.
        contradict Hpath.
        simpl. auto.
      }
    }
    { inversion Hderiv ; omega. }
  }
  {
    destruct Hderiv as [Hderiv | Hderiv].
    {
      inversion Hderiv;
        try (eapply IHn with (j := n1) ; eauto with zarith ; fail);
        try (
          try (rewrite H3 in HinD);
          try (rewrite H4 in HinD);
          try (rewrite H5 in HinD);
          destruct (unionInv _ _ _ HinD);
          [eapply IHn with (j := n1) | eapply IHn with (j := n2)];
          eauto with zarith;
          fail).
      {
        rewrite H5 in HinD.
        destruct (unionInv _ _ _ HinD).
        { eapply IHn with (j := n1) ; eauto with zarith. }
        {
          destruct (transferCorrect2 _ _ _ H15) as [du' [? Heqdu]].
          subst du.
          rewrite (DelUse.disjPathTransfer du' rn0) in *.
          destruct (name_dec rn rn0).
          { subst. eautoflam. }
          {
            assert ([rn0 ⊢ r ≽ DelUse.disjPath du']).
            {
              destruct du' as [d' path'].
              simpl in *.
              eapply IHn with (j := n2) ; eauto with zarith.
              intro.
              contradict n3.
              eapply emptyPathIsLocal ; eauto.
            }

            destruct H17.
            assert (exists n, [cn ⊢ r ≽ DelUse.disjPath du' # n] /\ n <= x)
              as Hcoerce.
            { eapply staticCoerce ; eauto. }
            destruct Hcoerce as [? [? ?]].
            eexists. eautoflam.
          }
        }
      }
      {
        rewrite H7 in HinD.
        destruct (unionInv3 _ _ _ _ HinD) as [?|[?|?]];
          [ eapply IHn with (j := n1)
           | eapply IHn with (j := n2)
           | eapply IHn with (j := n3)];
          eauto with zarith.
      }
    }
    {
      inversion Hderiv;
        try (
          rewrite H4 in HinD;
          destruct (unionInv _ _ _ HinD);
          [eapply IHn with (j := n1) | eapply IHn with (j := n2)];
          eauto with zarith;
          fail);
        try (
          try (rewrite H4 in HinD);
          try (rewrite H7 in HinD);
          destruct (unionInv3 _ _ _ _ HinD) as [?|[?|?]];
          [  eapply IHn with (j := n1)
           | eapply IHn with (j := n2)
           | eapply IHn with (j := n3)];
          eauto with zarith;
          fail).
      {
        rewrite H2 in HinD.
        contradict HinD.
        apply emptyIsEmpty.
      }
      { eapply IHn with (j := n1) ; eauto with zarith. }
      {
        rewrite H5 in HinD.
        destruct (unionInv _ _ _ HinD).
        { eapply IHn with (j := n1) ; eauto with zarith. }
        {
          destruct (transferCorrect2 _ _ _ H15) as [du' [? Heqdu]].
          subst du.
          rewrite (DelUse.disjPathTransfer du' rn0) in *.
          destruct (name_dec rn rn0).
          { subst. eautoflam. }
          {
            assert ([rn0 ⊢ r ≽ DelUse.disjPath du']).
            {
              destruct du' as [d' path'].
              simpl in *.
              eapply IHn with (j := n2) ; eauto with zarith.
              intro.
              contradict n3.
              eapply emptyPathIsLocal ; eauto.
            }

            destruct H17.
            assert (exists n, [cn ⊢ r ≽ DelUse.disjPath du' # n] /\ n <= x)
              as Hcoerce.
            { eapply staticCoerce ; eauto. }
            destruct Hcoerce as [? [? ?]].
            eexists. eautoflam.
          }
        }
      }
    }
  }
Qed.

Theorem robAuth:
  forall H c cn pc l p q D n nn du p' q' l' N x k,
       c = native cn
    -> n = native nn
    -> [H;cn;pc;l ⊢ p ≽ q # D;x] \/ [H;cn;pc;l ⊩ p ≽ q # D;x]
    -> In du D
    -> DelUse.getDel du = (DelT.make p' q' l' nn)
    -> k = length (DelUse.getPath du)
    -> N = DelUse.disjPath du
    ->     [H;cn;pc;l ⊩ l' ∨ N ⊑ l | D]
        /\ [H;cn;pc;l ⊩ N ≽ pc→ ∧ l← | D]
        /\ (k > 0 -> [H;cn;pc;l ⊩ c ≽ (l' ∨ N)→ | D]).
Proof.
  assert (forall n H c cn pc l p q D r rn du p' q' l' N,
         c = native cn
      -> r = native rn
      -> [H;cn;pc;l ⊢ p ≽ q # D;n] \/ [H;cn;pc;l ⊩ p ≽ q # D;n]
      -> In du D
      -> DelUse.getDel du = (DelT.make p' q' l' rn)
      -> N = DelUse.disjPath du
      ->     [H;cn;pc;l ⊩ l' ∨ N ⊑ l | D]                    (* concl 1 *)
          /\ [H;cn;pc;l ⊩ N ≽ pc→ ∧ l← | D]                  (* concl 2 *)
          /\ (rn <> cn -> [H;cn;pc;l ⊩ c ≽ (l' ∨ N)→ | D])). (* concl 3 *)
  {
    induction n as [|n IHn] using strong_induction;
      intros H c cn pc l p q D r rn du p' q' l' N Hcn Hrn Hderiv HinD Hdel
        HeqN.
    {
      (* base case (n=0). *)
      destruct Hderiv as [Hderiv | Hderiv].
      {
        (* base case (n=0). *)
        (* non-robust judgements. *)
        inversion Hderiv;
          try (omega);
          try (
            setoid_replace D with empty in HinD by auto;
            contradict HinD;
            apply emptyIsEmpty;
            fail).
        {
          (* base case (n=0). *)
          (* subcase: non-robust judgements. *)
          (* subcase: p ≽ q derived w/ AF_Del. *)
          rewrite H6 in HinD.
          pose proof (initInv _ _ HinD).
          subst du.
          rewrite DelUse.disjPathInit in HeqN.
          subst N. setoid_rewrite disjIdent.

          rewrite H4 in Hdel.
          simpl in Hdel.
          inversion Hdel.

          split ; [|split].
          {
            (* base case (n=0). *)
            (* subcase: non-robust judgements. *)
            (* subcase: p ≽ q derived w/ AF_Del. *)
            (* showing conclusion 1. *)
            setoid_rewrite H3.
            rewrite H17.
            apply robStatic. eautoflam.
          }
          {
            (* base case (n=0). *)
            (* subcase: non-robust judgements. *)
            (* subcase: p ≽ q derived w/ AF_Del. *)
            (* showing conclusion 2. *)
            apply robStatic. eautoflam.
          }
          {
            (* base case (n=0). *)
            (* subcase: non-robust judgements. *)
            (* subcase: p ≽ q derived w/ AF_Del. *)
            (* showing conclusion 3. *)
            tauto.
          }
        }
      }
      {
        (* base case (n=0). *)
        (* robust judgements. *)
        inversion Hderiv ; omega.
      }
    }
    {
      (* inductive case (n > 0). *)

      (* A specialized form of the IH, which will be useful later. *)
      assert (forall j p0 q0 D0,
           j <= n
        -> In du D0
        -> included D0 D
        -> [H;cn;pc;l ⊢ p0 ≽ q0 # D0;j] \/ [H;cn;pc;l ⊩ p0 ≽ q0 # D0;j]
        -> (    [H;cn;pc;l ⊩ l' ∨ N ⊑ l | D]
             /\ [H;cn;pc;l ⊩ N ≽ pc→ ∧ l← | D]
             /\ (rn <> cn -> [H;cn;pc;l ⊩ c ≽ (l' ∨ N)→ | D]))) as IHhelp.
      {
        intros j p0 q0 D0 Hcount HinD0 Hincl Hsubderiv.
        split;
          [|split ; [|intro Hrnec]];
          apply robWeakenD with (D' := D0); auto;
          eapply IHn with (j := j); eautoflamz.
      }

      destruct Hderiv as [Hderiv|Hderiv].
      {
        (* inductive case *)
        (* subcase: non-robust judgements *)
        inversion Hderiv.
        {
          (* inductive case *)
          (* subcase: non-robust judgements *)
          (* inductive case: p ≽ q derived w/ AF_Proj_conf. *)
          eapply IHn with (j := n1) ; eauto with zarith.
        }
        {
          (* inductive case *)
          (* subcase: non-robust judgements *)
          (* inductive case: p ≽ q derived w/ AF_Proj_integ. *)
          eapply IHn with (j := n1) ; eauto with zarith.
        }
        {
          (* inductive case *)
          (* subcase: non-robust judgements *)
          (* inductive case: p ≽ q derived w/ AF_Own1. *)
          remember H5 as Hrewrite.
          setoid_rewrite Hrewrite.
          rewrite Hrewrite in HinD.
          setoid_rewrite Hrewrite in IHhelp.
          destruct (unionInv _ _ _ HinD);
            [eapply IHhelp with (j := n1) | eapply IHhelp with (j := n2)];
            eauto with zarith;
            autoflam.
        }
        {
          (* inductive case: p ≽ q derived w/ AF_Own2. *)
          remember H5 as Hrewrite.
          setoid_rewrite Hrewrite.
          rewrite Hrewrite in HinD.
          setoid_rewrite Hrewrite in IHhelp.
          destruct (unionInv _ _ _ HinD);
            [eapply IHhelp with (j := n1) | eapply IHhelp with (j := n2)];
            eauto with zarith;
            autoflam.
        }
        {
          (* inductive case: p ≽ q derived w/ AF_ConjL1. *)
          eapply IHn with (j := n1) ; eauto with zarith.
        }
        {
          (* inductive case: p ≽ q derived w/ AF_ConjR. *)
          remember H4 as Hrewrite.
          setoid_rewrite Hrewrite.
          rewrite Hrewrite in HinD.
          setoid_rewrite Hrewrite in IHhelp.
          destruct (unionInv _ _ _ HinD);
            [eapply IHhelp with (j := n1) | eapply IHhelp with (j := n2)];
            eauto with zarith;
            autoflam.
        }
        {
          (* inductive case: p ≽ q derived w/ AF_DisjL. *)
          remember H4 as Hrewrite.
          setoid_rewrite Hrewrite.
          rewrite Hrewrite in HinD.
          setoid_rewrite Hrewrite in IHhelp.
          destruct (unionInv _ _ _ HinD);
            [eapply IHhelp with (j := n1) | eapply IHhelp with (j := n2)];
            eauto with zarith;
            autoflam.
        }
        {
          (* inductive case: p ≽ q derived w/ AF_DisjR1. *)
          eapply IHn with (j := n1) ; eauto with zarith.
        }
        {
          (* inductive case: p ≽ q derived w/ AF_Trans. *)
          remember H3 as Hrewrite.
          setoid_rewrite Hrewrite.
          rewrite Hrewrite in HinD.
          setoid_rewrite Hrewrite in IHhelp.
          destruct (unionInv _ _ _ HinD);
            [eapply IHhelp with (j := n1) | eapply IHhelp with (j := n2)];
            eauto with zarith;
            autoflam.
        }
        {
          (* inductive case: p ≽ q derived w/ AF_Fwd. *)
          subst c0. rewrite <- Hcn in *.
          rewrite H5 in HinD.
          destruct (unionInv _ _ _ HinD).
          {
            (* inductive case: p ≽ q derived w/ AF_Fwd. *)
            (* subcase: delegation in D1. *)
            eapply IHhelp with (D0 := D1) (j := n1) ; eauto with zarith.
            rewrite H5. autoflam.
          }
          {
            (* inductive case: p ≽ q derived w/ AF_Fwd. *)
            (* subcase: delegation in D2. *)
            destruct (transferCorrect2 _ _ _ H1) as [du' [? Heqdu]].

            remember (DelUse.disjPath du') as N'.
            assert (
                   [H;rn0;pc ⊔ l ⊔ c←;l ⊓ c→ ⊩ l' ∨ N' ⊑ l ⊓ c→ | D2]
                /\ [H;rn0;pc ⊔ l ⊔ c←;l ⊓ c→ ⊩
                      N' ≽ (pc ⊔ l ⊔ c←)→ ∧ (l ⊓ c→)← | D2]
                /\ (rn <> rn0
                      -> [H;rn0;pc ⊔ l ⊔ c←;l ⊓ c→ ⊩ r0 ≽ (l' ∨ N')→ | D2]))
              as IH.
            {
              eapply IHn with (j := n2) ; eauto with zarith.
              {
                destruct du' as [d path].
                rewrite Heqdu in Hdel.
                simpl in *. eauto.
              }
            }

            split ; [|split ; [|intro Hnec]].
            {
              (* inductive case: p ≽ q derived w/ AF_Fwd. *)
              (* subcase: delegation in D2. *)
              (* showing conclusion 1. *)
              rewrite Heqdu, DelUse.disjPathTransfer, <- H2, <- HeqN' in HeqN.
              subst N.
              eapply robConjR.
              {
                (* inductive case: p ≽ q derived w/ AF_Fwd. *)
                (* subcase: delegation in D2. *)
                (* showing conclusion 1. *)
                (* confidentiality half *)
                eapply robConjL1.
                eapply robFwd with (r := r0) (D1 := D) (D2 := D2) ; eauto.
                {
                  setoid_rewrite H5.
                  apply robConjR.
                  { eautoflam. }
                  {
                    apply robTrans with (q := ∇(r0→));
                      apply robStatic.
                    { erewrite voiceOfNativeConf ; eauto. eautoflam. }
                    {
                      setoid_rewrite <- disjAssoc.
                      setoid_rewrite (disjComm l' r0).
                      setoid_rewrite disjAssoc.
                      setoid_rewrite disjConf.
                      apply voiceDisjMonotonic.
                    }
                  }
                }
                {
                  apply robTrans with (q := (l' ∨ N')→).
                  {
                    apply robFlow2Conf.
                    apply robFlowTrans with (q := l ⊓ c→).
                    { apply IH. }
                    { apply robStatic. eautoflam. }
                  }
                  {
                    apply robStatic.
                    setoid_rewrite <- disjAssoc.
                    setoid_rewrite (disjComm l' r0).
                    setoid_rewrite disjAssoc.
                    algMagic.
                    eautoflam.
                  }
                }
                { rewrite H5. autoflam. }
              }
              {
                (* inductive case: p ≽ q derived w/ AF_Fwd. *)
                (* showing conclusion 1. *)
                (* subcase: delegation in D2. *)
                (* integrity half *)
                setoid_rewrite H5.
                apply robConjL2.
                setoid_rewrite <- disjAssoc.
                setoid_rewrite (disjComm l' r0).
                setoid_rewrite disjAssoc.
                setoid_rewrite disjInteg.
                apply robDisjL.
                {
                  apply robProj_integ.
                  apply robTrans with (q := pc→ ∧ l).
                  { apply robWeakenD with (D' := D1) ; eautoflam. }
                  { apply robStatic. eautoflam. }
                }
                {
                  eapply robFwd with (r := r0) (D1 := D1) (D2 := D2) ; eauto.
                  {
                    setoid_replace ((pc→ ∧ l) ∧ ∇(l←)) with (pc→ ∧ l).
                    { eautoflam. }
                    { rewrite conjAssoc. algMagic. }
                  }
                  {
                    apply robFlow2Integ.
                    apply robFlowTrans with (q := l ⊓ c→).
                    { apply IH. }
                    { apply robStatic. eautoflam. }
                  }
                  { autoflam. }
                }
              }
            }
            {
              (* inductive case: p ≽ q derived w/ AF_Fwd. *)
              (* subcase: delegation in D2. *)
              (* showing conclusion 2. *)
              rewrite Heqdu, DelUse.disjPathTransfer, <- H2, <- HeqN' in HeqN.
              subst N.
              apply robDisjL.
              {
                eapply robTrans with (q := pc→ ∧ l).
                {
                  setoid_rewrite H5.
                  eapply robWeakenD with (D' := D1) ; eautoflam.
                }
                {
                  eapply robStatic.
                  setoid_rewrite <- (confIntegBasis l) at 1.
                  eautoflam.
                }
              }
              {
                eapply robFwd with (r := r0) (D1 := D1) (D2 := D2) ; eauto.
                {
                  eapply robConjR.
                  { eautoflam. }
                  {
                    eapply robTrans with (q := r0←).
                    { eautoflam. }
                    {
                      setoid_rewrite <- (voiceOfNative r0 rn0 H2).
                      eapply robActImpliesVoiceAct.
                      eapply robTrans with (q := pc→ ∧ l).
                      { eautoflam. }
                      {
                        eapply robStatic.
                        setoid_rewrite <- (confIntegBasis l) at 1.
                        eautoflam.
                      }
                    }
                  }
                }
                {
                  eapply robTrans with (q := (pc ⊔ l ⊔ c←)→ ∧ (l ⊓ c→)←).
                  {
                    eapply IHn with (j := n2) (r := r) (N := N');
                      eauto with zarith.
                    {
                      destruct du' as [d path].
                      rewrite Heqdu in Hdel.
                      simpl in *. eauto.
                    }
                  }
                  {
                    apply robStatic. algMagic.
                    eexists. eapply AF_ConjR ; eautoflam.
                  }
                }
                { rewrite H5. autoflam. }
              }
            }
            {
              (* inductive case: p ≽ q derived w/ AF_Fwd. *)
              (* subcase: delegation in D2. *)
              (* showing conclusion 3. *)
              rewrite Heqdu, DelUse.disjPathTransfer, <- H2, <- HeqN' in HeqN.
              subst N.
              eapply robFwd with (r := r0) (D1 := D1) (D2 := D2) ; eauto.
              {
                apply robConjR.
                { eautoflam. }
                {
                  apply robTrans with (q := ∇(r0→));
                    apply robStatic.
                  { erewrite voiceOfNativeConf ; eauto. eautoflam. }
                  {
                    setoid_rewrite <- disjAssoc.
                    setoid_rewrite (disjComm l' r0).
                    setoid_rewrite disjAssoc.
                    setoid_rewrite disjConf.
                    apply voiceDisjMonotonic.
                  }
                }
              }
              {
                apply robTrans with (q := (l' ∨ N')→).
                {
                  apply robTrans with (q := (l ⊓ c→)→).
                  { apply robStatic. algMagic. eautoflam. }
                  { apply robFlow2Conf. apply IH. }
                }
                {
                  apply robStatic.
                  setoid_rewrite <- disjAssoc.
                  setoid_rewrite (disjComm l' r0).
                  setoid_rewrite disjAssoc.
                  algMagic.
                  eautoflam.
                }
              }
              { rewrite H5. autoflam. }
            }
          }
        }
        {
          (* inductive case: p ≽ q derived w/ AF_Weaken. *)
          assert (
              forall pc0 l0 D0,
                   In du D0
                -> (    [H1;cn;pc0;l0 ⊩ (l' ∨ N) ⊑ l0 | D0]
                     /\ [H1;cn;pc0;l0 ⊩ N ≽ pc0→ ∧ l0← | D0]
                     /\ (rn <> cn -> [H1;cn;pc0;l0 ⊩ c ≽ (l' ∨ N)→ | D0]))
                -> included D0 D
                -> robWeakensTo H cn pc0 l0 pc l D
                ->    [H;cn;pc;l ⊩ (l' ∨ N) ⊑ l | D]
                   /\ [H;cn;pc;l ⊩ N ≽ pc→ ∧ l← | D]
                   /\ (rn <> cn -> [H;cn;pc;l ⊩ c ≽ (l' ∨ N)→ | D]))
            as IHhelp'.
          {
            intros pcx lx Dx Hlocal IH Hincl Hweaken.
            assert ([H; cn; pc; l ⊩ pc ⊑ pcx | D]).
            {
              destruct Hweaken.
              eapply robWeaken with (pc0 := pc ⊔ lx) (l0 := l) ; auto.
              { eexists ; apply robStatic ; eautoflam. }
            }

            assert ([H; cn; pc; l ⊩ lx ⊑ l | D]).
            {
              destruct Hweaken.
              eapply robWeaken with (pc0 := pc ⊔ lx) (l0 := l) ; auto.
              { eexists ; apply robStatic ; eautoflam. }
            }

            destruct IH as [IH1 [IH2 IH3]].
            split ; [|split].
            {
              (* specialized IH, conclusion 1. *)
              eapply robFlowTrans with (q := lx) ; auto.
              eapply robWeaken ; eauto.
              eapply robWeakenH with (H1 := H1) ; eauto.
              eapply robWeakenD with (D' := Dx) ; auto.
            }
            {
              (* specialized IH, conclusion 2. *)
              apply robTrans with (q := pcx→ ∧ lx←).
              {
                eapply robWeaken ; eauto.
                eapply robWeakenH with (H1 := H1) ; eauto.
                eapply robWeakenD with (D' := Dx) ; auto.
              }
              {
                apply robConjR.
                { apply robConjL1. apply robFlow2Conf. auto. }
                { apply robConjL2. apply robFlow2Integ. auto. }
              }
            }
            {
              (* specialized IH, conclusion 3. *)
              intro Hrnec.
              eapply robWeaken ; eauto.
              eapply robWeakenH with (H1 := H1) ; eauto.
              eapply robWeakenD with (D' := Dx) ; auto.
            }
          }

          assert (robWeakensTo H cn pc' l'0 pc l D).
          {
            eexists.
            {
              eapply robStaticWeaken with (pc0 := pc ⊔ l'0) (l0 := l).
              { eautoflam. }
              { eautoflam. }
              {
                eapply robWeakenD with (D' := D2).
                { eapply robWeakenH with (H1 := H1) ; eautoflam. }
                { setoid_rewrite H7. autoflam. }
              }
            }
            {
              eapply robStaticWeaken with (pc0 := pc ⊔ l'0) (l0 := l).
              { eautoflam. }
              { eautoflam. }
              {
                eapply robWeakenD with (D' := D3).
                { eapply robWeakenH with (H1 := H1) ; eautoflam. }
                { setoid_rewrite H7. eautoflam. }
              }
            }
          }

          assert (robWeakensTo H cn (pc ⊔ l'0) l pc l D).
          { eexists ; apply robStatic ; eautoflam. }

          rewrite H7 in HinD.
          destruct (unionInv3 _ _ _ _ HinD) as [?|[?|?]];
            [  eapply IHhelp' with (pc0 := pc') (l0 := l'0) (D0 := D1)
             | eapply IHhelp' with (pc0 := pc ⊔ l'0) (l0 := l) (D0 := D2)
             | eapply IHhelp' with (pc0 := pc ⊔ l'0) (l0 := l) (D0 := D3)];
            eauto with zarith;
            try (eapply IHn with (j := n1) ; eauto with zarith ; fail);
            try (eapply IHn with (j := n2) ; eauto with zarith ; fail);
            try (eapply IHn with (j := n3) ; eauto with zarith ; fail);
            try (rewrite H7 ; autoflam).
        }
      }
      {
        (* inductive case *)
        (* subcase: robust judgements *)

        (*
         * A specialized form of the IH, which will be useful for RAF_Lift
         * and RAF_LiftPC.
         *)
        assert (forall j o p0 q0 D0 ,
               j <= n
            -> included D0 D
            ->    [H;cn;pc;l ∧ ∇(o) ⊢ p0 ≽ q0 # D0;j]
               \/ [H;cn;pc;l ∧ ∇(o) ⊩ p0 ≽ q0 # D0;j]
            -> In du D0
            ->    [H;cn;pc;l ⊩ l' ∨ N ⊑ l | D]
               /\ [H;cn;pc;l ⊩ N ≽ pc→ ∧ l← | D]
               /\ (rn <> cn -> [H;cn;pc;l ⊩ c ≽ (l' ∨ N)→ | D])) as IHliftHelp.
        {
          intros j o p0 q0 D0 Hcount Hincl Hsubderiv' HinD0.

          assert (robWeakensTo H cn pc (l ∧ ∇(o)) pc l D0).
          {
            destruct (voiceIsInteg o) as [x Hvoice].
            rewrite Hvoice in *.
            eexists;
                eapply robStatic;
                [|algMagic ; setoid_rewrite <- conjAssoc ; algMagic];
                eautoflam.
          }

          destruct (voiceIsInteg o) as [x Hvoice].
          rewrite Hvoice in *.

          split ; [|split].
          {
            eapply robWeakenD with (D' := D0) ; auto.
            eapply robWeaken with (pc0 := pc) (l0 := l ∧ x←) ; auto.
            eapply robFlowTrans with (q := l ∧ x←).
            { eapply IHn with (j := j) (rn := rn) ; eauto with zarith. }
            {
              eapply robStatic. algMagic.
              setoid_rewrite <- conjAssoc. algMagic.
              eautoflam.
            }
          }
          {
            eapply robWeakenD with (D' := D0) ; auto.
            eapply robWeaken with (pc0 := pc) (l0 := l ∧ x←) ; auto.
            eapply robTrans with (q := pc→ ∧ (l ∧ x←)←).
            { eapply IHn with (j := j) (rn := rn) ; eauto with zarith. }
            { eapply robStatic. algMagic. eautoflam. }
          }
          {
            intro Hrnec.
            eapply robWeakenD with (D' := D0) ; auto.
            eapply robWeaken with (pc0 := pc) (l0 := l ∧ x←) ; auto.
            eapply IHn with (j := j) (rn := rn) ; eauto with zarith.
          }
        }

        inversion Hderiv.
        {
          (* inductive case: p ≽ q derived w/ RAF_Static. *)
          rewrite H2 in HinD.
          contradict HinD.
          apply emptyIsEmpty.
        }
        {
          (* inductive case: p ≽ q derived w/ RAF_Lift. *)
          rewrite H4 in HinD.
          destruct (unionInv3 _ _ _ _ HinD) as [?|[?|?]].
          {
            (* subcase: delegation in D1. *)
            eapply IHliftHelp with (j := n1) (D0 := D1) ; eauto with zarith.
            rewrite H4. autoflam.
          }
          {
            (* subcase: delegation in D2. *)
            eapply IHliftHelp with (j := n2) (D0 := D2) (o := ⊥);
              eauto with zarith.
            { rewrite H4. autoflam. }
            { right. setoid_replace (l ∧ ∇(⊥)) with l ; eauto. algMagic. }
          }
          {
            (* subcase: delegation in D3. *)
            eapply IHliftHelp with (j := n3) (D0 := D3) (o := ⊥);
              eauto with zarith.
            { rewrite H4. autoflam. }
            { right. setoid_replace (l ∧ ∇(⊥)) with l ; eauto. algMagic. }
          }
        }
        {
          (* inductive case: p ≽ q derived w/ RAF_LiftPC. *)
          eapply IHliftHelp with (j := n1) (D0 := D) ; eauto with zarith.
          autoflam.
        }
        {
          (* inductive case: p ≽ q derived w/ RAF_ConjR. *)
          rewrite H4 in HinD. setoid_rewrite H4 in IHhelp. setoid_rewrite H4.
          destruct (unionInv _ _ _ HinD).
          { eapply IHhelp with (j := n1) (D0 := D1) ; eautoflamz. }
          { eapply IHhelp with (j := n2) (D0 := D2) ; eautoflamz. }
        }
        {
          (* inductive case: p ≽ q derived w/ RAF_DisjR. *)
          rewrite H4 in HinD. setoid_rewrite H4 in IHhelp. setoid_rewrite H4.
          destruct (unionInv _ _ _ HinD).
          { eapply IHhelp with (j := n1) (D0 := D1) ; eautoflamz. }
          { eapply IHhelp with (j := n2) (D0 := D2) ; eautoflamz. }
        }
        {
          (* inductive case: p ≽ q derived w/ RAF_Weaken. *)
          assert (
              forall pc0 l0 D0,
                   In du D0
                -> (    [H1;cn;pc0;l0 ⊩ (l' ∨ N) ⊑ l0 | D0]
                     /\ [H1;cn;pc0;l0 ⊩ N ≽ pc0→ ∧ l0← | D0]
                     /\ (rn <> cn -> [H1;cn;pc0;l0 ⊩ c ≽ (l' ∨ N)→ | D0]))
                -> included D0 D
                -> robWeakensTo H cn pc0 l0 pc l D
                ->    [H;cn;pc;l ⊩ (l' ∨ N) ⊑ l | D]
                   /\ [H;cn;pc;l ⊩ N ≽ pc→ ∧ l← | D]
                   /\ (rn <> cn -> [H;cn;pc;l ⊩ c ≽ (l' ∨ N)→ | D]))
            as IHhelp'.
          {
            intros pcx lx Dx Hlocal IH Hincl Hweaken.
            assert ([H; cn; pc; l ⊩ pc ⊑ pcx | D]).
            {
              destruct Hweaken.
              eapply robWeaken with (pc0 := pc ⊔ lx) (l0 := l) ; auto.
              { eexists ; apply robStatic ; eautoflam. }
            }

            assert ([H; cn; pc; l ⊩ lx ⊑ l | D]).
            {
              destruct Hweaken.
              eapply robWeaken with (pc0 := pc ⊔ lx) (l0 := l) ; auto.
              { eexists ; apply robStatic ; eautoflam. }
            }

            destruct IH as [IH1 [IH2 IH3]].
            split ; [|split].
            {
              (* specialized IH, conclusion 1. *)
              eapply robFlowTrans with (q := lx) ; auto.
              eapply robWeaken ; eauto.
              eapply robWeakenH with (H1 := H1) ; eauto.
              eapply robWeakenD with (D' := Dx) ; auto.
            }
            {
              (* specialized IH, conclusion 2. *)
              apply robTrans with (q := pcx→ ∧ lx←).
              {
                eapply robWeaken ; eauto.
                eapply robWeakenH with (H1 := H1) ; eauto.
                eapply robWeakenD with (D' := Dx) ; auto.
              }
              {
                apply robConjR.
                { apply robConjL1. apply robFlow2Conf. auto. }
                { apply robConjL2. apply robFlow2Integ. auto. }
              }
            }
            {
              (* specialized IH, conclusion 3. *)
              intro Hrnec.
              eapply robWeaken ; eauto.
              eapply robWeakenH with (H1 := H1) ; eauto.
              eapply robWeakenD with (D' := Dx) ; auto.
            }
          }

          assert (robWeakensTo H cn pc' l'0 pc l D).
          {
            eexists.
            {
              eapply robStaticWeaken with (pc0 := pc ⊔ l'0) (l0 := l).
              { eautoflam. }
              { eautoflam. }
              {
                eapply robWeakenD with (D' := D2).
                { eapply robWeakenH with (H1 := H1) ; eautoflam. }
                { setoid_rewrite H7. autoflam. }
              }
            }
            {
              eapply robStaticWeaken with (pc0 := pc ⊔ l'0) (l0 := l).
              { eautoflam. }
              { eautoflam. }
              {
                eapply robWeakenD with (D' := D3).
                { eapply robWeakenH with (H1 := H1) ; eautoflam. }
                { setoid_rewrite H7. eautoflam. }
              }
            }
          }

          assert (robWeakensTo H cn (pc ⊔ l'0) l pc l D).
          { eexists ; apply robStatic ; eautoflam. }

          rewrite H7 in HinD.
          destruct (unionInv3 _ _ _ _ HinD) as [?|[?|?]];
            [  eapply IHhelp' with (pc0 := pc') (l0 := l'0) (D0 := D1)
             | eapply IHhelp' with (pc0 := pc ⊔ l'0) (l0 := l) (D0 := D2)
             | eapply IHhelp' with (pc0 := pc ⊔ l'0) (l0 := l) (D0 := D3)];
            eauto with zarith;
            try (eapply IHn with (j := n1) ; eauto with zarith ; fail);
            try (eapply IHn with (j := n2) ; eauto with zarith ; fail);
            try (eapply IHn with (j := n3) ; eauto with zarith ; fail);
            try (rewrite H7 ; autoflam).
        }
        {
          (* inductive case: p ≽ q derived w/ RAF_Trans. *)
          rewrite H4 in HinD. setoid_rewrite H4 in IHhelp. setoid_rewrite H4.
          destruct (unionInv3 _ _ _ _ HinD) as [?|[?|?]].
          { eapply IHhelp with (j := n1) (D0 := D1) ; eautoflamz. }
          { eapply IHhelp with (j := n2) (D0 := D2) ; eautoflamz. }
          { eapply IHhelp with (j := n3) (D0 := D3) ; eautoflamz. }
        }
        {
          (* inductive case: p ≽ q derived w/ RAF_Fwd. *)
          subst c0. rewrite <- Hcn in *.
          rewrite H5 in HinD.
          destruct (unionInv _ _ _ HinD).
          {
            (* inductive case: p ≽ q derived w/ RAF_Fwd. *)
            (* subcase: delegation in D1. *)
            eapply IHhelp with (D0 := D1) (j := n1) ; eauto with zarith.
            rewrite H5. autoflam.
          }
          {
            (* inductive case: p ≽ q derived w/ RAF_Fwd. *)
            (* subcase: delegation in D2. *)
            destruct (transferCorrect2 _ _ _ H1) as [du' [? Heqdu]].

            remember (DelUse.disjPath du') as N'.
            assert (
                   [H;rn0;pc ⊔ l ⊔ c←;l ⊓ c→ ⊩ l' ∨ N' ⊑ l ⊓ c→ | D2]
                /\ [H;rn0;pc ⊔ l ⊔ c←;l ⊓ c→ ⊩
                      N' ≽ (pc ⊔ l ⊔ c←)→ ∧ (l ⊓ c→)← | D2]
                /\ (rn <> rn0
                      -> [H;rn0;pc ⊔ l ⊔ c←;l ⊓ c→ ⊩ r0 ≽ (l' ∨ N')→ | D2]))
              as IH.
            {
              eapply IHn with (j := n2) ; eauto with zarith.
              {
                destruct du' as [d path].
                rewrite Heqdu in Hdel.
                simpl in *. eauto.
              }
            }

            split ; [|split ; [|intro Hnec]].
            {
              (* inductive case: p ≽ q derived w/ RAF_Fwd. *)
              (* subcase: delegation in D2. *)
              (* showing conclusion 1. *)
              rewrite Heqdu, DelUse.disjPathTransfer, <- H2, <- HeqN' in HeqN.
              subst N.
              eapply robConjR.
              {
                (* inductive case: p ≽ q derived w/ RAF_Fwd. *)
                (* subcase: delegation in D2. *)
                (* showing conclusion 1. *)
                (* confidentiality half *)
                eapply robConjL1.
                eapply robFwd with (r := r0) (D1 := D) (D2 := D2) ; eauto.
                {
                  setoid_rewrite H5.
                  apply robConjR.
                  {
                    apply robTrans with (q := (pc→ ∧ l) ∧ ∇(q)).
                    { apply robWeakenD with (D' := D1) ; eautoflam. }
                    { apply robStatic. eautoflam. }
                  }
                  {
                    apply robTrans with (q := ∇(r0→));
                      apply robStatic.
                    { erewrite voiceOfNativeConf ; eauto. eautoflam. }
                    {
                      setoid_rewrite <- disjAssoc.
                      setoid_rewrite (disjComm l' r0).
                      setoid_rewrite disjAssoc.
                      setoid_rewrite disjConf.
                      apply voiceDisjMonotonic.
                    }
                  }
                }
                {
                  apply robTrans with (q := (l' ∨ N')→).
                  {
                    apply robFlow2Conf.
                    apply robFlowTrans with (q := l ⊓ c→).
                    { apply IH. }
                    { apply robStatic. eautoflam. }
                  }
                  {
                    apply robStatic.
                    setoid_rewrite <- disjAssoc.
                    setoid_rewrite (disjComm l' r0).
                    setoid_rewrite disjAssoc.
                    algMagic.
                    eautoflam.
                  }
                }
                { rewrite H5. autoflam. }
              }
              {
                (* inductive case: p ≽ q derived w/ RAF_Fwd. *)
                (* showing conclusion 1. *)
                (* subcase: delegation in D2. *)
                (* integrity half *)
                setoid_rewrite H5.
                apply robConjL2.
                setoid_rewrite <- disjAssoc.
                setoid_rewrite (disjComm l' r0).
                setoid_rewrite disjAssoc.
                setoid_rewrite disjInteg.
                apply robDisjL.
                {
                  apply robProj_integ.
                  apply robTrans with (q := (pc→ ∧ l) ∧ ∇(q)).
                  { apply robWeakenD with (D' := D1) ; eautoflam. }
                  { apply robStatic. eautoflam. }
                }
                {
                  eapply robFwd with (r := r0) (D1 := D1) (D2 := D2) ; eauto.
                  {
                    setoid_replace ((pc→ ∧ l) ∧ ∇(l←)) with (pc→ ∧ l).
                    {
                      apply robTrans with (q := (pc→ ∧ l) ∧ ∇(q)).
                      { apply robWeakenD with (D' := D1) ; eautoflam. }
                      { apply robStatic. eautoflam. }
                    }
                    { rewrite conjAssoc. algMagic. }
                  }
                  {
                    apply robFlow2Integ.
                    apply robFlowTrans with (q := l ⊓ c→).
                    { apply IH. }
                    { apply robStatic. eautoflam. }
                  }
                  { autoflam. }
                }
              }
            }
            {
              (* inductive case: p ≽ q derived w/ RAF_Fwd. *)
              (* subcase: delegation in D2. *)
              (* showing conclusion 2. *)
              rewrite Heqdu, DelUse.disjPathTransfer, <- H2, <- HeqN' in HeqN.
              subst N.
              apply robDisjL.
              {
                eapply robTrans with (q := (pc→ ∧ l) ∧ ∇(q)).
                {
                  setoid_rewrite H5.
                  eapply robWeakenD with (D' := D1) ; eautoflam.
                }
                {
                  apply robConjL1.
                  eapply robStatic.
                  setoid_rewrite <- (confIntegBasis l) at 1.
                  eautoflam.
                }
              }
              {
                eapply robFwd with (r := r0) (D1 := D1) (D2 := D2) ; eauto.
                {
                  eapply robConjR.
                  {
                    apply robTrans with (q := (pc→ ∧ l) ∧ ∇(q)).
                    { apply robWeakenD with (D' := D1) ; eautoflam. }
                    { apply robStatic. eautoflam. }
                  }
                  {
                    eapply robTrans with (q := r0←).
                    { eautoflam. }
                    {
                      setoid_rewrite <- (voiceOfNative r0 rn0 H2).
                      eapply robActImpliesVoiceAct.
                      eapply robTrans with (q := (pc→ ∧ l) ∧ ∇(q)).
                      { apply robWeakenD with (D' := D1) ; eautoflam. }
                      {
                        apply robConjL1.
                        eapply robStatic.
                        setoid_rewrite <- (confIntegBasis l) at 1.
                        eautoflam.
                      }
                    }
                  }
                }
                {
                  eapply robTrans with (q := (pc ⊔ l ⊔ c←)→ ∧ (l ⊓ c→)←).
                  {
                    eapply IHn with (j := n2) (r := r) (N := N');
                      eauto with zarith.
                    {
                      destruct du' as [d path].
                      rewrite Heqdu in Hdel.
                      simpl in *. eauto.
                    }
                  }
                  {
                    apply robStatic. algMagic.
                    eexists. eapply AF_ConjR ; eautoflam.
                  }
                }
                { rewrite H5. autoflam. }
              }
            }
            {
              (* inductive case: p ≽ q derived w/ RAF_Fwd. *)
              (* subcase: delegation in D2. *)
              (* showing conclusion 3. *)
              rewrite Heqdu, DelUse.disjPathTransfer, <- H2, <- HeqN' in HeqN.
              subst N.
              eapply robFwd with (r := r0) (D1 := D1) (D2 := D2) ; eauto.
              {
                apply robConjR.
                {
                  apply robTrans with (q := (pc→ ∧ l) ∧ ∇(q)).
                  { apply robWeakenD with (D' := D1) ; eautoflam. }
                  { apply robStatic. eautoflam. }
                }
                {
                  apply robTrans with (q := ∇(r0→));
                    apply robStatic.
                  { erewrite voiceOfNativeConf ; eauto. eautoflam. }
                  {
                    setoid_rewrite <- disjAssoc.
                    setoid_rewrite (disjComm l' r0).
                    setoid_rewrite disjAssoc.
                    setoid_rewrite disjConf.
                    apply voiceDisjMonotonic.
                  }
                }
              }
              {
                apply robTrans with (q := (l' ∨ N')→).
                {
                  apply robTrans with (q := (l ⊓ c→)→).
                  { apply robStatic. algMagic. eautoflam. }
                  { apply robFlow2Conf. apply IH. }
                }
                {
                  apply robStatic.
                  setoid_rewrite <- disjAssoc.
                  setoid_rewrite (disjComm l' r0).
                  setoid_rewrite disjAssoc.
                  algMagic.
                  eautoflam.
                }
              }
              { rewrite H5. autoflam. }
            }
          }
        }
      }
    }
  }

  intros.
  split ; [|split] ; try (eapply H ; eauto ; fail).
  intro.
  destruct (name_dec nn cn).
  {
    subst nn.
    apply robTrans with (q := N) ; apply robStatic.
    {
      subst N.
      eapply delHostOnPath ; eauto.
      destruct (DelUse.getPath du).
      { subst k. contradict H8. simpl. omega. }
      {
        intro Hfalse. symmetry in Hfalse. contradict Hfalse.
        apply List.nil_cons.
      }
    }
    { setoid_rewrite disjConf. eautoflam. }
  }
  { eapply H ; eauto. }
Qed.
