(* begin hide *)
(* 45678901234567890123456789012345678901234567890123456789012345678901234567 *)
Require Import Morphisms.
Require Import Omega.
Require Import StrongInduction.

Require Export FLAM_Algebra.
Require Export FLAM_DelSet.
Require Export FLAM_DelUseSet.
Require Export FLAM_Norm.
Require Export FLAM_PH.
(* end hide *)

Import DelUseSet.

Notation "D1 =D= D2" := (sameSet D1 D2)
  (at level 70, no associativity).
Notation "H1 =H= H2" := (PH.phEq H1 H2)
  (at level 70, no associativity).

(**
 * Principal voice
 *)

Definition voice_ q : sigT normalized :=
  match projT2 (norm q) with
  | norm_J a pfa => existT _ a← (norm_integ _ pfa)
  | norm_conf a pfa => existT _ a← (norm_integ _ pfa)
  | norm_integ a pfa => existT _ a← (norm_integ _ pfa)
  | norm_decomp a b pfa pfb =>
      existT _ (a ∧ b)←
        (norm_integ _ (normJ_conj _ _ pfa pfb))
  end.

Definition voice q := projT1 (voice_ q).

Notation "∇( q )" := (voice q).

Instance voiceProper : Proper (prinEq ==> prinEq) voice.
Proof.
  intro. intros.
  pose proof (normProper _ _ H).
  unfold voice, voice_.
  unfold normEq in *.
  destruct (norm x), (norm y).
  destruct n, n0 ; simpl in * ; autoalg ; try (rewrite H0 ; autoalg ; fail).
    assert (p == ⊥).
      eapply normJIsConfImpliesBot ; eauto.
    assert (p0 == ⊥).
      eapply normJConfIsBotImpliesBot ; eautoalg.
    rewrite H1, H2.
    autoalg.

    assert (p == p0 /\ p == q).
      apply normJDecomp with (p := p) (q := p0) (r := q) ; auto.
    destruct H1.
    rewrite <- H1, <- H2.
    autoalg.

    assert (p0 == ⊥).
      apply normJIsConfImpliesBot with (q := p) ; autoalg.
    assert (p == ⊥).
      apply normJConfIsBotImpliesBot ; eautoalg.
    rewrite H1, H2.
    autoalg.

    assert (p == p0).
      eapply normJConfEq ; eauto.
    rewrite H1.
    autoalg.

    assert (p == ⊥).
      apply normJConfIsBotImpliesBot ; auto.
      eapply confIsIntegImpliesBot.
      eauto.
    assert (p0 == ⊥).
      apply normJIntegIsBotImpliesBot ; auto.
      rewrite H1 ,botConf in H0.
      autoalg.
    rewrite H1, H2.
    autoalg.

    assert (q == ⊥).
      apply normJIntegIsBotImpliesBot ; auto.
      setoid_replace ⊥ with (p→←) by autoalg.
      setoid_replace (q←) with ((p0→ ∧ q←)←).
        rewrite H0. autoalg.
        rewrite conjInteg. autoalg.
    assert (p == p0).
      apply normJConfEq ; auto.
      setoid_replace (p→) with (p→→) by autoalg.
      setoid_replace (p0→) with ((p0→ ∧ q←)→).
        rewrite H0. autoalg.
        rewrite conjConf. autoalg.
    rewrite H1, H2.
    autoalg.

    rewrite <- integIdemp.
    rewrite H0.
    autoalg.

    assert (p0 == ⊥).
      apply normJConfIsBotImpliesBot ; auto.
      eapply confIsIntegImpliesBot.
      symmetry.
      eauto.
    rewrite H1 in *.
    rewrite botConf, botInteg in *.
    auto.

    assert (p0 == ⊥).
      apply normJConfIsBotImpliesBot ; auto.
      setoid_replace ⊥ with (p←→) by autoalg.
      setoid_replace (p0→) with ((p0→ ∧ q←)→).
        rewrite H0. autoalg.
        rewrite conjConf. autoalg.
    assert (p == q).
      apply normJIntegEq ; auto.
      setoid_replace (p←) with (p←←) by autoalg.
      setoid_replace (q←) with ((p0→ ∧ q←)←).
        rewrite H0. autoalg.
        rewrite conjInteg. autoalg.
    rewrite H1, H2.
    autoalg.

    assert (p0 == p /\ p0 == q).
      apply normJDecomp ; autoalg.
    destruct H1.
    rewrite <- H1, <- H2.
    autoalg.

    assert (q == ⊥).
      apply normJIntegIsBotImpliesBot ; auto.
      setoid_replace ⊥ with (p0→←) by autoalg.
      setoid_replace (q←) with ((p→ ∧ q←)←).
        rewrite H0. autoalg.
        rewrite conjInteg. autoalg.
    assert (p == p0).
      apply normJConfEq ; auto.
      setoid_replace (p0→) with (p0→→) by autoalg.
      setoid_replace (p→) with ((p→ ∧ q←)→).
        rewrite H0. autoalg.
        rewrite conjConf. autoalg.
    rewrite H1, H2.
    autoalg.

    assert (p == ⊥).
      apply normJConfIsBotImpliesBot ; auto.
      setoid_replace ⊥ with (p0←→) by autoalg.
      setoid_replace (p→) with ((p→ ∧ q←)→).
        rewrite H0. autoalg.
        rewrite conjConf. autoalg.
    assert (q == p0).
      apply normJIntegEq ; auto.
      setoid_replace (p0←) with (p0←←) by autoalg.
      setoid_replace (q←) with ((p→ ∧ q←)←).
        rewrite H0. autoalg.
        rewrite conjInteg. autoalg.
    rewrite H1, H2.
    autoalg.

    assert (p == p0).
      apply normJConfEq ; auto.
      setoid_replace (p→) with ((p→ ∧ q←)→) by (rewrite conjConf ; autoalg).
      setoid_replace (p0→) with ((p0→ ∧ q0←)→) by (rewrite conjConf ; autoalg).
      apply confSubst.
      auto.
    assert (q == q0).
      apply normJIntegEq ; auto.
      setoid_replace (q←) with ((p→ ∧ q←)←) by (rewrite conjInteg ; autoalg).
      setoid_replace (q0←) with ((p0→ ∧ q0←)←) by (rewrite conjInteg ; autoalg).
      apply integSubst.
      auto.
    rewrite H1, H2.
    autoalg.
Qed.

(**
 ** Properties
 *)

Lemma voiceIsInteg:
  forall p, exists q, ∇(p) == q←.
Proof.
  intro.
  unfold voice, voice_.
  destruct (projT2 (norm p)) ; simpl ; eautoalg.
Qed.

Hint Resolve voiceIsInteg : prinalg.

Lemma confOfVoice:
  forall p, ∇(p)→ == ⊥.
Proof.
  intros.
  pose proof voiceIsInteg p.
  destruct H.
  rewrite H.
  autoalg.
Qed.

Hint Resolve confOfVoice : prinalg.  Hint Rewrite confOfVoice : prinalg.

Lemma integOfVoice:
  forall p, ∇(p)← == ∇(p).
Proof.
  intros.
  pose proof voiceIsInteg p.
  destruct H.
  rewrite H.
  autoalg.
Qed.

Hint Resolve integOfVoice : prinalg. Hint Rewrite integOfVoice : prinalg.

Lemma voiceOfNorm:
  forall p (pfp : normJ p) q (pfq : normJ q), ∇(p→ ∧ q←) = (p ∧ q)←.
Proof.
  intros.
  unfold voice, voice_.
  pose proof (normIdemp p pfp q pfq).
  do 2 destruct H.
  rewrite H.
  autoalg.
Qed.

Hint Resolve voiceOfNorm : prinalg.

Lemma voiceOfNormJ:
  forall p (pfp : normJ p), ∇(p) = p←.
Proof.
  intros.
  unfold voice, voice_.
  pose proof normIdemp_J p pfp.
  destruct H.
  rewrite H.
  autoalg.
Qed.

Hint Resolve voiceOfNormJ : prinalg.

Lemma voiceOfNative:
  forall p pn, p = native pn -> ∇(p) = p←.
Proof.
  intros.
  apply voiceOfNormJ.
  rewrite H.
  autonorm.
Qed.

Hint Resolve voiceOfNative : prinalg.

Lemma voiceOfConj:
  forall p q, ∇(p ∧ q) == ∇(p) ∧ ∇(q).
Proof.
  intros.
  pose proof (normForm p).
  pose proof (normForm q).
  do 4 destruct H.
  do 4 destruct H0.
  rewrite H.
  rewrite H0.
  rewrite conjPermute2in4.
  rewrite <- conjConf.
  rewrite <- conjInteg.
  repeat (rewrite voiceOfNorm) ; autonorm.
  rewrite conjPermute2in4.
  autoalg.
Qed.

Hint Resolve voiceOfConj : prinalg.

Lemma voiceOfNormJConf:
  forall p (pfp : normJ p), ∇(p→) = p←.
Proof.
  intros.
  unfold voice, voice_.
  simpl.
  pose proof normIdemp_J p pfp.
  destruct H.
  rewrite H.
  auto.
Qed.

Hint Resolve voiceOfNormJConf : prinalg.

Lemma voiceOfNativeConf:
  forall p pn, p = native pn -> ∇(p→) = p←.
Proof.
  intros.
  apply voiceOfNormJConf.
  rewrite H.
  autonorm.
Qed.

Hint Resolve voiceOfNativeConf : prinalg.

Lemma voiceOfInteg:
  forall p, ∇(p←) == p←.
Proof.
  intros.
  pose proof (normForm p).
  do 4 destruct H.
  rewrite H.
  repeat (rewrite conjInteg).
  autoalg.
  setoid_replace (x0←) with (⊥→ ∧ x0←) by autoalg.
  rewrite voiceOfNorm ; autoalg ; autonorm.
Qed.

Hint Resolve voiceOfInteg : prinalg.

Lemma voiceIdemp:
  forall p, ∇(∇(p)) == ∇(p).
Proof.
  intro.
  assert (exists q, ∇(p) == q←) by apply voiceIsInteg.
  destruct H.
  rewrite H.
  rewrite voiceOfInteg.
  autoalg.
Qed.

Hint Resolve voiceIdemp : prinalg.       Hint Rewrite voiceIdemp : prinalg.

(**
 * Acts-for
 *)

(*
 * Notation:
 *   [[H;c;pc;l ⊢ p ≽ q # n]] means p acts for q in context H;c;pc;l, and the
 *   derivation consists of n+1 rule applications.
 *)
Reserved Notation "[ H ; c ; pc ; l ⊢ p ≽ q # D ; n ]"
  (at level 40).
Reserved Notation "[ H ; c ; pc ; l ⊢ p ⊑ q # D ; n ]"
  (at level 40, only parsing).
Reserved Notation "[ H ; c ; pc ; l ⊩ p ≽ q # D ; n ]"
  (at level 40).
Reserved Notation "[ H ; c ; pc ; l ⊩ p ⊑ q # D ; n ]"
  (at level 40, only parsing).
Reserved Notation "[ c ⊢ p ≽ q # n ]" (at level 40).

Inductive actsfor
    : PH.t -> name -> Label -> Label -> Prin -> Prin
        -> DelUseSet.t -> nat -> Prop :=
  | AF_Bot : forall H c pc l p q D,
      q == ⊥ ->
      D =D= empty ->
      [H;c;pc;l ⊢ p ≽ q # D;0]

  | AF_Top : forall H c pc l p q D,
      p == ⊤ ->
      D =D= empty ->
      [H;c;pc;l ⊢ p ≽ q # D;0]

  | AF_Refl : forall H c pc l p q D,
      p == q ->
      D =D= empty ->
      [H;c;pc;l ⊢ p ≽ q # D;0]

  | AF_Proj_conf : forall H c pc l p q D n n1 p' q',
      p == p'→ ->
      q == q'→ ->
      [H;c;pc;l ⊢ p' ≽ q' # D;n1] ->
      n = n1 + 1 ->
      [H;c;pc;l ⊢ p ≽ q # D;n]

  | AF_Proj_integ : forall H c pc l p q D n n1 p' q',
      p == p'← ->
      q == q'← ->
      [H;c;pc;l ⊢ p' ≽ q' # D;n1] ->
      n = n1 + 1 ->
      [H;c;pc;l ⊢ p ≽ q # D;n]

  | AF_ProjR_conf : forall H c pc l p q D,
      q == p→ ->
      D =D= empty ->
      [H;c;pc;l ⊢ p ≽ q # D;0]

  | AF_ProjR_integ : forall H c pc l p q D,
      q == p← ->
      D =D= empty ->
      [H;c;pc;l ⊢ p ≽ q # D;0]

  | AF_ProjR_owns : forall H c pc l p q q' D,
      q == p:q' ->
      D =D= empty ->
      [H;c;pc;l ⊢ p ≽ q # D;0]

  | AF_Own1 : forall H c pc l p q D D1 D2 n n1 n2 o r o' r',
      p == o:r ->
      q == o':r' ->
      [H;c;pc;l ⊢ o ≽ o' # D1;n1] ->
      [H;c;pc;l ⊢ r ≽ r' # D2;n2] ->
      D =D= union D1 D2 ->
      n = n1 + n2 + 1 ->
      [H;c;pc;l ⊢ p ≽ q # D;n]

  | AF_Own2 : forall H c pc l p q D D1 D2 n n1 n2 o r o' r',
      p == o:r ->
      q == o':r' ->
      [H;c;pc;l ⊢ o ≽ o' # D1;n1] ->
      [H;c;pc;l ⊢ r ≽ o':r' # D2;n2] ->
      D =D= union D1 D2 ->
      n = n1 + n2 + 1 ->
      [H;c;pc;l ⊢ p ≽ q # D;n]

  | AF_ConjL1 : forall H c pc l p q D n n1 p1 p2,
      p == p1 ∧ p2 ->
      [H;c;pc;l ⊢ p1 ≽ q # D;n1] ->
      n = n1 + 1 ->
      [H;c;pc;l ⊢ p ≽ q # D;n]

  | AF_ConjR : forall H c pc l p q D D1 D2 n n1 n2 q1 q2,
      q == q1 ∧ q2 ->
      [H;c;pc;l ⊢ p ≽ q1 # D1;n1] ->
      [H;c;pc;l ⊢ p ≽ q2 # D2;n2] ->
      D =D= union D1 D2 ->
      n = n1 + n2 + 1 ->
      [H;c;pc;l ⊢ p ≽ q # D;n]

  | AF_DisjL : forall H c pc l p q D D1 D2 n n1 n2 p1 p2,
      p == p1 ∨ p2 ->
      [H;c;pc;l ⊢ p1 ≽ q # D1;n1] ->
      [H;c;pc;l ⊢ p2 ≽ q # D2;n2] ->
      D =D= union D1 D2 ->
      n = n1 + n2 + 1 ->
      [H;c;pc;l ⊢ p ≽ q # D;n]

  | AF_DisjR1 : forall H c pc l p q D n n1 q1 q2,
      q == q1 ∨ q2 ->
      [H;c;pc;l ⊢ p ≽ q1 # D;n1] ->
      n = n1 + 1 ->
      [H;c;pc;l ⊢ p ≽ q # D;n]

  | AF_Trans : forall H c pc l p q r D D1 D2 n n1 n2,
      [H;c;pc;l ⊢ p ≽ q # D1;n1] ->
      [H;c;pc;l ⊢ q ≽ r # D2;n2] ->
      D =D= union D1 D2 ->
      n = n1 + n2 + 1 ->
      [H;c;pc;l ⊢ p ≽ r # D;n]

  | AF_Del : forall H c pc l p q p' q' l' d D,
      p == p' ->
      q == q' ->
      l == l' ->
      d = DelT.make p' q' l' c ->
      PH.In d H ->
      D =D= init d ->
      [H;c;pc;l ⊢ p ≽ q # D;0]

  | AF_Fwd : forall H cn pc l p q rn c r D D1 D2 n n1 n2,
      c = native cn ->
      r = native rn ->
      [H;cn;pc;l ⊩ r ≽ pc→ ∧ l # D1;n1] ->
      [H;rn;pc ⊔ l ⊔ c←;l ⊓ c→ ⊢ p ≽ q # D2;n2] ->
      D =D= union D1 (transfer D2 rn) ->
      n = n1 + n2 + 1 ->
      [H;cn;pc;l ⊢ p ≽ q # D;n]

  | AF_Weaken : forall H c pc l p q H1 H2 pc' l' D D1 D2 D3 n n1 n2 n3,
      [H1;c;pc';l' ⊢ p ≽ q # D1;n1] ->
      [H1;c;pc ⊔ l';l ⊩ pc ⊑ pc' # D2;n2] ->
      [H1;c;pc ⊔ l';l ⊩ l' ⊑ l # D3;n3] ->
      H =H= PH.union H1 H2 ->
      D =D= union D1 (union D2 D3) ->
      n = n1 + n2 + n3 + 1 ->
      [H;c;pc;l ⊢ p ≽ q # D;n]

with actsforRobust
    : PH.t -> name -> Label -> Label -> Prin -> Prin
        -> DelUseSet.t -> nat -> Prop :=
  | RAF_Static : forall H c pc l p q n n1 D,
      [c ⊢ p ≽ q # n1] ->
      D =D= empty ->
      n = n1 + 1 ->
      [H;c;pc;l ⊩ p ≽ q # D;n]

  | RAF_Lift : forall H c pc l p q D D1 D2 D3 n n1 n2 n3,
      [H;c;pc;l ∧ ∇(q) ⊢ p ≽ q # D1;n1] ->
      [H;c;pc;l ⊩ pc ≽ ∇(q) # D2;n2] ->
      [H;c;pc;l ⊩ ∇(p→) ≽ ∇(q→) # D3;n3] ->
      D =D= union D1 (union D2 D3) ->
      n = n1 + n2 + n3 + 1 ->
      [H;c;pc;l ⊩ p ≽ q # D;n]

  | RAF_LiftPC : forall H c pc l p q D n n1 q',
      p == pc ->
      q == ∇(q') ->
      [H;c;pc;l ∧ ∇(q') ⊢ pc ≽ ∇(q') # D;n1] ->
      n = n1 + 1 ->
      [H;c;pc;l ⊩ p ≽ q # D;n]

  | RAF_ConjR : forall H c pc l p q D D1 D2 n n1 n2 q1 q2,
      q == q1 ∧ q2 ->
      [H;c;pc;l ⊩ p ≽ q1 # D1;n1] ->
      [H;c;pc;l ⊩ p ≽ q2 # D2;n2] ->
      D =D= union D1 D2 ->
      n = n1 + n2 + 1 ->
      [H;c;pc;l ⊩ p ≽ q # D;n]

  | RAF_DisjL : forall H c pc l p q D D1 D2 n n1 n2 p1 p2,
      p == p1 ∨ p2 ->
      [H;c;pc;l ⊩ p1 ≽ q # D1;n1] ->
      [H;c;pc;l ⊩ p2 ≽ q # D2;n2] ->
      D =D= union D1 D2 ->
      n = n1 + n2 + 1 ->
      [H;c;pc;l ⊩ p ≽ q # D;n]

  | RAF_Weaken : forall H c pc l p q H1 H2 pc' l' D D1 D2 D3 n n1 n2 n3,
      [H1;c;pc';l' ⊩ p ≽ q # D1;n1] ->
      [H1;c;pc ⊔ l';l ⊩ pc ⊑ pc' # D2;n2] ->
      [H1;c;pc ⊔ l';l ⊩ l' ⊑ l # D3;n3] ->
      H =H= PH.union H1 H2 ->
      D =D= union D1 (union D2 D3) ->
      n = n1 + n2 + n3 + 1 ->
      [H;c;pc;l ⊩ p ≽ q # D;n]

  | RAF_Trans : forall H c pc l p q r D D1 D2 D3 n n1 n2 n3,
      [H;c;pc;l ⊩ p ≽ q # D1;n1] ->
      [H;c;pc;l ⊩ q ≽ r # D2;n2] ->
      [H;c;pc;l ⊩ pc ≽ ∇(r→) # D3;n3] ->
      D =D= union D1 (union D2 D3) ->
      n = n1 + n2 + n3 + 1 ->
      [H;c;pc;l ⊩ p ≽ r # D;n]

  | RAF_Fwd : forall H cn pc l p q rn c r D D1 D2 n n1 n2,
      c = native cn ->
      r = native rn ->
      [H;cn;pc;l ⊩ r ≽ (pc→ ∧ l) ∧ ∇(q) # D1;n1] ->
      [H;rn;pc ⊔ l ⊔ c←;l ⊓ c→ ⊩ p ≽ q # D2;n2] ->
      D =D= union D1 (transfer D2 rn) ->
      n = n1 + n2 + 1 ->
      [H;cn;pc;l ⊩ p ≽ q # D;n]

where "[ H ; c ; pc ; l ⊢ p ≽ q # D ; n ]" :=
  (actsfor H c pc l p q D n)
and "[ H ; c ; pc ; l ⊢ p ⊑ q # D ; n ]" :=
  ([H;c;pc;l ⊢ q→ ∧ p← ≽ p→ ∧ q← # D;n])
and "[ c ⊢ p ≽ q # n ]" :=
  ([PH.empty;c;⊤→;⊤← ⊢ p ≽ q # empty;n])
and "[ H ; c ; pc ; l ⊩ p ≽ q # D ; n ]" :=
  (actsforRobust H c pc l p q D n)
and "[ H ; c ; pc ; l ⊩ p ⊑ q # D ; n ]" :=
  ([H;c;pc;l ⊩ q→ ∧ p← ≽ p→ ∧ q← # D;n]).

Notation "[ c ⊢ p ≽ q ]" := (exists n, [c ⊢ p ≽ q # n])
  (at level 40).
Notation "[ c ⊢ p ⊑ q # n ]" := ([PH.empty;c;⊤→;⊤← ⊢ p ⊑ q # empty;n])
  (at level 40).
Notation "[ c ⊢ p ⊑ q ]" := (exists n, [c ⊢ p ⊑ q # n])
  (at level 40).
Notation "[ c ⊩ p ≽ q # n ]" := ([PH.empty;c;⊤→;⊤← ⊩ p ≽ q # empty;n])
  (at level 40).
Notation "[ c ⊩ p ≽ q ]" := (exists n, [c ⊩ p ≽ q # n])
  (at level 40).
Notation "[ c ⊩ p ⊑ q # n ]" := ([PH.empty;c;⊤→;⊤← ⊩ p ⊑ q # empty;n])
  (at level 40).
Notation "[ c ⊩ p ⊑ q ]" := (exists n, [c ⊩ p ⊑ q # n])
  (at level 40).
Notation "[ H ; c ; pc ; l ⊢ p ≽ q | D ]" :=
    (exists D', (exists n, [H;c;pc;l ⊢ p ≽ q # D';n]) /\ included D' D)
  (at level 40).
Notation "[ H ; c ; pc ; l ⊢ p ⊑ q | D ]" :=
    (exists D', (exists n, [H;c;pc;l ⊢ p ⊑ q # D';n]) /\ included D' D)
  (at level 40).
Notation "[ H ; c ; pc ; l ⊩ p ≽ q | D ]" :=
    (exists D', (exists n, [H;c;pc;l ⊩ p ≽ q # D';n]) /\ included D' D)
  (at level 40).
Notation "[ H ; c ; pc ; l ⊩ p ⊑ q | D ]" :=
    (exists D', (exists n, [H;c;pc;l ⊩ p ⊑ q # D';n]) /\ included D' D)
  (at level 40).

Hint Resolve AF_Bot : flam.
Hint Resolve AF_Top : flam.
Hint Resolve AF_Refl : flam.
Hint Resolve AF_Proj_conf : flam.
Hint Resolve AF_Proj_integ : flam.
Hint Resolve AF_ProjR_conf : flam.
Hint Resolve AF_ProjR_integ : flam.
Hint Resolve AF_ProjR_owns : flam.
Hint Resolve AF_Own1 : flam.
Hint Resolve AF_Own2 : flam.
Hint Resolve AF_ConjL1 : flam.
Hint Resolve AF_ConjR : flam.
Hint Resolve AF_DisjL : flam.
Hint Resolve AF_DisjR1 : flam.
Hint Resolve AF_Del : flam.
Hint Resolve RAF_Static : flam.
Hint Resolve RAF_Lift : flam.
Hint Resolve RAF_LiftPC : flam.
Hint Resolve RAF_ConjR : flam.
Hint Resolve RAF_DisjL : flam.
(* begin hide *)
(*
These three make the auto magic diverge.
Hint Resolve AF_Trans : flam.
Hint Resolve AF_Fwd : flam.
Hint Resolve AF_Weaken : flam.
*)
(* end hide *)

Tactic Notation "autoflam" := auto using refl with flam.
Tactic Notation "eautoflam" := eauto using refl with flam.

(* Auto with FLAM rules and integer arithmetic. *)
Tactic Notation "autoflamz" := auto using refl with flam zarith.

(* Auto with existentials, FLAM rules, and integer arithmetic. *)
Tactic Notation "eautoflamz" := eauto using refl with flam zarith.

Ltac algMagic :=
  repeat progress (
    try setoid_rewrite disjConjAbsorb;
    try setoid_rewrite conjDisjAbsorb;
    try setoid_rewrite conjIdemp;
    try setoid_rewrite disjIdemp;
    try setoid_rewrite conjIdent;
    try setoid_rewrite disjIdent;
    try setoid_rewrite conjTop;
    try setoid_rewrite disjBot;
    try setoid_rewrite ownDistConj;
    try setoid_rewrite ownDistDisj;
    try setoid_rewrite confIdemp;
    try setoid_rewrite integIdemp;
    try setoid_rewrite ownIdemp;
    try setoid_rewrite confInteg;
    try setoid_rewrite integConf;
    try setoid_rewrite confDisjInteg;
    try setoid_rewrite ownsSelf;
    try setoid_rewrite ownsBot;
    try setoid_rewrite conjDistOwn;
    try setoid_rewrite disjDistOwn;
    try setoid_rewrite ownConf;
    try setoid_rewrite ownInteg;
    try setoid_rewrite confIntegBasis;
    try setoid_rewrite botConf;
    try setoid_rewrite botInteg;
    try setoid_rewrite botOwn;
    try setoid_rewrite conjIdentL;
    try setoid_rewrite disjIdentL;
    try setoid_rewrite conjTopL;
    try setoid_rewrite disjBotL;
    try setoid_rewrite integDisjConf;
    try setoid_rewrite conjOwnConfL;
    try setoid_rewrite conjOwnConfR;
    try setoid_rewrite conjOwnIntegL;
    try setoid_rewrite conjOwnIntegR;
    try setoid_rewrite disjOwnConfL;
    try setoid_rewrite disjOwnConfR;
    try setoid_rewrite disjOwnIntegL;
    try setoid_rewrite disjOwnIntegR;
    try setoid_rewrite joinTopConf;
    try setoid_rewrite joinTopConf';
    try setoid_rewrite meetTopInteg;
    try setoid_rewrite meetTopInteg';
    try setoid_rewrite confJoin;
    try setoid_rewrite integJoin;
    try setoid_rewrite confMeet;
    try setoid_rewrite integMeet;
    try setoid_rewrite conjConf;
    try setoid_rewrite disjConf;
    try setoid_rewrite conjInteg;
    try setoid_rewrite disjInteg;
    try setoid_rewrite confOfVoice;
    try setoid_rewrite integOfVoice;
    try setoid_rewrite voiceOfConj;
    try setoid_rewrite voiceOfInteg;
    try setoid_rewrite voiceIdemp;
    try setoid_replace (∇(⊥)) with ⊥ by (unfold voice ; simpl ; autoalg)
  );
  autoalg;
  match goal with
  | [|- ?X == ?Y] => ring_simplify X Y
  | [|- context[[_;_;_;_ ⊢ ?X ≽ ?Y # _;_]]] => ring_simplify X Y
  | _ => trivial
  end;
  simpl;
  autoalg.

Axiom actsForEq:
  forall c p q, [c ⊢ p ≽ q] -> [c ⊢ q ≽ p] -> p == q.

(**
 ** Properties
 *)
Lemma actsforProperHelper:
  forall n H0 pc0 l0 p0 q0 D0 H c pc l p q D,
       H0 =H= H
    -> pc0 == pc
    -> l0 == l
    -> p0 == p
    -> q0 == q
    -> D0 =D= D
    ->   ([H0;c;pc0;l0 ⊢ p0 ≽ q0 # D0;n] -> [H;c;pc;l ⊢ p ≽ q # D;n])
      /\ ([H0;c;pc0;l0 ⊩ p0 ≽ q0 # D0;n] -> [H;c;pc;l ⊩ p ≽ q # D;n]).
Proof.
  induction n as [|n IHn] using strong_induction;
      intros H0 pc0 l0 p0 q0 D0;
      intros H c pc l p q D;
      intros HH Hpc Hl Hp Hq HD ; split ; intro Hderiv.
  {
    inversion Hderiv ; try (omega).
    {
      eapply AF_Bot.
      rewrite <- Hq. auto.
      eapply sameSetTrans with (y := D0) ; eauto.
      symmetry. auto.
    }
    {
      eapply AF_Top.
      rewrite <- Hp. auto.
      eapply sameSetTrans with (y := D0) ; eauto.
      symmetry. auto.
    }
    {
      eapply AF_Refl.
      rewrite <- Hp, <- Hq. auto.
      eapply sameSetTrans with (y := D0) ; eauto.
      symmetry. auto.
    }
    {
      eapply AF_ProjR_conf ; eauto.
      rewrite <- Hp, <- Hq. auto.
      eapply sameSetTrans with (y := D0) ; eauto.
      symmetry. auto.
    }
    {
      eapply AF_ProjR_integ ; eauto.
      rewrite <- Hp, <- Hq. auto.
      eapply sameSetTrans with (y := D0) ; eauto.
      symmetry. auto.
    }
    {
      eapply AF_ProjR_owns ; eauto.
      rewrite <- Hp, <- Hq. eauto.
      eapply sameSetTrans with (y := D0) ; eauto.
      symmetry. auto.
    }
    {
      eapply AF_Del ; eauto.
      { rewrite <- Hp. auto. }
      { rewrite <- Hq. auto. }
      { rewrite <- Hl. auto. }
      { rewrite <- HH. auto. }
      { rewrite <- HD. auto. }
    }
  }
  { inversion Hderiv ; omega. }
  {
    inversion Hderiv.
    {
      eapply AF_Proj_conf ; eauto.
        rewrite <- Hp. eauto.
        rewrite <- Hq. eauto.
        eapply IHn with (j := n1) ; eauto with zarith ; autoalg.
    }
    {
      eapply AF_Proj_integ ; eauto.
        rewrite <- Hp. eauto.
        rewrite <- Hq. eauto.
        eapply IHn with (j := n1) ; eauto with zarith ; autoalg.
    }
    {
      eapply AF_Own1 with (D1 := D2) (D2 := D3) ; eauto.
        rewrite <- Hp. eauto.
        rewrite <- Hq. eauto.
        eapply IHn with (j := n1) ; eauto with zarith ; autoflam.
        eapply IHn with (j := n2) ; eauto with zarith ; autoflam.
        rewrite <- HD. eauto.
    }
    {
      eapply AF_Own2 with (D1 := D2) (D2 := D3) ; eauto.
        rewrite <- Hp. eauto.
        rewrite <- Hq. eauto.
        eapply IHn with (j := n1) ; eauto with zarith ; autoflam.
        eapply IHn with (j := n2) (p0 := r) (q0 := o':r');
            eauto with zarith;
            autoflam.
        rewrite <- HD. auto.
    }
    {
      eapply AF_ConjL1 ; eauto.
        rewrite <- Hp. eauto.
        eapply IHn with (j := n1) ; eauto with zarith ; autoalg.
    }
    {
      eapply AF_ConjR with (D1 := D2) (D2 := D3) ; eauto.
        rewrite <- Hq. eauto.
        eapply IHn with (j := n1) ; eauto with zarith ; autoflam.
        eapply IHn with (j := n2) ; eauto with zarith ; autoflam.
        rewrite <- HD. auto.
    }
    {
      eapply AF_DisjL with (D1 := D2) (D2 := D3) ; eauto.
        rewrite <- Hp. eauto.
        eapply IHn with (j := n1) ; eauto with zarith ; autoflam.
        eapply IHn with (j := n2) ; eauto with zarith ; autoflam.
        rewrite <- HD. auto.
    }
    {
      eapply AF_DisjR1 ; eauto.
        rewrite <- Hq. eauto.
        eapply IHn with (j := n1) ; eauto with zarith ; autoalg.
    }
    {
      eapply AF_Trans with (q := q1) (D1 := D2) (D2 := D3) ; eauto.
        eapply IHn with (j := n1) ; eauto with zarith ; autoflam.
        eapply IHn with (j := n2) ; eauto with zarith ; autoflam.
        rewrite <- HD. auto.
    }
    {
      eapply AF_Fwd with (rn := rn) (D1 := D2) (D2 := D3) ; eauto.
        eapply IHn with (j := n1);
            eauto with zarith;
            try (setoid_replace pc0 with pc by auto);
            try (setoid_replace l0 with l by auto);
            autoflam.
        eapply IHn with (j := n2);
            eauto with zarith;
            try (setoid_replace pc0 with pc by auto);
            try (setoid_replace l0 with l by auto);
            autoflam.
        rewrite <- HD. auto.
    }
    {
      eapply AF_Weaken
          with (pc' := pc') (l' := l') (H1 := H2)
            (D1 := D2) (D2 := D3) (D3 := D4) ; eauto.
        eapply IHn with (j := n1) ; eauto with zarith ; autoflam.
        eapply IHn with (j := n2);
            eauto with zarith;
            try (setoid_replace pc0 with pc by auto);
            autoflam.
        eapply IHn with (j := n3);
            eauto with zarith;
            try (setoid_replace pc0 with pc by auto);
            try (setoid_replace l0 with l by auto);
            autoflam.
        rewrite <- HH. eauto.
        rewrite <- HD. auto.
    }
  }
  {
    inversion Hderiv.
    {
      eapply RAF_Static ; eauto.
        eapply IHn with (j := n1) (H0 := PH.empty) (D0 := empty);
            eauto with zarith;
            autoflam.
        rewrite <- HD. auto.
    }
    {
      eapply RAF_Lift with (D1 := D2) (D2 := D3) (D3 := D4) ; eauto.
        eapply IHn with (j := n1);
            eauto with zarith;
            try (setoid_replace l0 with l by auto);
            try (setoid_replace q0 with q by auto);
            autoflam.
        eapply IHn with (j := n2);
            eauto with zarith;
            try (setoid_replace q0 with q by auto);
            autoflam.
        eapply IHn with (j := n3);
            eauto with zarith;
            try (setoid_replace p0 with p by auto);
            try (setoid_replace q0 with q by auto);
            autoflam.
        rewrite <- HD. auto.
    }
    {
      eapply RAF_LiftPC ; eauto.
        rewrite <- Hp, <- Hpc. auto.
        rewrite <- Hq. eauto.
        eapply IHn with (j := n1) (p0 := pc0) (q0 := ∇(q'));
            eauto with zarith;
            try (setoid_replace l0 with l by auto);
            autoflam.
    }
    {
      eapply RAF_ConjR with (D1 := D2) (D2 := D3) ; eauto.
        rewrite <- Hq. eauto.
        eapply IHn with (j := n1) ; eauto with zarith ; autoflam.
        eapply IHn with (j := n2) ; eauto with zarith ; autoflam.
        rewrite <- HD. auto.
    }
    {
      eapply RAF_DisjL with (D1 := D2) (D2 := D3) ; eauto.
        rewrite <- Hp. eauto.
        eapply IHn with (j := n1) ; eauto with zarith ; autoflam.
        eapply IHn with (j := n2) ; eauto with zarith ; autoflam.
        rewrite <- HD. auto.
    }
    {
      eapply RAF_Weaken
          with (pc' := pc') (l' := l') (H1 := H2)
            (D1 := D2) (D2 := D3) (D3 := D4) ; eauto.
        eapply IHn with (j := n1) ; eauto with zarith ; autoflam.
        eapply IHn with (j := n2);
            eauto with zarith;
            try (setoid_replace pc0 with pc by auto);
            autoflam.
        eapply IHn with (j := n3);
            eauto with zarith;
            try (setoid_replace pc0 with pc by auto);
            try (setoid_replace l0 with l by auto);
            autoflam.
        rewrite <- HH. eauto.
        rewrite <- HD. auto.
    }
    {
      eapply RAF_Trans with (q := q1) (D1 := D2) (D2 := D3) (D3 := D4) ; eauto.
        eapply IHn with (j := n1) ; eauto with zarith ; autoflam.
        eapply IHn with (j := n2) ; eauto with zarith ; autoflam.
        eapply IHn with (j := n3);
            eauto with zarith;
            try (setoid_replace q0 with q by auto);
            autoflam.
        rewrite <- HD. auto.
    }
    {
      eapply RAF_Fwd with (rn := rn) (D1 := D2) (D2 := D3) ; eauto.
        eapply IHn with (j := n1);
            eauto with zarith;
            try (setoid_replace pc0 with pc by auto);
            try (setoid_replace l0 with l by auto);
            try (setoid_replace q0 with q by auto);
            autoflam.

        eapply IHn with (j := n2);
            eauto with zarith;
            try (setoid_replace pc0 with pc by auto);
            try (setoid_replace l0 with l by auto);
            autoflam.

        rewrite <- HD. auto.
    }
  }
Qed.

Instance actsforProper:
  Proper
    (PH.phEq ==> eq ==> prinEq ==> prinEq ==> prinEq ==> prinEq
      ==> sameSet ==> eq ==> iff)
    actsfor.
Proof.
  intros H0 H HH.
  intros c0 c Hc.
  intros pc0 pc Hpc.
  intros l0 l Hl.
  intros p0 p Hp.
  intros q0 q Hq.
  intros D0 D HD.
  intros n0 n Hn.
  rewrite Hc, Hn.
  split ; intro.
    eapply actsforProperHelper ; eauto.
    eapply actsforProperHelper
        with (H0 := H) (pc0 := pc) (l0 := l) (p0 := p) (q0 := q) (D0 := D);
      autoalg;
      symmetry;
      auto.
Qed.

Instance actsForRobustProper:
  Proper
    (PH.phEq ==> eq ==> prinEq ==> prinEq ==> prinEq ==> prinEq
      ==> sameSet ==> eq ==> iff)
    actsforRobust.
Proof.
  intros H0 H HH.
  intros c0 c Hc.
  intros pc0 pc Hpc.
  intros l0 l Hl.
  intros p0 p Hp.
  intros q0 q Hq.
  intros D0 D HD.
  intros n0 n Hn.
  rewrite Hc, Hn.
  split ; intro.
    eapply actsforProperHelper ; eauto.
    eapply actsforProperHelper
        with (H0 := H) (pc0 := pc) (l0 := l) (p0 := p) (q0 := q) (D0 := D);
      autoalg;
      symmetry;
      auto.
Qed.

Lemma conjIsBot:
  forall p q, p ∧ q == ⊥ -> p == ⊥ /\ q == ⊥.
Proof.
  assert (forall p q, p ∧ q == ⊥ -> p == ⊥).
    intros.
    eapply actsForEq with (c := Name "c") ; eautoflam.
    setoid_rewrite <- H.
    eautoflam.

  intros.
  split ; eauto.
  rewrite conjComm in H0 ; eauto.
Qed.

Lemma disjIsTop:
  forall p q, p ∨ q == ⊤ -> p == ⊤ /\ q == ⊤.
Proof.
  assert (forall p q, p ∨ q == ⊤ -> p == ⊤).
    intros.
    eapply actsForEq with (c := Name "c") ; eautoflam.
    setoid_rewrite <- H.
    eautoflam.

  intros.
  split ; eauto.
  rewrite disjComm in H0 ; eauto.
Qed.

(*
 * The above definition omits AF_Proj_owns. We now show that it is admissible
 * under the existing rules.
 *)
Lemma AF_Proj_owns:
  forall H c pc l p q r D n n1,
    [H;c;pc;l ⊢ p ≽ q # D;n1]
      -> n = n1+1
      -> [H;c;pc;l ⊢ p:r ≽ q:r # D;n].
Proof.
  intros.
  eapply AF_Own1 ; eautoflamz.
  rewrite unionComm.
  symmetry.
  repeat rewrite emptyZero.
  autoflam.
Qed.

Hint Resolve AF_Proj_owns : flam.

Lemma AF_ConjL2:
  forall H c pc l p1 p2 p D n n1,
    [H;c;pc;l ⊢ p2 ≽ p # D;n1]
      -> n = n1 + 1
      -> [H;c;pc;l ⊢ p1 ∧ p2 ≽ p # D;n].
Proof.
  intros.
  rewrite conjComm.
  eautoflamz.
Qed.

Hint Resolve AF_ConjL2 : flam.

Lemma AF_DisjR2:
  forall H c pc l p1 p2 p D n n1,
    [H;c;pc;l ⊢ p ≽ p2 # D;n1]
      -> n = n1 + 1
      -> [H;c;pc;l ⊢ p ≽ p1 ∨ p2 # D;n].
Proof.
  intros.
  rewrite disjComm.
  eautoflamz.
Qed.

Hint Resolve AF_DisjR2 : flam.

(* The DelUseSet output by a derivation is a subset of H. *)
Lemma useSetInH:
  forall n H c pc l p q D d,
    [H;c;pc;l ⊢ p ≽ q # D;n] \/ [H;c;pc;l ⊩ p ≽ q # D;n]
      -> DelIn d D
      -> PH.In d H.
Proof.
  induction n as [|n IHn] using strong_induction ; intros.
  {
    (* base case *)
    destruct H0.
    {
      (* non-robust judgements *)
      inversion H0;
        try (omega);
        try (
          contradict H1;
          setoid_replace D with empty by auto;
          apply emptyNotDelIn).
      replace d with d0 ; auto.
      rewrite H8 in H1.
      do 2 destruct H1.
      unfold init in H1.
      pose proof (singletonInv _ _ H1).
      subst. auto.
    }
    { inversion H0 ; omega. }
  }
  {
    (* inductive case *)
    destruct H0.
    {
      (* non-robust judgements *)
      inversion H0;
        try (eapply IHn with (j := n1) ; eauto with zarith ; fail);
        try (
          setoid_replace D with (union D1 D2) in H1 by auto;
          destruct (unionInvDel _ _ _ H1);
          [eapply IHn with (j := n1)|eapply IHn with (j := n2)];
          eauto with zarith
        ).
      {
        setoid_replace D with (union D1 (transfer D2 rn)) in H1 by auto.
        destruct (unionInvDel _ _ _ H1).
        { eapply IHn with (j := n1) ; eauto with zarith. }
        {
          eapply IHn with (j := n2) ; eauto with zarith.
          eapply delInTransferInv. eauto.
        }
      }
      {
        setoid_replace D with (union D1 (union D2 D3)) in H1 by auto.
        rewrite H8.
        apply PH.unionIntro. left.
        destruct (unionInvDel3 _ _ _ _ H1) as [?|[?|?]].
        { eapply IHn with (j := n1) ; eauto with zarith. }
        { eapply IHn with (j := n2) ; eauto with zarith. }
        { eapply IHn with (j := n3) ; eauto with zarith. }
      }
    }
    {
      (* robust judgements *)
      inversion H0;
        try (
          setoid_replace D with (union D1 (union D2 D3)) in H1 by auto;
          destruct (unionInvDel3 _ _ _ _ H1) as [?|[?|?]];
          [eapply IHn with (j := n1) | eapply IHn with (j := n2)
            | eapply IHn with (j := n3)];
          eauto with zarith;
          fail
        );
        try (
          setoid_replace D with (union D1 D2) in H1 by auto;
          destruct (unionInvDel _ _ _ H1);
          [eapply IHn with (j := n1)|eapply IHn with (j := n2)];
          eauto with zarith
        ).
      {
        setoid_replace D with empty in H1 by auto.
        contradict H1. apply emptyNotDelIn.
      }
      { eapply IHn with (j := n1) ; eauto with zarith. }
      {
        setoid_replace D with (union D1 (union D2 D3)) in H1 by auto.
        rewrite H8.
        apply PH.unionIntro. left.
        destruct (unionInvDel3 _ _ _ _ H1) as [?|[?|?]].
        { eapply IHn with (j := n1) ; eauto with zarith. }
        { eapply IHn with (j := n2) ; eauto with zarith. }
        { eapply IHn with (j := n3) ; eauto with zarith. }
      }
      {
        setoid_replace D with (union D1 (transfer D2 rn)) in H1 by auto.
        destruct (unionInvDel _ _ _ H1).
        { eapply IHn with (j := n1) ; eauto with zarith. }
        {
          eapply IHn with (j := n2) ; eauto with zarith.
          eapply delInTransferInv. eauto.
        }
      }
    }
  }
Qed.

(* Derivations in empty trust configurations don't use delegations. *)
Lemma emptyHEmptyD:
  forall c pc l p q D n,
    [PH.empty;c;pc;l ⊢ p ≽ q # D;n] -> D =D= empty.
Proof.
  intros.
  apply emptySameAsEmpty.
  intro. intro.
  eapply PH.emptyIsEmpty.
  eapply useSetInH ; eauto.
  unfold DelIn. eauto.
Qed.

Hint Resolve emptyHEmptyD : flam.

(**
 *** Static judgements
 *)

Lemma botActsFor:
  forall c p n, [c ⊢ ⊥ ≽ p # n] -> p == ⊥.
Proof.
  intros.
  eapply actsForEq.
    eautoflam.
    eauto.
Qed.

Lemma actsForTop:
  forall c p n, [c ⊢ p ≽ ⊤ # n] -> p == ⊤.
Proof.
  intros.
  eapply actsForEq.
    eauto.
    eautoflam.
Qed.

Lemma flowsFromTopConf:
  forall c p n, [c ⊢ ⊤→ ⊑ p # n] -> p == ⊤→.
Proof.
  intros.
  rewrite <- (confIntegBasis p).
  eapply actsForEq.
  {
    eexists.
    eapply AF_ConjL1.
    { eapply refl. }
    {
      eapply AF_Proj_conf with (p' := p→ ∧ ⊤→←) (q' := ⊤→→ ∧ p←).
      { autoalg. }
      { rewrite conjConf. autoalg. }
      { eauto. }
      { auto. }
    }
    { auto. }
  }
  {
    eexists.
    eapply AF_ConjR.
    { eapply refl. }
    { eautoflam. }
    {
      eapply AF_ConjL1 with (p1 := ⊤→←) (p2 := ⊤→).
      { algMagic. }
      {
        eapply AF_Proj_integ with (p' := p→ ∧ ⊤→←) (q' := ⊤→→ ∧ p←);
          try (algMagic ; fail) ; eauto.
      }
      { autoflam. }
    }
    { rewrite unionIdemp. auto. }
    { auto. }
  }
Qed.

Lemma flowsToTopInteg:
  forall c p n, [c ⊢ p ⊑ ⊤← # n] -> p == ⊤←.
Proof.
  intros.
  rewrite <- (confIntegBasis p).
  eapply actsForEq.
  {
    eexists.
    eapply AF_ConjL2.
    {
      eapply AF_Proj_integ with (p' := ⊤←→ ∧ p←) (q' := p→ ∧ ⊤←←);
        try (algMagic ; fail);
        eauto.
    }
    { auto. }
  }
  {
    eexists.
    eapply AF_ConjR.
    { eapply refl. }
    {
      eapply AF_ConjL1 with (p1 := ⊥) (p2 := ⊤←).
      { algMagic. }
      {
        eapply AF_Proj_conf with (p' := ⊤←→ ∧ p←) (q' := p→ ∧ ⊤←←);
          try (algMagic ; fail);
          eauto.
      }
      { auto. }
    }
    { eautoflam. }
    { rewrite unionIdemp. auto. }
    { auto. }
  }
Qed.

Lemma staticActsForConf:
  forall c p q n, [c ⊢ p→ ≽ q # n] -> q == q→.
Proof.
  intros.
  eapply actsForEq.
  { eautoflam. }
  {
    eexists.
    eapply AF_ConjR with (q1 := q→) (q2 := q←).
    { algMagic. }
    { autoflam. }
    {
      eapply AF_ConjL1 with (p1 := ⊥) (p2 := q→).
      { algMagic. }
      {
        eapply AF_Proj_integ with (p' := p→) (q' := q);
          try (algMagic ; fail);
          eauto.
      }
      { auto. }
    }
    { rewrite unionIdemp. auto. }
    { auto. }
  }
Qed.

(* In static contexts, robust and non-robust judgements are equivalent. *)
Lemma staticRobust:
  forall n c p q,
    [c ⊩ p ≽ q # n] -> exists n', [c ⊢ p ≽ q # n'] /\ n' < n.
Proof.
  induction n as [|n IHn] using strong_induction ; intros.
  { inversion H ; omega. }
  {
    inversion H.
    { eauto with zarith. }
    {
      pose proof emptyUnion3 _ _ _ H4 as HD.
      destruct HD as [HD1 [HD2 HD3]].
      rewrite HD1 in *.
      assert (⊤← ∧ ∇(q) == ⊤←) as Hrewrite.
        destruct (voiceIsInteg q) as [x Hvoice].
        rewrite Hvoice , <- conjInteg.
        autoalg.
      setoid_rewrite Hrewrite in H1.
      eauto with zarith.
    }
    {
      assert (⊤← ∧ ∇(q') == ⊤←) as Hrewrite.
        destruct (voiceIsInteg q') as [x Hvoice].
        rewrite Hvoice , <- conjInteg.
        autoalg.
      setoid_rewrite Hrewrite in H3.
      setoid_rewrite H1.
      setoid_rewrite H2.
      eauto with zarith.
    }
    {
      pose proof emptyUnion2 _ _ H4 as HD.
      destruct HD as [HD1 HD2].
      rewrite HD1, HD2 in *.

      assert (exists n', [c ⊢ p ≽ q1 # n'] /\ n' < n1).
        eapply IHn with (j := n1) ; eauto with zarith.
      assert (exists n', [c ⊢ p ≽ q2 # n'] /\ n' < n2).
        eapply IHn with (j := n2) ; eauto with zarith.
      do 2 destruct H14, H15.

      eautoflamz.
    }
    {
      pose proof emptyUnion2 _ _ H4 as HD.
      destruct HD as [HD1 HD2].
      rewrite HD1, HD2 in *.

      assert (exists n', [c ⊢ p1 ≽ q # n'] /\ n' < n1).
        eapply IHn with (j := n1) ; eauto with zarith.
      assert (exists n', [c ⊢ p2 ≽ q # n'] /\ n' < n2).
        eapply IHn with (j := n2) ; eauto with zarith.
      do 2 destruct H14, H15.

      eautoflamz.
    }
    {
      destruct (PH.emptyUnion2 _ _ H6) as [HH1 HH2].
      destruct (emptyUnion3 _ _ _ H7) as [HD1 [HD2 HD3]].
      rewrite HH1, HH2, HD1, HD2, HD3 in *.
      assert (⊤→ ⊔ l' == ⊤→) as Hrewrite1 by algMagic.
      rewrite Hrewrite1 in H4, H5.

      assert (exists n', [c ⊢ ⊤→ ⊑ pc' # n'] /\ n' < n2).
        eapply IHn with (j := n2) ; eauto with zarith.
      assert (exists n', [c ⊢ l' ⊑ ⊤← # n'] /\ n' < n3).
        eapply IHn with (j := n3) ; eauto with zarith.
      do 2 destruct H17, H18.

      assert (pc' == ⊤→) as Hpc by (eapply flowsFromTopConf ; eauto).
      assert (l' == ⊤←) as Hl by (eapply flowsToTopInteg ; eauto).
      rewrite Hpc, Hl in *.

      assert (exists n', [c ⊢ p ≽ q # n'] /\ n' < n1).
        eapply IHn with (j := n1) ; eauto with zarith.
      do 2 destruct H21.

      eautoflamz.
    }
    {
      pose proof emptyUnion3 _ _ _ H4 as HD.
      destruct HD as [HD1 [HD2 HD3]].
      rewrite HD1, HD2, HD3 in *.

      assert (exists n', [c ⊢ p ≽ q0 # n'] /\ n' < n1).
        eapply IHn with (j := n1) ; eauto with zarith.
      assert (exists n', [c ⊢ q0 ≽ q # n'] /\ n' < n2).
        eapply IHn with (j := n2) ; eauto with zarith.
      do 2 destruct H14, H15.

      eexists.
      split.
        eapply AF_Trans with (q := q0) ; eautoflam.
        omega.
    }
    {
      pose proof emptyUnion2 _ _ H5 as HD.
      destruct HD as [HD1 HtransferD2].
      pose proof transferIsEmpty _ _ HtransferD2 as HD2.
      rewrite HD1, HD2 in *.

      assert (⊤→ ⊔ ⊤← ⊔ c0← == ⊤→) as Hrewrite1 by autoalg.
      assert (⊤← ⊓ c0→ == ⊤←) as Hrewrite2 by autoalg.
      setoid_replace ((⊤→→ ∧ ⊤←) ∧ ∇(q)) with (⊤→→ ∧ ⊤←) in H3 by autoalg.
      assert (exists n', [rn ⊢ p ≽ q # n'] /\ n' < n2).
        rewrite Hrewrite1, Hrewrite2 in H4.
        eapply IHn with (j := n2) ; eauto with zarith.
      do 2 destruct H15.

      eexists.
      split.
        eapply AF_Fwd with (rn := rn) (n2 := x) ; eauto.
          rewrite Hrewrite1, Hrewrite2. eauto.
        omega.
    }
  }
Qed.

Lemma staticActsForInteg:
  forall c p q n, [c ⊢ p← ≽ q # n] -> q == q←.
Proof.
  intros.
  eapply actsForEq.
  { eautoflam. }
  {
    eexists.
    eapply AF_ConjR with (q1 := q→) (q2 := q←).
    { algMagic. }
    {
      eapply AF_ConjL1 with (p1 := ⊥) (p2 := q←).
      { algMagic. }
      {
        eapply AF_Proj_conf with (p' := p←) (q' := q);
          try (algMagic ; fail) ; eauto.
      }
      { auto. }
    }
    { eautoflam. }
    { rewrite unionIdemp. auto. }
    { auto. }
  }
Qed.

(*
 * Judgements in empty trust hierarchies hold regardless of where they are
 * made.
 *)
Lemma staticCoerce:
  forall c p q n c',
       ([c ⊢ p ≽ q # n] -> (exists n', [c' ⊢ p ≽ q # n'] /\ n' <= n))
    /\ ([c ⊩ p ≽ q # n] -> (exists n', [c' ⊩ p ≽ q # n'] /\ n' <= n)).
Proof.
  assert (
      forall n c p q,
           ([c ⊢ p ≽ q # n]
               -> forall c', exists n', [c' ⊢ p ≽ q # n'] /\ n' <= n)
        /\ ([c ⊩ p ≽ q # n]
               -> forall c', exists n', [c' ⊩ p ≽ q # n'] /\ n' <= n)).
  {
    intro.
    induction n as [|n IHn] using strong_induction.
    {
      intros ; split ; intros.
      {
        inversion H ; try (omega) ; eautoflam.
        contradict H6. intro.
        apply (initNotEmpty d). apply emptySameAsEmpty.
        symmetry. auto.
      }
      { inversion H ; omega. }
    }
    {
      intros ; split ; intros.
      {
        inversion H.
        {
          assert (exists n', [c' ⊢ p' ≽ q' # n'] /\ n' <= n1).
            eapply IHn with (j := n1) ; eauto with zarith.
          do 2 destruct H13.

          eautoflamz.
        }
        {
          assert (exists n', [c' ⊢ p' ≽ q' # n'] /\ n' <= n1).
            eapply IHn with (j := n1) ; eauto with zarith.
          do 2 destruct H13.

          eautoflamz.
        }
        {
          destruct (emptyUnion2 _ _ H5) as [HD1 HD2].
          rewrite HD1, HD2 in *.

          assert (exists n', [c' ⊢ o ≽ o' # n'] /\ n' <= n1).
            eapply IHn with (j := n1) ; eauto with zarith.
          assert (exists n', [c' ⊢ r ≽ r' # n'] /\ n' <= n2).
            eapply IHn with (j := n2) ; eauto with zarith.
          do 2 destruct H15, H16.

          eautoflamz.
        }
        {
          destruct (emptyUnion2 _ _ H5) as [HD1 HD2].
          rewrite HD1, HD2 in *.

          assert (exists n', [c' ⊢ o ≽ o' # n'] /\ n' <= n1).
            eapply IHn with (j := n1) ; eauto with zarith.
          assert (exists n', [c' ⊢ r ≽ o':r' # n'] /\ n' <= n2).
            eapply IHn with (j := n2) ; eauto with zarith.
          do 2 destruct H15, H16.

          eautoflamz.
        }
        {
          assert (exists n', [c' ⊢ p1 ≽ q # n'] /\ n' <= n1).
            eapply IHn with (j := n1) ; eauto with zarith.
          do 2 destruct H12.

          eautoflamz.
        }
        {
          destruct (emptyUnion2 _ _ H4) as [HD1 HD2].
          rewrite HD1, HD2 in *.

          assert (exists n', [c' ⊢ p ≽ q1 # n'] /\ n' <= n1).
            eapply IHn with (j := n1) ; eauto with zarith.
          assert (exists n', [c' ⊢ p ≽ q2 # n'] /\ n' <= n2).
            eapply IHn with (j := n2) ; eauto with zarith.
          do 2 destruct H14, H15.

          eautoflamz.
        }
        {
          destruct (emptyUnion2 _ _ H4) as [HD1 HD2].
          rewrite HD1, HD2 in *.

          assert (exists n', [c' ⊢ p1 ≽ q # n'] /\ n' <= n1).
            eapply IHn with (j := n1) ; eauto with zarith.
          assert (exists n', [c' ⊢ p2 ≽ q # n'] /\ n' <= n2).
            eapply IHn with (j := n2) ; eauto with zarith.
          do 2 destruct H14, H15.

          eautoflamz.
        }
        {
          assert (exists n', [c' ⊢ p ≽ q1 # n'] /\ n' <= n1).
            eapply IHn with (j := n1) ; eauto with zarith.
          do 2 destruct H12.

          eautoflamz.
        }
        {
          destruct (emptyUnion2 _ _ H3).
          rewrite H13, H14 in *.

          assert (exists n', [c' ⊢ p ≽ q0 # n'] /\ n' <= n1).
            eapply IHn with (j := n1) ; eauto with zarith.
          assert (exists n', [c' ⊢ q0 ≽ q # n'] /\ n' <= n2).
            eapply IHn with (j := n2) ; eauto with zarith.
          do 2 destruct H15, H16.

          eexists.
          split.
            eapply AF_Trans ; eauto with zarith.
            omega.
        }
        {
          destruct (emptyUnion2 _ _ H5) as [HD1 HtransferD2].
          pose proof transferIsEmpty _ _ HtransferD2 as HD2.
          rewrite HD2, joinTopConf, meetTopInteg in *.

          assert (exists n, [c' ⊢ p ≽ q # n] /\ n <= n2).
            eapply IHn with (j := n2) ; eauto with zarith.
          do 2 destruct H15.

          eauto with zarith.
        }
        {
          destruct (PH.emptyUnion2 _ _ H6) as [HH1 HH2].
          destruct (emptyUnion3 _ _ _ H7) as [HD1 [HD2 HD3]].
          rewrite HH1, HH2, HD1, HD2, HD3, joinTopConf in *.

          assert (pc' == ⊤→).
            pose proof staticRobust n2 c (pc'→ ∧ ⊤→←) (⊤→→ ∧ pc'←) H4.
            do 2 destruct H17.
            eapply flowsFromTopConf ; eauto with zarith.

          assert (l' == ⊤←).
            pose proof staticRobust n3 c (⊤←→ ∧ l'←) (l'→ ∧ ⊤←←) H5.
            do 2 destruct H18.
            eapply flowsToTopInteg ; eauto with zarith.
          rewrite H17, H18 in *.

          assert (exists n, [c' ⊢ p ≽ q # n] /\ n <= n1).
            eapply IHn with (j := n1) ; eauto with zarith.
          do 2 destruct H19.

          eauto with zarith.
        }
      }
      {
        inversion H.
        {
          assert (exists n, [c' ⊢ p ≽ q # n] /\ n <= n1).
            eapply IHn with (j := n1) ; eauto with zarith.
          do 2 destruct H12.

          eautoflamz.
        }
        {
          destruct (emptyUnion3 _ _ _ H4) as [HD1 [HD2 HD3]].
          assert (⊤← ∧ ∇(q) == ⊤←) as Hrewrite1.
            destruct (voiceIsInteg q) as [x Hvoice].
            rewrite Hvoice, <- conjInteg.
            autoalg.
          rewrite HD1, HD2, HD3, Hrewrite1 in *.

          assert (exists n, [c' ⊢ p ≽ q # n] /\ n <= n1).
            eapply IHn with (j := n1) ; eauto with zarith.
          do 2 destruct H14.

          eautoflamz.
        }
        {
          assert (exists n, [PH.empty;c';⊤→;⊤← ∧ ∇(q') ⊢ ⊤→ ≽ ∇(q') # empty;n]
              /\ n <= n1).
            assert (⊤← ∧ ∇(q') == ⊤←) as Hrewrite2.
              destruct (voiceIsInteg q') as [x Hvoice].
              rewrite Hvoice, <- conjInteg.
              autoalg.
            setoid_rewrite Hrewrite2.
            rewrite Hrewrite2 in *.
            eapply IHn with (j := n1) ; eauto with zarith.
          do 2 destruct H13.

          eautoflamz.
        }
        {
          destruct (emptyUnion2 _ _ H4) as [HD1 HD2].
          rewrite HD1, HD2 in *.

          assert (exists n', [c' ⊩ p ≽ q1 # n'] /\ n' <= n1).
            eapply IHn with (j := n1) ; eauto with zarith.
          assert (exists n', [c' ⊩ p ≽ q2 # n'] /\ n' <= n2).
            eapply IHn with (j := n2) ; eauto with zarith.
          do 2 destruct H14, H15.

          eautoflamz.
        }
        {
          destruct (emptyUnion2 _ _ H4) as [HD1 HD2].
          rewrite HD1, HD2 in *.

          assert (exists n', [c' ⊩ p1 ≽ q # n'] /\ n' <= n1).
            eapply IHn with (j := n1) ; eauto with zarith.
          assert (exists n', [c' ⊩ p2 ≽ q # n'] /\ n' <= n2).
            eapply IHn with (j := n2) ; eauto with zarith.
          do 2 destruct H14, H15.

          eautoflamz.
        }
        {
          destruct (PH.emptyUnion2 _ _ H6) as [HH1 HH2].
          destruct (emptyUnion3 _ _ _ H7) as [HD1 [HD2 HD3]].
          assert (⊤→ ⊔ l' == ⊤→) as Hrewrite1 by algMagic.
          rewrite HH1, HH2, HD1, HD2, HD3, Hrewrite1 in *.

          assert (exists n, [c ⊢ ⊤→ ⊑ pc' # n] /\ n < n2)
              by (eapply staticRobust ; eauto).
          assert (exists n, [c ⊢ l' ⊑ ⊤← # n] /\ n < n3)
              by (eapply staticRobust ; eauto).
          do 2 destruct H17, H18.

          assert (pc' == ⊤→) as Hpc by (eapply flowsFromTopConf ; eauto).
          assert (l' == ⊤←) as Hl by (eapply flowsToTopInteg ; eauto).
          rewrite Hpc, Hl in *.

          assert (exists n, [c' ⊩ p ≽ q # n] /\ n <= n1).
            eapply IHn with (j := n1) ; eauto with zarith.
          do 2 destruct H21.

          eauto with zarith.
        }
        {
          destruct (emptyUnion3 _ _ _ H4) as [HD1 [HD2 HD3]].
          rewrite HD1, HD2, HD3 in *.

          assert (exists n, [c' ⊩ p ≽ q0 # n] /\ n <= n1).
            eapply IHn with (j := n1) ; eauto with zarith.
          assert (exists n, [c' ⊩ q0 ≽ q # n] /\ n <= n2).
            eapply IHn with (j := n2) ; eauto with zarith.
          assert (exists n, [c' ⊩ ⊤→ ≽ ∇(q→) # n] /\ n <= n3).
            eapply IHn with (j := n3) ; eauto with zarith.
          do 2 destruct H14, H15, H16.

          eexists.
          split.
            eapply RAF_Trans ; eauto.
            omega.
        }
        {
          destruct (emptyUnion2 _ _ H5) as [HD1 HtransferD2].
          pose proof transferIsEmpty _ _ HtransferD2 as HD2.
          rewrite HD1, HD2 in *.
          rewrite joinTopConf, meetTopInteg in H4.

          assert (exists n, [c' ⊩ p ≽ q # n] /\ n <= n2).
            eapply IHn with (j := n2) ; eauto with zarith.
          do 2 destruct H15.

          eauto with zarith.
        }
      }
    }
  }

  intros.
  pose proof H n c p q.
  destruct H0.
  auto.
Qed.

(* Static judgements may be derived by any host. *)
Lemma staticSubst:
  forall rn cn p q, [rn ⊢ p ≽ q] -> [cn ⊢ p ≽ q].
Proof.
  intros.
  destruct H.
  assert (exists n' : nat, [cn ⊢ p ≽ q # n'] /\ n' <= x).
    eapply staticCoerce.
    eapply H.
  do 2 destruct H0.

  eautoflam.
Qed.

Lemma staticFlowCoerce:
  forall c p q n c',
    [c ⊢ p ⊑ q # n] -> exists n', [c' ⊢ p ⊑ q # n'] /\ n' <= n.
Proof.
  intros.
  eapply staticCoerce ; eauto.
Qed.

(* Judgements with empty delegation sets can be derived statically. *)
Lemma emptyDIsStatic:
  forall n H cn pc l p q D,
      D =D= empty
      -> ([H;cn;pc;l ⊢ p ≽ q # D;n] \/ [H;cn;pc;l ⊩ p ≽ q # D;n]
          -> [cn ⊢ p ≽ q]).
(* begin hide *)
Proof.
  induction n as [|n IHn] using strong_induction ; intros.
  {
    rewrite H0 in H1.
    destruct H1.
    {
      inversion H1 ; try (omega) ; try (eautoflam ; fail).
      {
        contradict H8. intro.
        apply (initNotEmpty d). apply emptySameAsEmpty.
        symmetry. auto.
      }
    }
    { inversion H1 ; try (omega). }
  }
  {
    destruct H1.
    {
      (* non-robust inductive case *)
      inversion H1 ; rewrite H0 in *.
      {
        (* AF_Proj_conf *)
        assert ([cn ⊢ p' ≽ q']).
        { eapply IHn with (j := n1) (D := empty) ; eautoflamz. }
        destruct H15.
        eautoflam.
      }
      {
        (* AF_Proj_integ *)
        assert ([cn ⊢ p' ≽ q']).
        { eapply IHn with (j := n1) (D := empty) ; eautoflamz. }
        destruct H15.
        eautoflam.
      }
      {
        (* AF_Own1 *)
        destruct (emptyUnion2 _ _ H7) as [HD1 HD2].
        rewrite HD1, HD2 in *.

        assert ([cn ⊢ o ≽ o']) as IH1.
        { eapply IHn with (j := n1) (D := empty) ; eautoflamz. }

        assert ([cn ⊢ r ≽ r']) as IH2.
        { eapply IHn with (j := n2) (D := empty) ; eautoflamz. }
        destruct IH1, IH2.

        eautoflam.
      }
      {
        (* AF_Own2 *)
        destruct (emptyUnion2 _ _ H7) as [HD1 HD2].
        rewrite HD1, HD2 in *.

        assert ([cn ⊢ o ≽ o']) as IH1.
        { eapply IHn with (j := n1) (D := empty) ; eautoflamz. }

        assert ([cn ⊢ r ≽ o':r']) as IH2.
        { eapply IHn with (j := n2) (D := empty) ; eautoflamz. }
        destruct IH1, IH2.

        eautoflam.
      }
      {
        (* AF_ConjL1 *)
        assert ([cn ⊢ p1 ≽ q]) as IH1.
        { eapply IHn with (j := n1) (D := empty) ; eautoflamz. }
        destruct IH1.

        eautoflam.
      }
      {
        (* AF_ConjR *)
        destruct (emptyUnion2 _ _ H6) as [HD1 HD2].
        rewrite HD1, HD2 in *.

        assert ([cn ⊢ p ≽ q1]) as IH1.
        { eapply IHn with (j := n1) (D := empty) ; eautoflamz. }

        assert ([cn ⊢ p ≽ q2]) as IH2.
        { eapply IHn with (j := n2) (D := empty) ; eautoflamz. }
        destruct IH1, IH2.

        eautoflam.
      }
      {
        (* AF_DisjL *)
        destruct (emptyUnion2 _ _ H6) as [HD1 HD2].
        rewrite HD1, HD2 in *.

        assert ([cn ⊢ p1 ≽ q]) as IH1.
        { eapply IHn with (j := n1) (D := empty) ; eautoflamz. }

        assert ([cn ⊢ p2 ≽ q]) as IH2.
        { eapply IHn with (j := n2) (D := empty) ; eautoflamz. }
        destruct IH1, IH2.

        eautoflam.
      }
      {
        (* AF_DisjR1 *)
        assert ([cn ⊢ p ≽ q1]) as IH1.
        { eapply IHn with (j := n1) (D := empty) ; eautoflamz. }
        destruct IH1.

        eautoflam.
      }
      {
        (* AF_Trans *)
        destruct (emptyUnion2 _ _ H5) as [HD1 HD2].
        rewrite HD1, HD2 in *.

        assert ([cn ⊢ p ≽ q0]) as IH1.
        { eapply IHn with (j := n1) (D := empty) ; eautoflamz. }

        assert ([cn ⊢ q0 ≽ q]) as IH2.
        { eapply IHn with (j := n2) (D := empty) ; eautoflamz. }
        destruct IH1, IH2.

        eexists.
        eapply AF_Trans with (q := q0) ; eauto.
      }
      {
        (* AF_Fwd *)
        destruct (emptyUnion2 _ _ H7) as [HD1 HtransferD2].
        pose proof (transferIsEmpty _ _ HtransferD2) as HD2.
        rewrite HD1, HD2 in *.

        eapply staticSubst with (rn := rn).
        eapply IHn with (j := n2) (D := empty) ; eautoflamz.
      }
      {
        (* AF_Weaken *)
        destruct (emptyUnion3 _ _ _ H9) as [HD1 [HD2 HD3]].
        rewrite HD1, HD2, HD3 in *.

        eapply IHn with (j := n1) (D := empty) ; eautoflamz.
      }
    }
    {
      (* robust inductive case *)
      inversion H1 ; rewrite H0 in *.
      {
        (* RAF_Static *)
        eautoflam.
      }
      {
        (* RAF_Lift *)
        destruct (emptyUnion3 _ _ _ H6) as [HD1 [HD2 HD3]].
        rewrite HD1, HD2, HD3 in *.

        eapply IHn with (j := n1) (D := empty) ; eautoflamz.
      }
      {
        (* RAF_LiftPC *)
        eapply IHn with (j := n1) (D := empty) ; eautoflamz.
        left.
        rewrite H3, H4, H13.
        eauto.
      }
      {
        (* RAF_ConjR *)
        destruct (emptyUnion2 _ _ H6) as [HD1 HD2].
        rewrite HD1, HD2 in *.

        assert ([cn ⊢ p ≽ q1]) as IH1.
        { eapply IHn with (j := n1) (D := empty) ; eautoflamz. }

        assert ([cn ⊢ p ≽ q2]) as IH2.
        { eapply IHn with (j := n2) (D := empty) ; eautoflamz. }
        destruct IH1, IH2.

        eautoflam.
      }
      {
        (* RAF_DisjL *)
        destruct (emptyUnion2 _ _ H6) as [HD1 HD2].
        rewrite HD1, HD2 in *.

        assert ([cn ⊢ p1 ≽ q]) as IH1.
        { eapply IHn with (j := n1) (D := empty) ; eautoflamz. }

        assert ([cn ⊢ p2 ≽ q]) as IH2.
        { eapply IHn with (j := n2) (D := empty) ; eautoflamz. }
        destruct IH1, IH2.

        eautoflam.
      }
      {
        (* RAF_Weaken *)
        destruct (emptyUnion3 _ _ _ H9) as [HD1 [HD2 HD3]].
        rewrite HD1, HD2, HD3 in *.

        eapply IHn with (j := n1) (D := empty) ; eautoflamz.
      }
      {
       (* RAF_Trans *)
        destruct (emptyUnion3 _ _ _ H6) as [HD1 [HD2 HD3]].
        rewrite HD1, HD2, HD3 in *.

        assert ([cn ⊢ p ≽ q0]) as IH1.
        { eapply IHn with (j := n1) (D := empty) ; eautoflamz. }

        assert ([cn ⊢ q0 ≽ q]) as IH2.
        { eapply IHn with (j := n2) (D := empty) ; eautoflamz. }
        destruct IH1, IH2.

        eexists.
        eapply AF_Trans with (q := q0) ; eautoflam.
      }
      {
        (* AF_Fwd *)
        destruct (emptyUnion2 _ _ H7) as [HD1 HtransferD2].
        pose proof (transferIsEmpty _ _ HtransferD2) as HD2.
        rewrite HD1, HD2 in *.

        eapply staticSubst with (rn := rn).
        eapply IHn with (j := n2) (D := empty) ; eautoflamz.
      }
    }
  }
Qed.
(* end hide *)

Lemma staticConjActsFor:
  forall c p1 p2 q,
    [c ⊢ p1 ∧ p2 ≽ q]
      -> exists q1 q2, q == q1 ∧ q2 /\ [c ⊢ p1 ≽ q1] /\ [c ⊢ p2 ≽ q2].
Proof.
  intros.
  exists (p1 ∨ q), (p2 ∨ q).
  split ; [|split ; eautoflam].
  eapply actsForEq with (c := c).
  { eexists. eapply AF_ConjR ; eautoflam. }
  {
    destruct H.
    setoid_rewrite <- disjRDistConj.
    eautoflam.
  }
Qed.

Lemma staticActsForDisj:
  forall c p q1 q2,
    [c ⊢ p ≽ q1 ∨ q2]
      -> exists p1 p2, p == p1 ∨ p2 /\ [c ⊢ p1 ≽ q1] /\ [c ⊢ p2 ≽ q2].
Proof.
  intros.
  exists (p ∧ q1), (p ∧ q2).
  split ; [|split ; eautoflam].
  eapply actsForEq with (c := c).
  {
    destruct H.
    setoid_rewrite <- conjDistDisj.
    eautoflam.
  }
  { eexists. eapply AF_DisjL ; eautoflam. }
Qed.

(**
 *** Flows-to relationships
 *)

(* ⊤→ is the top element in the information-flow lattice. *)
Lemma flowToTopConf:
  forall H c pc l p,
    [H;c;pc;l ⊢ p ⊑ ⊤→ # empty;2].
Proof.
  intros.
  autoalg.
  eautoflam.
Qed.

Hint Resolve flowToTopConf : flam.

(* ⊤← is the bottom element in the information-flow lattice. *)
Lemma flowFromTopInteg:
  forall H c pc l p,
    [H;c;pc;l ⊢ ⊤← ⊑ p # empty;2].
Proof.
  intros.
  autoalg.
  eautoflam.
Qed.

Hint Resolve flowFromTopInteg : flam.

Lemma flowToJoin:
  forall H c pc l p q,
    [H;c;pc;l ⊢ p ⊑ p ⊔ q # empty;5].
Proof.
  intros.
  algMagic.
  eapply AF_ConjR ; eautoflam.
Qed.

Hint Resolve flowToJoin : flam.

Lemma flowToJoin':
  forall H c pc l p q,
    [H;c;pc;l ⊢ q ⊑ p ⊔ q # empty;5].
Proof.
  intros.
  setoid_replace (p ⊔ q) with (q ⊔ p) by algMagic.
  apply flowToJoin.
Qed.

Hint Resolve flowToJoin' : flam.

Lemma flowFromMeet:
  forall H c pc l p q,
    [H;c;pc;l ⊢ p ⊓ q ⊑ p # empty;4].
Proof.
  intros.
  algMagic.
  eapply AF_ConjR ; eautoflam.
Qed.

Hint Resolve flowFromMeet : flam.

Lemma flowFromMeet':
  forall H c pc l p q,
    [H;c;pc;l ⊢ p ⊓ q ⊑ q # empty;4].
Proof.
  intros.
  setoid_replace (p ⊓ q) with (q ⊓ p) by algMagic.
  apply flowFromMeet.
Qed.

Hint Resolve flowFromMeet' : flam.

(*
 * Converts a flows-to relationship to an acts-for relationship between
 * confidentiality projections.
 *)
Lemma flow2Conf:
  forall H c pc l p q D n,
    [H;c;pc;l ⊢ p ⊑ q # D;n]
      -> [H;c;pc;l ⊢ q→ ≽ p→ # D;n+1].
Proof.
  intros.
  setoid_replace (p→) with ((p→ ∧ q←)→) by (rewrite conjConf ; autoalg).
  setoid_replace (q→) with ((q→ ∧ p←)→) by (rewrite conjConf ; autoalg).
  eautoflam.
Qed.

Hint Resolve flow2Conf : flam.

(*
 * Converts a flows-to relationship to an acts-for relationship between
 * integrity projections.
 *)
Lemma flow2Integ:
  forall H c pc l p q D n,
    [H;c;pc;l ⊢ p ⊑ q # D;n]
      -> [H;c;pc;l ⊢ p← ≽ q← # D;n+1].
Proof.
  intros.
  setoid_replace (p←) with ((q→ ∧ p←)←) by (rewrite conjInteg ; autoalg).
  setoid_replace (q←) with ((p→ ∧ q←)←) by (rewrite conjInteg ; autoalg).
  eautoflam.
Qed.

Hint Resolve flow2Integ : flam.

(* Flows-to is transitive. *)
Lemma flowTrans:
  forall H c pc l p q r D1 n1 D2 n2 D n,
    [H;c;pc;l ⊢ p ⊑ q # D1;n1]
      -> [H;c;pc;l ⊢ q ⊑ r # D2;n2]
      -> D =D= union D1 D2
      -> n = n1 + n1 + n2 + n2 + 9
      -> [H;c;pc;l ⊢ p ⊑ r # D;n].
Proof.
  intros.
  assert ([H;c;pc;l ⊢ r→ ≽ q→ # D2;n2+1]) by eautoflam.
  assert ([H;c;pc;l ⊢ q← ≽ r← # D2;n2+1]) by eautoflam.
  assert ([H;c;pc;l ⊢ q→ ≽ p→ # D1;n1+1]) by eautoflam.
  assert ([H;c;pc;l ⊢ p← ≽ q← # D1;n1+1]) by eautoflam.
  eapply AF_ConjR ; auto using refl.
  {
    eapply AF_ConjL1 ; eauto using refl.
    eapply AF_Trans ; eautoflam.
  }
  {
    eapply AF_ConjL2 ; eauto using refl.
    eapply AF_Trans ; eautoflam.
  }
  {
    rewrite (unionComm D2 D1).
    rewrite <- H2.
    rewrite unionIdemp ; autoflam.
  }
  { omega. }
Qed.

(**
 *** Monotonicity
 *)

(* Conjunction is monotonic with respect to acts-for. *)
Lemma conjMonotonic:
  forall H c pc l p q r D n,
    [H;c;pc;l ⊢ p ≽ q # D;n]
      -> [H;c;pc;l ⊢ r ∧ p ≽ r ∧ q # D;n+4].
Proof.
  intros.
  eapply AF_Trans with (q := r ∧ p ∧ q).
  {
    rewrite <- conjAssoc.
    eautoflam.
  }
  {
    rewrite <- conjAssoc, conjPermute.
    eautoflam.
  }
  {
    repeat rewrite emptyZero ; autoflam.
    rewrite unionComm. autoflam.
  }
  { omega. }
Qed.

Hint Resolve conjMonotonic : flam.

(**
 *** Voice
 *)
(* Principal voice is monotonic with respect to confidentiality projection. *)
Lemma voiceConfMonotonic:
  forall c p,
    [c ⊢ ∇(p) ≽ ∇(p→)].
Proof.
  intros.
  setoid_rewrite <- (confIntegBasis p).
  setoid_rewrite voiceOfConj.
  algMagic.
  eautoflam.
Qed.

Hint Resolve voiceConfMonotonic : flam.

Lemma voiceActsForInteg:
  forall c p, [c ⊢ ∇(p) ≽ p←].
Proof.
  intros.
  setoid_rewrite <- (confIntegBasis p).
  setoid_rewrite voiceOfConj.
  algMagic.
  eautoflam.
Qed.

Hint Resolve voiceActsForInteg : flam.

Lemma voiceActsForIntegVoice:
  forall c p, [c ⊢ ∇(p) ≽ ∇(p←)].
Proof.
  intros.
  algMagic.
  eapply voiceActsForInteg.
Qed.

Hint Resolve voiceActsForIntegVoice : flam.

(* Principal voice is monotonic with respect to disjunction. *)
Lemma voiceDisjMonotonic:
  forall c p q,
    [c ⊢ ∇(p) ≽ ∇(p ∨ q)].
Proof.
  intros.
  destruct (normForm p) as [pc [pi [pfpc [pfpi Hp]]]].
  destruct (normForm q) as [qc [qi [pfqc [pfqi Hq]]]].
  setoid_rewrite Hp.
  setoid_rewrite Hq.
  setoid_rewrite disjLDistConj.
  setoid_rewrite disjRDistConj.
  algMagic.
  setoid_rewrite <- disjConf.
  setoid_rewrite <- disjInteg.
  repeat setoid_rewrite voiceOfInteg.
  destruct (normJorJ _ pfpc _ pfqc) as [pcORqc [pfpcORqc ?]].
  pose proof voiceOfNormJConf _ pfpcORqc.
  assert (∇(pcORqc→) == pcORqc←).
    rewrite H0. autoalg.
  assert (pc ∨ qc == pcORqc).
    assert (pcORqc = projT1 (norm (pc ∨ qc))) by (rewrite H ; auto).
    rewrite H2.
    apply norm_correct.
  setoid_rewrite <- H2 in H1.
  setoid_rewrite H1.
  rewrite voiceOfNormJConf ; auto.
  repeat setoid_rewrite disjInteg.
  eexists.
  eapply AF_ConjR ; eautoflam.
Qed.

Hint Resolve voiceDisjMonotonic : flam.

Lemma staticBasisDeriv:
 forall n c p q r s,
       [c ⊢ p→ ∧ r← ≽ q→ ∧ s← # n]   
       -> (exists n1 n2, [c ⊢ p→ ≽ q→ # n1]
                         /\ [c ⊢ r← ≽ s← # n2] /\ n1 <= n /\ n2 <= n)
          \/ p→ ∧ r← == ⊤.
(* begin hide *)
Proof.
  assert (forall n c p q r s,
       [c ⊢ p→ ∧ r← ≽ q→ ∧ s← # n]   
       -> (exists n1 n2, [c ⊢ p→ ≽ q→ # n1]
                         /\ [c ⊢ r← ≽ s← # n2] /\ n1 <= n /\ n2 <= n)
          \/ p→ ∧ r← == ⊤).
    induction n as [|n IHn] using strong_induction ; intros.
      inversion H ; try (omega).
      (* bot *)
      assert (q → == ⊥ /\ s← == ⊥).
      eapply conjIsBot; auto.
      destruct H10.
      left.
      do 2 eexists.
      split; eautoflam.

      (* top *)
      right; auto.

      (* refl *)
       assert (p→ == q→).
        assert (p→ == (p→ ∧ r←)→) by algMagic.
        setoid_rewrite H10.
        setoid_rewrite H1.
        algMagic.
      setoid_rewrite H10.

      assert (r← == s←).
        assert (r← == (p→ ∧ r←)←) by algMagic.
        setoid_rewrite H11.
        setoid_rewrite H1.
        algMagic.
      setoid_rewrite H11.
      
      left.
      do 2 eexists.
      split.
      eapply AF_Refl; auto using refl.
      split.
      eapply AF_Refl; auto using refl.
      auto.

      (* projR_conf *)
      assert (p→ == q→).
        assert (p→ == (p→ ∧ r←)→) by algMagic.
        setoid_rewrite H10.
        setoid_rewrite <- confIdemp.
        setoid_rewrite <- H1.
        algMagic.
      setoid_rewrite H10.
      
      assert (s← == ⊥).
        assert (⊥ == (q→ ∧ s←)←).
          setoid_rewrite H1.
          algMagic.
        setoid_rewrite H11.
        algMagic.
      setoid_rewrite H11.

      left.
      do 2 eexists.
      split.
      eapply AF_Refl; auto using refl.
      split.
      eapply AF_Bot; auto using refl.
      auto.

       (* projR_integ *)
      assert (r← == s←).
        assert (r← == (p→ ∧ r←)←) by algMagic.
        setoid_rewrite H10.
        setoid_rewrite <- integIdemp.
        setoid_rewrite <- H1.
        algMagic.
      setoid_rewrite H10.
      
      assert (q→ == ⊥).
        assert (⊥ == (q→ ∧ s←)→).
          setoid_rewrite H1.
          algMagic.
        setoid_rewrite H11.
        algMagic.
      setoid_rewrite H11.

      left.
      do 2 eexists.
      split.
      eapply AF_Bot; auto using refl.
      split.
      eapply AF_Refl; auto using refl.
      auto.

      (* projR_owns *)
      assert (q → ∧ s ← == (q → ∧ s ←): q').
        setoid_rewrite H1.
        algMagic.
      assert (q→ == p:q'→).
        assert (q→ == (q→ ∧ s←)→) by algMagic.
        setoid_rewrite H10 in H11.
        setoid_rewrite H11.
        setoid_rewrite H1.
        algMagic.
      setoid_rewrite H11.

      assert (s← == r:q'←).
        assert (s← == (q→ ∧ s←)←) by algMagic.
        setoid_rewrite H10 in H12.
        setoid_rewrite H12.
        setoid_rewrite H1.
        algMagic.
      setoid_rewrite H12.

      left.
      do 2 eexists.
      split.
      setoid_rewrite ownConfAssoc.
      setoid_rewrite <- ownConf.
      eapply AF_ProjR_owns; auto using refl.
      split.
      setoid_rewrite ownIntegAssoc.
      setoid_rewrite <- ownInteg.
      eapply AF_ProjR_owns; auto using refl.
      auto.

      (* AF_Del *)
      contradict H5.
      apply PH.emptyIsEmpty.
      
      (* inductive cases *)
      inversion H.

        (* AF_Proj_conf *)
        setoid_rewrite <- (confIntegBasis p') in H3.
        setoid_rewrite <- (confIntegBasis q') in H3.
        assert (n1 <= n) by omega.
        pose proof IHn n1 H13 _ _ _ _ _ H3.
        
        setoid_rewrite <- H1 in H14.
        setoid_rewrite <- H2 in H14.
        
        assert (r← == ⊥).
          assert ((p→ ∧ r←)← == ⊥).
            setoid_rewrite H1. algMagic.
          setoid_rewrite <- H15. algMagic.
        setoid_rewrite H15.    
        setoid_rewrite H15 in H14.    
        
        assert (s← == ⊥).
          assert ((q→ ∧ s←)← == ⊥).
            setoid_rewrite H2. algMagic.
          setoid_rewrite <- H16. algMagic.
        setoid_rewrite H16 .    
        setoid_rewrite H16 in H14.    

        setoid_rewrite conjIdent in H14.    
          
        destruct H14 as [|isTop].
          (* p' is not top *)
          left.
          do 2 destruct H14. 
          destruct H14.
          destruct H17.
          destruct H18.
          exists x, 0.
          split. eauto.
          split. eautoflam.
          split;omega.
          (* p' is top *)
          left.
          assert (p→ == ⊤→).
            setoid_rewrite <- isTop.
            algMagic.
          exists 1, 0.
          setoid_rewrite H14.
          split. eautoflam.
          split. eautoflam.
          split; omega.

        (* AF_Proj_integ *)
        setoid_rewrite <- (confIntegBasis p') in H3.
        setoid_rewrite <- (confIntegBasis q') in H3.
        assert (n1 <= n) by omega.
        pose proof IHn n1 H13 _ _ _ _ _ H3.
        
        setoid_rewrite <- H1 in H14.
        setoid_rewrite <- H2 in H14.
        
        assert (p→ == ⊥).
          assert ((p→ ∧ r←)→ == ⊥).
            setoid_rewrite H1. algMagic.
          setoid_rewrite <- H15. algMagic.
        setoid_rewrite H15.    
        setoid_rewrite H15 in H14.    
        
        assert (q→ == ⊥).
          assert ((q→ ∧ s←)→ == ⊥).
            setoid_rewrite H2. algMagic.
          setoid_rewrite <- H16. algMagic.
        setoid_rewrite H16 .    
        setoid_rewrite H16 in H14.    

        setoid_rewrite conjComm in H14.    
        setoid_rewrite conjIdent in H14.    
          
        destruct H14 as [|isTop].
          (* p' is not top *)
          left.
          do 2 destruct H14. 
          destruct H14.
          destruct H17.
          destruct H18.
          exists 0, x0.
          split. eautoflam.
          split. eauto.
          split;omega.
          (* p' is top *)
          left.
          assert (r← == ⊤←).
            setoid_rewrite <- isTop.
            algMagic.
          exists 0, 1.
          setoid_rewrite H14.
          split. eautoflam.
          split. eautoflam.
          split; omega.
        (* AF_Own1 *)
        destruct (emptyUnion2 _ _ H5) as [HD1 HD2].
        rewrite HD1, HD2 in *.

        assert (p→ == (o:r0)→) by (setoid_rewrite <- H1; algMagic).
        assert (q→ == (o':r')→) by (setoid_rewrite <- H2; algMagic).
        assert (r← == (o:r0)←) by (setoid_rewrite <- H1; algMagic).
        assert (s← == (o':r')←) by (setoid_rewrite <- H2; algMagic).

        setoid_rewrite <- (confIntegBasis o) in H3.
        setoid_rewrite <- (confIntegBasis o') in H3.
        assert (n1 <= n) by omega.
        pose proof IHn n1 H19 _ _ _ _ _ H3.
        
        setoid_rewrite <- (confIntegBasis r0) in H4.
        setoid_rewrite <- (confIntegBasis r') in H4.
        assert (n2 <= n) by omega.
        pose proof IHn n2 H21 _ _ _ _ _ H4.

        setoid_rewrite H15.
        setoid_rewrite H16.
        setoid_rewrite H17.
        setoid_rewrite H18.
        
        setoid_rewrite <- (confIdemp (o:r0)).
        setoid_rewrite <- (confIdemp (o':r')).
        setoid_rewrite <- (integIdemp (o:r0)).
        setoid_rewrite <- (integIdemp (o':r')).
        setoid_rewrite <- ownConf.
        setoid_rewrite <- ownInteg.
        setoid_rewrite <- ownConfAssoc.
        setoid_rewrite <- ownIntegAssoc.
        
        destruct H20, H22.
        (* neither is top *)
          left.
          do 2 destruct H20, H22. 
          exists (x+x0+1), (x1+x2+1).
          destruct H20.
          destruct H23.
          destruct H24.
          
          destruct H22.
          destruct H26.
          destruct H27.

          split.
          eapply AF_Own1 with (o := o→) (r := r0→)
                              (o' := o'→) (r':= r'→); eauto using refl.
          split.
          eapply AF_Own1 with (o := o←) (r := r0←)
                              (o' := o'←) (r':= r'←); eauto using refl.
          split; omega.
        (* r0 is top *)
          left.
          do 2 destruct H20.
          exists (x+1), (x0+1).
          destruct H20.
          destruct H23.
          destruct H24.

          assert (r0→ == ⊤→) by (setoid_rewrite <- H22; algMagic).
          assert (r0← == ⊤←) by (setoid_rewrite <- H22; algMagic).
          setoid_rewrite  H26.
          setoid_rewrite  H27.
          
          setoid_replace (o→:⊤→) with (o→:⊤).
          setoid_replace (o←:⊤←) with (o←:⊤).
          split.
          eapply AF_Own1 with (o := o→) (r := ⊤)
                              (o' := o'→) (r':= r'→); eauto using refl.
          autoflam.
          
          split.
          eapply AF_Own1 with (o := o←) (r := ⊤)
                              (o' := o'←) (r':= r'←); eauto using refl.
          autoflam.
          split; omega.

          algMagic.
          setoid_rewrite ownIntegAssoc.
          setoid_rewrite integIdemp.
          algMagic.
          
          algMagic.
          setoid_rewrite ownConfAssoc.
          setoid_rewrite confIdemp.
          algMagic.
        (* o is top *)
          left.
          do 2 destruct H22.
          exists (x+1), (x0+1).
          destruct H22.
          destruct H23.
          destruct H24.

          assert (o→ == ⊤→) by (setoid_rewrite <- H20; algMagic).
          assert (o← == ⊤←) by (setoid_rewrite <- H20; algMagic).
          setoid_rewrite  H26.
          setoid_rewrite  H27.
          
          setoid_replace (⊤→:r0→) with (⊤:r0→). 
          setoid_replace (⊤←:r0←) with (⊤:r0←). 
          split.
          eapply AF_Own1 with (o := ⊤) (r := r0→)
                              (o' := o'→) (r':= r'→); eauto using refl.
          autoflam. omega.
          
          split.
          eapply AF_Own1 with (o := ⊤) (r := r0←)
                              (o' := o'←) (r':= r'←); eauto using refl.
          autoflam. omega.
          split; omega.

          algMagic.
          setoid_rewrite ownIntegAssoc.
          setoid_rewrite integIdemp.
          algMagic.
          
          algMagic.
          setoid_rewrite ownConfAssoc.
          setoid_rewrite confIdemp.
          algMagic.
          
         (* o:r0 is top *)
          left.
          exists (1), (1).

          assert (o→ == ⊤→) by (setoid_rewrite <- H20; algMagic).
          assert (o← == ⊤←) by (setoid_rewrite <- H20; algMagic).
          assert (r0→ == ⊤→) by (setoid_rewrite <- H22; algMagic).
          assert (r0← == ⊤←) by (setoid_rewrite <- H22; algMagic).
          setoid_rewrite H23.
          setoid_rewrite H24.
          setoid_rewrite H25.
          setoid_rewrite H26.
         
          setoid_replace (⊤→:⊤→) with (⊤→). 
          setoid_replace (⊤←:⊤←) with (⊤←). 
          
          setoid_rewrite ownConfAssoc.
          setoid_rewrite ownIntegAssoc.
          
          split.
          eapply AF_Proj_conf; eauto using refl.
          autoflam. omega.
          split.
          eapply AF_Proj_integ; eauto using refl.
          autoflam. omega.
          split; omega.

          algMagic. algMagic.

         (* AF_Own2 *)
        destruct (emptyUnion2 _ _ H5) as [HD1 HD2].
        rewrite HD1, HD2 in *.

        assert (p→ == (o:r0)→) by (setoid_rewrite <- H1; algMagic).
        assert (q→ == (o':r')→) by (setoid_rewrite <- H2; algMagic).
        assert (r← == (o:r0)←) by (setoid_rewrite <- H1; algMagic).
        assert (s← == (o':r')←) by (setoid_rewrite <- H2; algMagic).

        setoid_rewrite <- (confIntegBasis o) in H3.
        setoid_rewrite <- (confIntegBasis o') in H3.
        assert (n1 <= n) by omega.
        pose proof IHn n1 H19 _ _ _ _ _ H3.
        
        setoid_rewrite <- (confIntegBasis r0) in H4.
        setoid_rewrite <- (confIntegBasis r') in H4.
        setoid_rewrite conjDistOwn in H4.
        setoid_rewrite ownConfAssoc in H4.
        setoid_rewrite ownIntegAssoc in H4.
        assert (n2 <= n) by omega.
        pose proof IHn n2 H21 _ _ _ _ _ H4.

        setoid_rewrite H15.
        setoid_rewrite H16.
        setoid_rewrite H17.
        setoid_rewrite H18.
        
        setoid_rewrite <- (confIdemp (o:r0)).
        setoid_rewrite <- (confIdemp (o':r')).
        setoid_rewrite <- (integIdemp (o:r0)).
        setoid_rewrite <- (integIdemp (o':r')).
        setoid_rewrite <- ownConf.
        setoid_rewrite <- ownInteg.
        setoid_rewrite <- ownConfAssoc.
        setoid_rewrite <- ownIntegAssoc.
        
        destruct H20, H22.
        (* neither is top *)
          left.
          do 2 destruct H20, H22. 
          exists (x+x0+1), (x1+x2+1).
          destruct H20.
          destruct H23.
          destruct H24.
          
          destruct H22.
          destruct H26.
          destruct H27.

          split.
          eapply AF_Own2 with (o := o→) (r := r0→)
                              (o' := o'→) (r':= r'→); eauto using refl.
          setoid_rewrite ownConfAssoc.
          setoid_rewrite ownConf.
          setoid_rewrite confIdemp.
          auto.
          split.
          eapply AF_Own2 with (o := o←) (r := r0←)
                              (o' := o'←) (r':= r'←); eauto using refl.
          setoid_rewrite ownIntegAssoc.
          setoid_rewrite ownInteg.
          setoid_rewrite integIdemp.
          auto.
          split; omega.
        (* r0 is top *)
          left.
          do 2 destruct H20.
          exists (x+1), (x0+1).
          destruct H20.
          destruct H23.
          destruct H24.

          assert (r0→ == ⊤→) by (setoid_rewrite <- H22; algMagic).
          assert (r0← == ⊤←) by (setoid_rewrite <- H22; algMagic).
          setoid_rewrite  H26.
          setoid_rewrite  H27.
          
          setoid_replace (o→:⊤→) with (o→:⊤).
          setoid_replace (o←:⊤←) with (o←:⊤).
          split.
          eapply AF_Own2 with (o := o→) (r := ⊤)
                              (o' := o'→) (r':= r'→); eauto using refl.
          autoflam.
          
          split.
          eapply AF_Own2 with (o := o←) (r := ⊤)
                              (o' := o'←) (r':= r'←); eauto using refl.
          autoflam.
          split; omega.

          algMagic.
          setoid_rewrite ownIntegAssoc.
          setoid_rewrite integIdemp.
          algMagic.
          
          algMagic.
          setoid_rewrite ownConfAssoc.
          setoid_rewrite confIdemp.
          algMagic.
        (* o is top *)
          left.
          do 2 destruct H22.
          exists (x+1), (x0+1).
          destruct H22.
          destruct H23.
          destruct H24.

          assert (o→ == ⊤→) by (setoid_rewrite <- H20; algMagic).
          assert (o← == ⊤←) by (setoid_rewrite <- H20; algMagic).
          setoid_rewrite  H26.
          setoid_rewrite  H27.
          
          setoid_replace (⊤→:r0→) with (⊤:r0→). 
          setoid_replace (⊤←:r0←) with (⊤:r0←). 
          split.
          eapply AF_Own2 with (n1 := 0)  (n2 := x) (o := ⊤) (r := r0→)
                              (o' := o'→) (r':= r'→); eauto using refl.
          autoflam.
          setoid_rewrite ownConfAssoc.
          setoid_rewrite ownConf.
          setoid_rewrite confIdemp.
          auto.
          
          split.
          eapply AF_Own2 with (n1 := 0)  (n2 := x0) (o := ⊤) (r := r0←)
                              (o' := o'←) (r':= r'←); eauto using refl.
          autoflam.
          setoid_rewrite ownIntegAssoc.
          setoid_rewrite ownInteg.
          setoid_rewrite integIdemp.
          auto.
          omega.

          algMagic.
          setoid_rewrite ownIntegAssoc.
          setoid_rewrite integIdemp.
          algMagic.
          
          algMagic.
          setoid_rewrite ownConfAssoc.
          setoid_rewrite confIdemp.
          algMagic.
          
         (* o:r0 is top *)
          left.
          exists (1), (1).

          assert (o→ == ⊤→) by (setoid_rewrite <- H20; algMagic).
          assert (o← == ⊤←) by (setoid_rewrite <- H20; algMagic).
          assert (r0→ == ⊤→) by (setoid_rewrite <- H22; algMagic).
          assert (r0← == ⊤←) by (setoid_rewrite <- H22; algMagic).
          setoid_rewrite H23.
          setoid_rewrite H24.
          setoid_rewrite H25.
          setoid_rewrite H26.
         
          setoid_replace (⊤→:⊤→) with (⊤→). 
          setoid_replace (⊤←:⊤←) with (⊤←). 
          
          setoid_rewrite ownConfAssoc.
          setoid_rewrite ownIntegAssoc.
          
          split.
          eapply AF_Proj_conf; eauto using refl.
          autoflam. omega.
          split.
          eapply AF_Proj_integ; eauto using refl.
          autoflam. omega.
          split; omega.

          algMagic. algMagic.
          
        (* AF_ConjL1*)
        setoid_rewrite <- (confIntegBasis p1) in H2.
        assert (n1 <= n) by omega.
        pose proof IHn n1 H12 _ _ _ _ _ H2.

        assert (p→ == p1→ ∧ p2→).
          setoid_rewrite <- conjConf.
          setoid_rewrite <- H1.
          algMagic.
          
        assert (r← == p1← ∧ p2←).
          setoid_rewrite <- conjInteg.
          setoid_rewrite <- H1.
          algMagic.

        setoid_rewrite H14.
        setoid_rewrite H15.
        
        destruct H13 as [|isTop].
         (* p1 is not top *)
          left.
          do 2 destruct H13. 
          destruct H13.
          destruct H16.
          destruct H17.
          exists (x+1), (x0+1).
          split.
          eapply AF_ConjL1 with (p1 := p1→); eauto using refl.
          split.
          eapply AF_ConjL1 with (p1 := p1←); eauto using refl.
          omega.
          (* p1 is top *)
          left.
          assert (p1→ == ⊤→).
            setoid_rewrite <- isTop.
            algMagic.
          assert (p→ == ⊤→).
            setoid_rewrite H14.
            setoid_rewrite H13.
            setoid_rewrite <- conjConf.
            algMagic.
          assert (p1← == ⊤←).
            setoid_rewrite <- isTop.
            algMagic.
          assert (r← == ⊤←).
            setoid_rewrite H15.
            setoid_rewrite H17.
            setoid_rewrite <- conjInteg.
            algMagic.
 
          exists 1, 1.
          setoid_rewrite H13.
          setoid_rewrite <- conjConf.
          setoid_rewrite H17.
          setoid_rewrite <- conjInteg.
          algMagic.
          
          split. eautoflam.
          split. eautoflam.
          split; omega.
         
        (* AF_ConjR*)
        destruct (emptyUnion2 _ _ H4) as [HD1 HD2].
        rewrite HD1, HD2 in *.

        assert (q→ == (q1 ∧ q2)→) by (setoid_rewrite <- H1; algMagic).
        assert (s← == (q1 ∧ q2)←) by (setoid_rewrite <- H1; algMagic).

        setoid_rewrite <- (confIntegBasis q1) in H2.
        assert (n1 <= n) by omega.
        pose proof IHn n1 H16 _ _ _ _ _ H2.
        
        setoid_rewrite <- (confIntegBasis q2) in H3.
        assert (n2 <= n) by omega.
        pose proof IHn n2 H18 _ _ _ _ _ H3.

        setoid_rewrite H14.
        setoid_rewrite H15.
        
        destruct H17, H19.
        (* neither is top *)
          left.
          do 2 destruct H17, H19. 
          exists (x+x0+1), (x1+x2+1).
          destruct H17.
          destruct H20.
          destruct H21.
          
          destruct H19.
          destruct H23.
          destruct H24.

          split.
          eapply AF_ConjR with (p := p→) (q1 := q1→)
                               (q2:= q2→); eauto using refl.
          algMagic.
          split.
          eapply AF_ConjR with (p := r←) (q1 := q1←)
                               (q2:= q2←); eauto using refl.
          algMagic.
          omega.
        (* p is top *)
          right; auto.
          right; auto.
          right; auto.
        (* AF_DisjL*)
        destruct (emptyUnion2 _ _ H4) as [HD1 HD2].
        rewrite HD1, HD2 in *.

        assert (p→ == (p1 ∨ p2)→) by (setoid_rewrite <- H1; algMagic).
        assert (r← == (p1 ∨ p2)←) by (setoid_rewrite <- H1; algMagic).

        setoid_rewrite <- (confIntegBasis p1) in H2.
        assert (n1 <= n) by omega.
        pose proof IHn n1 H16 _ _ _ _ _ H2.
        
        setoid_rewrite <- (confIntegBasis p2) in H3.
        assert (n2 <= n) by omega.
        pose proof IHn n2 H18 _ _ _ _ _ H3.

        setoid_rewrite H14.
        setoid_rewrite H15.
        
        destruct H17, H19.
        (* neither is top *)
          left.
          do 2 destruct H17, H19. 
          exists (x+x0+1), (x1+x2+1).
          destruct H17.
          destruct H20.
          destruct H21.
          
          destruct H19.
          destruct H23.
          destruct H24.

          split.
          eapply AF_DisjL with (q := q→) (p1 := p1→)
                               (p2:= p2→); eauto using refl.
          algMagic.
          split.
          eapply AF_DisjL with (q := s←) (p1 := p1←)
                               (p2:= p2←); eauto using refl.
          algMagic.
          omega.
        (* p2 is top *)
          left.
          do 2 destruct H17. 
          exists (x), (x0).
          destruct H17.
          destruct H20.
          destruct H21.
          
          setoid_rewrite (confIntegBasis p2) in H19.
          setoid_rewrite H19.
          setoid_rewrite disjIdent.
          
          split. auto.
          split. auto.
          omega.
         (* p1 is top *)
          left.
          do 2 destruct H19. 
          exists (x), (x0).
          destruct H19.
          destruct H20.
          destruct H21.
          
          setoid_rewrite (confIntegBasis p1) in H17.
          setoid_rewrite H17.
          setoid_rewrite disjComm.
          setoid_rewrite disjIdent.
          
          split. auto.
          split. auto.
          omega.
         (* p is top *)
          right.
          setoid_rewrite (confIntegBasis p1) in H17.
          setoid_rewrite (confIntegBasis p2) in H19.
          setoid_rewrite H17.
          setoid_rewrite H19.
          algMagic.
          
         (* AF_DisjR1*)
        setoid_rewrite <- (confIntegBasis q1) in H2.
        assert (n1 <= n) by omega.
        pose proof IHn n1 H12 _ _ _ _ _ H2.

        assert (q→ == q1→ ∨ q2→).
          setoid_rewrite <- disjConf.
          setoid_rewrite <- H1.
          algMagic.
          
        assert (s← == q1← ∨ q2←).
          setoid_rewrite <- disjInteg.
          setoid_rewrite <- H1.
          algMagic.

        setoid_rewrite H14.
        setoid_rewrite H15.
        
        destruct H13 as [|isTop].
         (* q1 is not top *)
          left.
          do 2 destruct H13. 
          destruct H13.
          destruct H16.
          destruct H17.
          exists (x+1), (x0+1).
          split.
          eapply AF_DisjR1 with (q1 := q1→); eauto using refl.
          split.
          eapply AF_DisjR1 with (q1 := q1←); eauto using refl.
          omega.
          (* q1 is top *)
          left.
          assert (p→ == ⊤→).
            setoid_rewrite <- isTop.
            algMagic.
          assert (r← == ⊤←).
            setoid_rewrite <- isTop.
            algMagic.
          setoid_rewrite H13.
          setoid_rewrite H16.
            
          setoid_rewrite <- disjConf.
          setoid_rewrite <- disjInteg.
          
          exists 1, 1.
          split.
          eapply AF_Proj_conf; eauto using refl; eautoflam.
          split.
          eapply AF_Proj_integ; eauto using refl; eautoflam.
          split; omega.
        (* AF_Trans *)
        destruct (emptyUnion2 _ _ H3) as [HD1 HD2].
        rewrite HD1, HD2 in *.

        setoid_rewrite <- (confIntegBasis q0) in H1.
        assert (n1 <= n) by omega.
        pose proof IHn n1 H13 _ _ _ _ _ H1.
        
        setoid_rewrite <- (confIntegBasis q0) in H2.
        assert (n2 <= n) by omega.
        pose proof IHn n2 H15 _ _ _ _ _ H2.

        destruct H14, H16.
        (* neither is top *)
          left.
          do 2 destruct H14, H16. 
          exists (x+x0+1), (x1+x2+1).
          destruct H14.
          destruct H17.
          destruct H18.
          
          destruct H16.
          destruct H20.
          destruct H21.

          split.
          eapply AF_Trans with (p := p→) (q := q0→)
                               (r:= q→); eauto using refl.
          split.
          eapply AF_Trans with (p := r←) (q := q0←)
                               (r:= s←); eauto using refl.
          omega.
        (* q0 is top *)
          right.
          do 3 destruct H14.
          assert (q0→ == ⊤→).
            setoid_rewrite  <- H16.
            algMagic.
          assert (q0← == ⊤←).
            setoid_rewrite  <- H16.
            algMagic.
          setoid_rewrite H18 in H14. 
          setoid_rewrite H19 in H17. 
          
          (* XXX: change after static actsfor equivalence refactor *)
          eapply actsForTop with (n:=x+x0+5). 
            setoid_rewrite <- (confIntegBasis ⊤). 
            eapply AF_ConjR with (n := x+x0+5); eauto using refl.
            eapply AF_ConjL1 with (n := x+2); eauto using refl.
            eapply AF_ConjL1 with (n := x+1); eauto using refl.
             algMagic.
             setoid_rewrite (confIntegBasis ⊤). 
             eapply H14.
             omega.
            setoid_rewrite (confIntegBasis ⊤). 
            setoid_rewrite conjComm.
            eapply AF_ConjL1 with (n := x0+2); eauto using refl.
            eapply AF_ConjL1 with (n := x0+1); eauto using refl.
              algMagic.
              destruct H17.
              eapply H17.
              omega.
            omega.
       (*  p → ∧ r ← = top *)
       right. auto.
       right. auto.
       (* AF_Fwd *)
        destruct (emptyUnion2 _ _ H5) as [HD1 HtransferD2].
        pose proof transferIsEmpty _ _ HtransferD2 as HD2.
        rewrite HD1, HD2, joinTopConf, meetTopInteg in *.

        assert (exists n', [c ⊢  p → ∧ r ← ≽ q → ∧ s ←# n']
                           /\ n' <= n2)
            by (eapply staticCoerce ; eauto with zarith).
        do 2 destruct H15.
        left.
        exists (x+1), (x+1).
        split.
        eapply AF_Proj_conf with (p' := p→ ∧ r←) (q' := q→ ∧ s←).
        algMagic.
        algMagic.
        eauto.
        auto.
        split.
        eapply AF_Proj_integ with (p' := p→ ∧ r←) (q' := q→ ∧ s←).
        algMagic.
        algMagic.
        eauto.
        auto.
        omega.

        (* AF_Weaken *)
        destruct (PH.emptyUnion2 _ _ H6) as [HH1 HH2].
        destruct (emptyUnion3 _ _ _ H7) as [HD1 [HD2 HD3]].
        rewrite HH1, HD1, HD2, HD3, (joinTopConf l') in *.

        assert (exists n', [c ⊢ ⊤→ ⊑ pc' # n'] /\ n' < n2)
            by (eapply staticRobust ; eauto with zarith).
        assert (exists n', [c ⊢ l' ⊑ ⊤← # n'] /\ n' < n3)
            by (eapply staticRobust ; eauto with zarith).
        do 2 destruct H17, H18.

        assert (pc' == ⊤→) as Hpc by (eapply flowsFromTopConf ; eauto).
        assert (l' == ⊤←) as Hl by (eapply flowsToTopInteg ; eauto).
        rewrite Hpc, Hl in *.

        left.
        exists (n1+1), (n1+1).
        split.
        eapply AF_Proj_conf with (p' := p→ ∧ r←) (q' := q→ ∧ s←).
        algMagic.
        algMagic.
        eauto.
        auto.
        split.
        eapply AF_Proj_integ with (p' := p→ ∧ r←) (q' := q→ ∧ s←).
        algMagic.
        algMagic.
        eauto.
        auto.
        omega.
auto.
Grab Existential Variables.
eapply (native c).
eapply (native c).
Qed.

Lemma normJConf:
  forall c p q n,
    normJ p
      -> normJ q
      -> [c ⊢ p→ ≽ q→ # n]
      -> [c ⊢ p ≽ q].
(* begin hide *)
Proof.
  assert (forall n c p q r,
      normJ p
        -> normJ q
        -> normJ r
        -> [c ⊢ p→ ∧ r← ≽ q→ # n]
        -> [c ⊢ p ≽ q]).
    induction n as [|n IHn] using strong_induction ; intros.
      inversion H2 ; try (omega).
        (* AF_Bot *)
        assert (q == ⊥) by (eapply normJConfIsBotImpliesBot ; eauto).
        setoid_rewrite H13.
        eautoflam.

        (* AF_Top *)
        assert (p == ⊤).
          eapply normJConfEq; eauto.
            apply (normJ_M ⊤ (normM_L ⊤ normL_top)).  
          setoid_rewrite <- H4.
          algMagic.
        setoid_rewrite H13.
        eautoflam.
        
        (* AF_Refl *)
        assert (p == q).
          eapply normJConfEq; auto.
          setoid_rewrite <- confIdemp.
          setoid_rewrite <- H4.
          algMagic.
        eautoflam.

        (* AF_ProjR_conf *)
        assert (p == q).
          eapply normJConfEq ; auto.
          setoid_rewrite H4.
          algMagic.
        setoid_rewrite H13.
        eautoflam.

        (* AF_ProjR_integ *)
        assert (q == ⊥).
          eapply normJConfIsBotImpliesBot ; eauto.
          rewrite <- confIdemp.
          setoid_rewrite H4.
          algMagic.
        setoid_rewrite H13.
        eautoflam.

        (* AF_ProjR_owns *)
        assert (Hobot: r : q'← == ⊥).
        {
          assert (q→ == ((p → ∧ r ←) : q')→).
          {
            setoid_rewrite <- confIdemp.
            setoid_rewrite H4.
            algMagic.
          }

          assert (((p → ∧ r ←) : q')← == ⊥).
          { setoid_rewrite <- H4. algMagic. }

          setoid_rewrite ownDistConj in H14.
          setoid_rewrite conjInteg in H14.
          assert ((r ← : q') ← == ⊥).
          { eapply conjIsBot. eapply H14. }

          setoid_rewrite <- ownIntegAssoc in H15.
          setoid_rewrite ownInteg in H15.
          setoid_rewrite <- ownIntegAssoc in H15.
          setoid_rewrite integIdemp in H15.
          auto.
        }

        pose proof norm_correct q' as Hnorm_q'.
        destruct (norm q') as [nq' pfnq'].
        simpl in Hnorm_q'.
        induction pfnq'.
        {
          (* q' normalized via norm_J *)
          pose proof normOwnsJJ_correct p H p1 n as Hnorm_pp1.
          destruct (normOwnsJJ p H p1 n) as [pn pfpp1].
          simpl in Hnorm_pp1.
          setoid_replace q with (p:p1). eautoflam.
          rewrite Hnorm_pp1.
          eapply normJConfEq ; eauto.
          rewrite <- Hnorm_pp1, <- Hnorm_q', <- ownConf.
          setoid_rewrite H4.
          setoid_rewrite ownDistConj.
          setoid_rewrite ownInteg.
          setoid_rewrite <- ownIntegAssoc.
          setoid_rewrite Hobot.
          algMagic.
        }
        {
          (* q' normalized via norm_conf *)
          pose proof normOwnsJJ_correct p H p1 n as Hnorm_pp1.
          destruct (normOwnsJJ p H p1 n) as [pp1 pfpp1].
          simpl in Hnorm_pp1.
          setoid_replace q with (p:p1). eautoflam.
          rewrite Hnorm_pp1.
          eapply normJConfEq ; eauto.
          rewrite <- Hnorm_pp1, <- (confIdemp (p:p1)), <- ownConf,
              <- ownConfAssoc.
          rewrite <- Hnorm_q', H4.
          algMagic.
          setoid_rewrite <- ownIntegAssoc.
          setoid_rewrite Hobot.
          algMagic.
        }
        {
          (* q' normalized via norm_integ *)
          assert (q→ == ⊥).
            rewrite H4, Hnorm_q', ownIntegAssoc.
            algMagic.
            rewrite <- ownIntegAssoc.
            rewrite <- integIdemp.
            rewrite <- Hnorm_q'.
            auto.

          assert (q == ⊥) as Hq.
          { eapply normJConfIsBotImpliesBot ; auto. }

          setoid_rewrite Hq. eautoflam.
        }
        {
          (* q' normalized via norm_decomp *)
          subst.
          pose proof normOwnsJJ_correct p H p1 n as Hnorm_pp1.
          destruct (normOwnsJJ p H p1 n) as [pp1 pfpp1].
          simpl in Hnorm_pp1.
          assert (q→ == (p:p1)→).
          {
            rewrite <- (confIdemp q).
            rewrite H4, Hnorm_q'.
            repeat rewrite ownDistConj.
            repeat rewrite conjDistOwn.
            repeat rewrite ownConf.
            repeat rewrite ownInteg.
            repeat rewrite ownConfAssoc.
            repeat rewrite ownIntegAssoc.
            algMagic.
          }

          setoid_replace q with (p:p1) ; eautoflam.
          rewrite Hnorm_pp1.
          eapply normJConfEq ; eauto.
          rewrite <- Hnorm_pp1.
          auto.
        }

        (* AF_Del *)
        contradict H8.
        apply PH.emptyIsEmpty.

      inversion H2.
        (* AF_Proj_conf *)
        destruct (normForm p') as [p'c [p'i [pfp'c [pfp'i Hp']]]].
        destruct (normForm q') as [q'c [q'i [pfq'c [pfq'i Hq']]]].

        assert (p→ ∧ r← == p→).
          setoid_rewrite H4.
          setoid_rewrite <- (confIdemp p').
          setoid_rewrite <- H4.
          algMagic.
          
        assert((p→ ∧ r←)← == r←) by algMagic.
        assert (r← == ⊥).
          setoid_rewrite <- H17.
          setoid_rewrite H16.
          algMagic.
 
        rewrite H18 in *.
        rewrite conjIdent in *.
        assert (p == p'c).      
          eapply normJConfEq; auto.
          setoid_rewrite Hp' in H4.
          setoid_rewrite H4.
          algMagic.
          
        setoid_rewrite <- (confIntegBasis p') in H6.
        setoid_rewrite <- (confIntegBasis q') in H6.
        assert((exists n', [c ⊢ p'→ ≽ q'→ # n'] /\ n' <= n1) \/ p' == ⊤).
          pose proof staticBasisDeriv _ _ _ _ _ _ H6 as Hbasis.
          destruct Hbasis.
            do 3 destruct H20.
            destruct H21.
            destruct H22.
            left.
            eexists.
            eautoflam.
            (* otherwise p' is top *)
            right.
            setoid_rewrite <- H20.
            algMagic.
        destruct H20.
        do 2 destruct H20.
        (* case1 there's an x <= n1 *)
        eapply IHn with (j := x).
          omega. auto. auto. apply H1.
          setoid_rewrite H18.
          setoid_rewrite conjIdent.
          setoid_rewrite H5.
          setoid_rewrite H4.
          eautoflam.
        (* case2 p' is top *)
        rewrite H20 in *.
        assert(p == ⊤).
          eapply normJConfEq; auto.
          apply (normJ_M ⊤ (normM_L ⊤ normL_top)).  
        setoid_rewrite H21.
        eautoflam.
        
        (* AF_Proj_integ *)
        assert (q→← == q'←).
        setoid_rewrite  H5.
        algMagic.
        assert (⊥ == q'←).
        setoid_rewrite <- H16.
        algMagic.
        
        assert (p→ == p'←→).
          assert (p→ == (p→ ∧ r←)→) by algMagic.
          setoid_rewrite H4 in H18.
          auto.
        
        assert (⊥ == p→).
        setoid_rewrite H18.
        algMagic.

        assert (q→← == q'←).
        setoid_rewrite H5.
        algMagic.
        assert (⊥ == q'←).
        setoid_rewrite <- H20.
        algMagic.
        assert (q→ == q'←→).
        setoid_rewrite <- H5.
        algMagic.
        assert (⊥ == q→).
        setoid_rewrite  H21.
        algMagic.
        
        assert (p == ⊥).
        eapply normJConfIsBotImpliesBot; algMagic.
         
        assert (q == ⊥).
        eapply normJConfIsBotImpliesBot; algMagic.
        
        setoid_rewrite H24.
        setoid_rewrite H25.
        eexists.
        eapply AF_Refl.
        algMagic.
        autoflam.
        
        (* AF_Own1 *)
        destruct (emptyUnion2 _ _ H8) as [HD1 HD2].
        rewrite HD1, HD2 in *.
        
        assert (p → == (o : r0)→) as Hp.
        setoid_rewrite <- H4.
        algMagic.
        
        assert (q → == (o' : r')→) as Hp'.
        setoid_rewrite <- H5.
        algMagic.
       
        destruct (normForm o) as [oc [oi [pfoc [pfoi Ho]]]].
        destruct (normForm r0) as [r0c [r0i [pfr0c [pfr0i Hr0]]]].
        remember (projT1 (normOwnsJJ oc pfoc r0c pfr0c)) as pOwn.  
        remember (projT2 (normOwnsJJ oc pfoc r0c pfr0c)) as pOwn_pf.  
        assert (HpOwnEquiv: oc:r0c== pOwn).
          rewrite HeqpOwn.
          eapply normOwnsJJ_correct.
        assert (normJ pOwn).
          rewrite HeqpOwn.
          auto.
          
        destruct (normForm o') as [o'c [o'i [pfo'c [pfo'i Ho']]]].
        destruct (normForm r') as [r'c [r'i [pfr'c [pfr'i Hr']]]].
        remember (projT1 (normOwnsJJ o'c pfo'c r'c pfr'c)) as pOwn'.  
        remember (projT2 (normOwnsJJ o'c pfo'c r'c pfr'c)) as pOwn'_pf.  
        assert (HpOwn'Equiv: o'c:r'c== pOwn').
          rewrite HeqpOwn'.
          eapply normOwnsJJ_correct.
        assert (normJ pOwn').
          rewrite HeqpOwn'.
          auto.
 
        assert (p == oc : r0c).
          setoid_rewrite HpOwnEquiv.
          eapply normJConfEq. auto. auto.
          setoid_rewrite <- HpOwnEquiv.
          setoid_rewrite Hp.
          setoid_rewrite Ho.
          setoid_rewrite Hr0.
          algMagic.
          setoid_rewrite ownConfAssoc.
          setoid_rewrite ownIntegAssoc.
          algMagic.
          
        assert (q == o'c : r'c).
          setoid_rewrite HpOwn'Equiv.
          eapply normJConfEq. auto. auto.
          setoid_rewrite <- HpOwn'Equiv.
          setoid_rewrite Hp'.
          setoid_rewrite Ho'.
          setoid_rewrite Hr'.
          algMagic.
          setoid_rewrite ownConfAssoc.
          setoid_rewrite ownIntegAssoc.
          algMagic.
          
        setoid_rewrite Ho in H6.
        setoid_rewrite Ho' in H6.
            
        assert([c ⊢ oc ≽ o'c]
               /\ (exists n', [c ⊢ oc→ ≽ o'c→ # n'] /\ n' <= n1)
                  \/ oc == ⊤).
          pose proof staticBasisDeriv _ _ _ _ _ _ H6 as Hbasis.
          destruct Hbasis as [|isTop]. 
          
          (* there exists an x <= n1 *)
          do 3 destruct H22. 
          destruct H23. 
          destruct H24. 
          
          assert ([c ⊢ oc ≽ o'c]).
            eapply IHn with (j := x) (r := ⊥); eauto with zarith.
              apply (normJ_M ⊥ (normM_L ⊥ normL_bot)).  
              algMagic.
              assert (oc→ == o→).
                setoid_rewrite Ho.
                algMagic.
            left.
            split. auto. exists x. split. auto. auto.
          (* oc is top *)
           right.
            assert (o→ == oc→).
              setoid_rewrite Ho.
              algMagic.
            eapply normJConfEq; eauto. 
            apply (normJ_M ⊤ (normM_L ⊤ normL_top)).  
            setoid_rewrite <- isTop.
            algMagic.
             
        setoid_rewrite Hr0 in H7.
        setoid_rewrite Hr' in H7.
        
        assert([c ⊢ r0c ≽ r'c]
               /\ (exists n', [c ⊢ r0c→ ≽ r'c→ # n'] /\ n' <= n2)
                  \/ r0c == ⊤).
        
          pose proof staticBasisDeriv _ _ _ _ _ _ H7 as Hbasis.
          destruct Hbasis as [|isTop]. 
          
          (* there exists an x <= n1 *)
          do 3 destruct H23. 
          destruct H24. 
          destruct H25. 
          
          assert ([c ⊢ r0c ≽ r'c]).
            eapply IHn with (j := x) (r := ⊥); eauto with zarith.
              apply (normJ_M ⊥ (normM_L ⊥ normL_bot)).  
              algMagic.
              assert (r0c→ == r0→).
                setoid_rewrite Hr0.
                algMagic.
            left.
            split. auto. exists x. split. auto. auto.
          (* r0c is top *)
           right.
            eapply normJConfEq; eauto. 
            apply (normJ_M ⊤ (normM_L ⊤ normL_top)).  
            setoid_rewrite <- isTop.
            algMagic.
             
        setoid_rewrite H20.
        setoid_rewrite H21.
        destruct H22, H23.
          (* neither o nor r is top *)
          do 2 destruct H22.
          do 2 destruct H23.
          eexists.
          eapply AF_Own1; eauto using refl. 
          (* r0c is top *)
          do 2 destruct H22.
          setoid_rewrite H23.
          eexists.
          eapply AF_Own1; eauto using refl. 
          eautoflam.
          
          (* oc is top *)
          do 2 destruct H23.
          setoid_rewrite H22.
          eexists.
          eapply AF_Own1; eauto using refl. 
          eautoflam.
          
          (* both are top *)
          setoid_rewrite H22.
          setoid_rewrite H23.
          algMagic. eautoflam.
               
        (* AF_Own2 *)
        destruct (emptyUnion2 _ _ H8) as [HD1 HD2].
        rewrite HD1, HD2 in *.
        
        assert (p → == (o : r0)→) as Hp.
        setoid_rewrite <- H4.
        algMagic.
        
        assert (q → == (o' : r')→) as Hp'.
        setoid_rewrite <- H5.
        algMagic.
       
        destruct (normForm o) as [oc [oi [pfoc [pfoi Ho]]]].
        destruct (normForm r0) as [r0c [r0i [pfr0c [pfr0i Hr0]]]].
        remember (projT1 (normOwnsJJ oc pfoc r0c pfr0c)) as pOwn.  
        remember (projT2 (normOwnsJJ oc pfoc r0c pfr0c)) as pOwn_pf.  
        assert (HpOwnEquiv: oc:r0c== pOwn).
          rewrite HeqpOwn.
          eapply normOwnsJJ_correct.
        assert (normJ pOwn).
          rewrite HeqpOwn.
          auto.
          
        destruct (normForm o') as [o'c [o'i [pfo'c [pfo'i Ho']]]].
        destruct (normForm r') as [r'c [r'i [pfr'c [pfr'i Hr']]]].
        remember (projT1 (normOwnsJJ o'c pfo'c r'c pfr'c)) as pOwn'.  
        remember (projT2 (normOwnsJJ o'c pfo'c r'c pfr'c)) as pOwn'_pf.  
        assert (HpOwn'Equiv: o'c:r'c== pOwn').
          rewrite HeqpOwn'.
          eapply normOwnsJJ_correct.
        assert (normJ pOwn').
          rewrite HeqpOwn'.
          auto.
 
        assert (p == oc : r0c).
          setoid_rewrite HpOwnEquiv.
          eapply normJConfEq. auto. auto.
          setoid_rewrite <- HpOwnEquiv.
          setoid_rewrite Hp.
          setoid_rewrite Ho.
          setoid_rewrite Hr0.
          algMagic.
          setoid_rewrite ownConfAssoc.
          setoid_rewrite ownIntegAssoc.
          algMagic.
          
        assert (q == o'c : r'c).
          setoid_rewrite HpOwn'Equiv.
          eapply normJConfEq. auto. auto.
          setoid_rewrite <- HpOwn'Equiv.
          setoid_rewrite Hp'.
          setoid_rewrite Ho'.
          setoid_rewrite Hr'.
          algMagic.
          setoid_rewrite ownConfAssoc.
          setoid_rewrite ownIntegAssoc.
          algMagic.
          
        setoid_rewrite Ho in H6.
        setoid_rewrite Ho' in H6.
            
        assert([c ⊢ oc ≽ o'c]
               /\ (exists n', [c ⊢ oc→ ≽ o'c→ # n'] /\ n' <= n1)
                  \/ oc == ⊤).
          pose proof staticBasisDeriv _ _ _ _ _ _ H6 as Hbasis.
          destruct Hbasis as [|isTop]. 
          
          (* there exists an x <= n1 *)
          do 3 destruct H22. 
          destruct H23. 
          destruct H24. 
          
          assert ([c ⊢ oc ≽ o'c]).
            eapply IHn with (j := x) (r := ⊥); eauto with zarith.
              apply (normJ_M ⊥ (normM_L ⊥ normL_bot)).  
              algMagic.
              assert (oc→ == o→).
                setoid_rewrite Ho.
                algMagic.
            left.
            split. auto. exists x. split. auto. auto.
          (* oc is top *)
           right.
            assert (o→ == oc→).
              setoid_rewrite Ho.
              algMagic.
            eapply normJConfEq; eauto. 
            apply (normJ_M ⊤ (normM_L ⊤ normL_top)).  
            setoid_rewrite <- isTop.
            algMagic.
             
        setoid_rewrite Hr0 in H7.
        setoid_rewrite Ho' in H7.
        setoid_rewrite Hr' in H7.
        assert ((o'c : r'c)→ ∧ (o'i : r'i)←
                             == (o'c → ∧ o'i ←) : (r'c → ∧ r'i ←)).
          algMagic.
          setoid_rewrite ownIntegAssoc.
          setoid_rewrite ownConfAssoc.
          algMagic.
        setoid_rewrite <- H23 in H7.
        clear H23.
          
        assert([c ⊢ r0c ≽ o'c : r'c]
               /\ (exists n', [c ⊢ r0c→ ≽ o'c : r'c→ # n'] /\ n' <= n2)
                  \/ r0c == ⊤).
        
          pose proof staticBasisDeriv _ _ _ _ _ _ H7 as Hbasis.
          destruct Hbasis as [|isTop]. 
          
          (* there exists an x <= n1 *)
          do 3 destruct H23. 
          destruct H24. 
          destruct H25. 
          
          assert ([c ⊢ r0c ≽ o'c : r'c]).
            setoid_rewrite HpOwn'Equiv.
            eapply IHn with (j := x) (r := ⊥); eauto with zarith.
              apply (normJ_M ⊥ (normM_L ⊥ normL_bot)).  
              algMagic.
              assert (r0c→ == r0→).
                setoid_rewrite Hr0.
                algMagic.
              setoid_rewrite <- HpOwn'Equiv.
              auto.
            left.
            split. auto. exists x. split.
            setoid_rewrite ownConfAssoc.
            auto. auto.
          (* r0c is top *)
           right.
            eapply normJConfEq; eauto. 
            apply (normJ_M ⊤ (normM_L ⊤ normL_top)).  
            setoid_rewrite <- isTop.
            algMagic.
             
        setoid_rewrite H20.
        setoid_rewrite H21.
        destruct H22, H23.
          (* neither o nor r is top *)
          do 2 destruct H22.
          do 2 destruct H23.
          eexists.
          eapply AF_Own2; eauto using refl. 
          (* r0c is top *)
          do 2 destruct H22.
          setoid_rewrite H23.
          eexists.
          eapply AF_Own2; eauto using refl. 
          eautoflam.
          
          (* oc is top *)
          do 2 destruct H23.
          setoid_rewrite H22.
          eexists.
          eapply AF_Own2; eauto using refl. 
          eautoflam.
          
          (* both are top *)
          setoid_rewrite H22.
          setoid_rewrite H23.
          algMagic. eautoflam.
          
        (* AF_ConjL1 *)
        assert (p→ == (p1 ∧ p2)→).
        setoid_rewrite <- H4.
        algMagic.
        destruct (normForm p1) as [p1c [p1i [pfp1c [pfp1i Hp1']]]].
        destruct (normForm p2) as [p2c [p2i [pfp2c [pfp2i Hp2']]]].

        assert (normJ (p1c ∧ p2c)).
          apply (normJ_conj p1c p2c pfp1c pfp2c).
        
        assert (p == p1c ∧ p2c).
          eapply normJConfEq ; auto.
          setoid_rewrite H15.
          setoid_rewrite Hp1'.
          setoid_rewrite Hp2'.
          algMagic.
          
        assert ((p1 ∧ p2)← == r←).
          setoid_rewrite <- H4.
          algMagic.
          
        setoid_rewrite Hp1' in H5.  
        setoid_replace (q→) with (q→ ∧ ⊥←) in H5 by (algMagic).
        pose proof staticBasisDeriv _ _ _ _ _ _ H5.
        
        destruct H19. 
          (* case 1 : there exists an x <= n1 *)
          do 3 destruct H19. 
          destruct H20. 
          destruct H21. 
          
          assert ([c ⊢ p1c ≽ q]).
            eapply IHn with (j := x) (r := ⊥); eauto with zarith.
              apply (normJ_M ⊥ (normM_L ⊥ normL_bot)).  
            algMagic.
            
          destruct H23.
          eexists.
          eapply AF_Trans with (q := p1c).  
          setoid_rewrite H17.
          eautoflam.
          apply H23.
          eautoflam.
          eauto.
          (* case 2 : p1 is top *)
          rewrite H19 in *.
          assert(p1c == ⊤).
            eapply normJConfEq; auto.
            apply (normJ_M ⊤ (normM_L ⊤ normL_top)).  
            setoid_rewrite <- H19.
            algMagic.
          rewrite H20 in *.
          assert(p == ⊤).
            rewrite H17.
            algMagic.
          eautoflam.
        
        (* AF_ConjR *)
        destruct (emptyUnion2 _ _ H7) as [HD1 HD2].
        rewrite HD1, HD2 in *.
        
        assert ((q1 ∧ q2)← == ⊥).
        setoid_rewrite <- H4.
        algMagic.

        assert (q1← == ⊥ /\ q2← == ⊥).
        eapply conjIsBot; auto.
        setoid_rewrite <- conjInteg.
        auto.
        destruct H18.
        
        assert (q1 == q1→).
          setoid_rewrite <- (confIntegBasis q1).
          setoid_rewrite H18.
          algMagic.
          
        assert (q2 == q2→).
          setoid_rewrite <- (confIntegBasis q2).
          setoid_rewrite H19.
          algMagic.
        
        destruct (normForm q1) as [q1c [q1i [pfq1c [pfq1i Hq1']]]].
        rewrite Hq1' in *.

        assert ((q1c→ ∧ q1i←)→ == q1c→) as Hrewrite_q1 by algMagic.
        rewrite Hrewrite_q1 in *.
        
        destruct (normForm q2) as [q2c [q2i [pfq2c [pfq2i Hq2']]]].
        rewrite Hq2' in *.

        assert ((q2c→ ∧ q2i←)→ == q2c→) as Hrewrite_q2 by algMagic.
        rewrite Hrewrite_q2 in *.
        
        pose proof staticBasisDeriv _ _ _ _ _ _ H5.
        rewrite or_comm in H22.
        destruct H22 as [isTop|]. 
          assert (p == ⊤) as pTop.
            eapply normJConfEq; eauto.
              apply (normJ_M ⊤ (normM_L ⊤ normL_top)).  
            setoid_rewrite <- isTop.
            algMagic.
          setoid_rewrite pTop.
          eautoflam.
        
        (* there exists an x <= n1 *)
        do 3 destruct H22. 
        destruct H23. 
        destruct H24. 
         
        assert ([c ⊢ p ≽ q1c]).
          eapply IHn with (j := x) (r := ⊥); eauto with zarith.
            apply (normJ_M ⊥ (normM_L ⊥ normL_bot)).  
            algMagic.
            auto.
            
        pose proof staticBasisDeriv _ _ _ _ _ _ H6.
        rewrite or_comm in H27.
        destruct H27 as [isTop|]. 
          assert (p == ⊤) as pTop.
            eapply normJConfEq; eauto.
              apply (normJ_M ⊤ (normM_L ⊤ normL_top)).  
            setoid_rewrite <- isTop.
            algMagic.
          setoid_rewrite pTop.
          eautoflam.
 
        (* there exists an x1 <= n2 *)
        do 3 destruct H27. 
        destruct H28. 
        destruct H29. 
        assert ([c ⊢ p ≽ q2c]).
          eapply IHn with (j := x1) (r := ⊥); eauto with zarith.
            apply (normJ_M ⊥ (normM_L ⊥ normL_bot)).  
            algMagic.
            auto.

        assert (normJ (q1c ∧ q2c)). 
          eapply normJ_conj. auto. auto.

        assert (q == (q1c ∧ q2c)). 
          eapply normJConfEq. auto. auto.
          setoid_rewrite <- (confIdemp q).
          setoid_rewrite H4. 
          algMagic.
          
        setoid_rewrite H33.
        destruct H26.
        destruct H31.
        eexists.
        eapply AF_ConjR ; eauto using refl.
        
        (* AF_DisjL *)
        destruct (emptyUnion2 _ _ H7) as [HD1 HD2].
        rewrite HD1, HD2 in *.
        
        assert (p → == (p1 ∨ p2)→) as Hp.
        setoid_rewrite <- H4.
        algMagic.
       
        destruct (normForm p1) as [p1c [p1i [pfp1c [pfp1i Hp1]]]].
        destruct (normForm p2) as [p2c [p2i [pfp2c [pfp2i Hp2]]]].
        remember (projT1 (normDisjJJ p1c pfp1c p2c pfp2c)) as pDisj.  
        remember (projT2 (normDisjJJ p1c pfp1c p2c pfp2c)) as pDisj_pf.  
        assert (p1c ∨ p2c == pDisj).
          rewrite HeqpDisj.
          eapply normDisjJJ_correct.
        assert (normJ pDisj).
          rewrite HeqpDisj.
          auto.
          
        assert (p == p1c ∨ p2c ).
          setoid_rewrite H17.
          eapply normJConfEq. auto. auto.
          setoid_rewrite <- H17.
          setoid_rewrite Hp.
          setoid_rewrite Hp1.
          setoid_rewrite Hp2.
          algMagic.
          
        setoid_rewrite <- (confIntegBasis p1) in H5.
        setoid_rewrite <- (conjIdent q→) in H5.
        setoid_rewrite <- botInteg in H5.
            
        assert([c ⊢ p1c ≽ q]
               /\ (exists n', [c ⊢ p1→ ≽ q→ # n'] /\ n' <= n1) \/ p1c == ⊤).
          pose proof staticBasisDeriv _ _ _ _ _ _ H5 as Hbasis.
          destruct Hbasis as [|isTop]. 
          
          (* there exists an x <= n1 *)
          do 3 destruct H20. 
          destruct H21. 
          destruct H22. 
          
          assert ([c ⊢ p1c ≽ q]).
            eapply IHn with (j := x) (r := ⊥); eauto with zarith.
              apply (normJ_M ⊥ (normM_L ⊥ normL_bot)).  
              algMagic.
              assert (p1c→ == p1→).
                setoid_rewrite Hp1.
                algMagic.
              setoid_rewrite H24.
              auto.
            left.
            split. auto. exists x. split. auto. auto.
          (* p1 is top *)
           right.
            assert (p1→ == p1c→).
              setoid_rewrite Hp1.
              algMagic.
            eapply normJConfEq; eauto. 
            apply (normJ_M ⊤ (normM_L ⊤ normL_top)).  
            setoid_rewrite <- (confIntegBasis p1) in H20.
            setoid_rewrite isTop in H20.
            algMagic.
             
        setoid_rewrite <- (confIntegBasis p2) in H6.
        setoid_rewrite <- (conjIdent q→) in H6.
        setoid_rewrite <- botInteg in H6.
        assert([c ⊢ p2c ≽ q]
                /\ (exists n', [c ⊢ p2→ ≽ q→ # n'] /\ n' <= n2) \/ p2c == ⊤).
          pose proof staticBasisDeriv _ _ _ _ _ _ H6 as Hbasis.
          destruct Hbasis as [|isTop]. 
          
          (* there exists an x <= n1 *)
          do 3 destruct H21. 
          destruct H22. 
          destruct H23. 
          
          assert ([c ⊢ p2c ≽ q]).
            eapply IHn with (j := x) (r := ⊥); eauto with zarith.
              apply (normJ_M ⊥ (normM_L ⊥ normL_bot)).  
              algMagic.
              assert (p2c→ == p2→).
                setoid_rewrite Hp2.
                algMagic.
              setoid_rewrite H25.
              auto.
            left. 
            split. auto. exists x. auto.
          (* p2 is top *)
          right.
            assert (p2→ == p2c→).
              setoid_rewrite Hp2.
              algMagic.
            eapply normJConfEq; eauto. 
            apply (normJ_M ⊤ (normM_L ⊤ normL_top)).  
            setoid_rewrite <- (confIntegBasis p2) in H21.
            setoid_rewrite isTop in H21.
            algMagic.
          auto.
             
        setoid_rewrite H19.
        destruct H20, H21.
          (* neither p1 nor p2 is top *)
          do 2 destruct H20.
          do 2 destruct H21.
          eexists.
          eapply AF_DisjL; eauto using refl. 
          (* p2 is top *)
          destruct H20.
          setoid_rewrite H21. algMagic.
          
          (* p1 is top *)
          destruct H21.
          setoid_rewrite H20. algMagic.
          
          (* both are top *)
          setoid_rewrite H20.
          setoid_rewrite H21.
          algMagic. eautoflam.
        
        (* AF_DisjR1 *)
        assert ((q1 ∨ q2)← == ⊥).
        setoid_rewrite <- H4.
        algMagic.

        destruct (normForm q1) as [q1c [q1i [pfq1c [pfq1i Hq1']]]].
        rewrite Hq1' in *.

        assert ((q1c→ ∧ q1i←)→ == q1c→) as Hrewrite_q1 by algMagic.
        
        destruct (normForm q2) as [q2c [q2i [pfq2c [pfq2i Hq2']]]].
        rewrite Hq2' in *.

        assert ((q2c→ ∧ q2i←)→ == q2c→) as Hrewrite_q2 by algMagic.
        
        assert ((q1c → ∨ q2c → ∧ q2i ←)→ == (q1c ∨ q2c)→)
          as Hrewrite_q by algMagic.
        assert (q → == (q1c ∨ q2c)→).
        setoid_rewrite <- confIdemp.
        setoid_rewrite H4.
        algMagic.

        assert((exists n', [c ⊢ p→ ≽ q1c→ # n'] /\ n' <= n1) \/ p == ⊤).
          pose proof staticBasisDeriv _ _ _ _ _ _ H5 as Hbasis.
          destruct Hbasis.
            do 3 destruct H17.
            destruct H18.
            destruct H19.
            left.
            eexists.
            eautoflam.
            (* otherwise p' is top *)
            right.
            eapply normJConfEq; auto.
              apply (normJ_M ⊤ (normM_L ⊤ normL_top)).  
              setoid_rewrite <- H17.
              algMagic.
              
        destruct H17 as [|pIsTop].
          (* an x <= n1 exists *)
          do 2 destruct H17.
          
          assert ([c ⊢ p ≽ q1c]).
            eapply IHn with (j := x) (r := ⊥) ; eauto with zarith.
              apply (normJ_M ⊥ (normM_L ⊥ normL_bot)).  
            algMagic.
          
          remember (projT1 (normDisjJJ q1c pfq1c q2c pfq2c)) as pDisj.  
          remember (projT2 (normDisjJJ q1c pfq1c q2c pfq2c)) as pDisj_pf.  
          assert (q1c ∨ q2c == pDisj).
            rewrite HeqpDisj.
            eapply normDisjJJ_correct.
          assert (normJ pDisj).
            rewrite HeqpDisj.
            auto.

          assert (q == (q1c ∨ q2c)). 
            setoid_rewrite H20.
            eapply normJConfEq. auto. auto.
            setoid_rewrite <- confIdemp.
            setoid_rewrite H4.
            setoid_rewrite <- H20.
            algMagic.
            
          setoid_rewrite H22.
          destruct H19.
          eexists.
          eapply AF_DisjR1 ; eauto using refl.
          (* p is top *)        
          setoid_rewrite pIsTop.
          eautoflam.

        (* AF_Trans *)
        destruct (emptyUnion2 _ _ H6) as [HD1 HD2].
        rewrite HD1, HD2 in *.

        destruct (normForm q0) as [q0c [q0i [pfq0c [pfq0i Hq0']]]].
        rewrite Hq0' in *.

        assert ((q0c→ ∧ q0i←)→ == q0c→) as Hrewrite_q0 by algMagic.

        assert([c ⊢ p ≽ q0c]
               /\ (exists n', [c ⊢ p→ ≽ q0c→ # n']
                              /\ n' <= n1) \/ p == ⊤).
          pose proof staticBasisDeriv _ _ _ _ _ _ H4 as Hbasis.
          destruct Hbasis as [|isTop]. 
            (* an x <= n1 exists *)
            do 3 destruct H16.
            destruct H17.
            destruct H18.
        
            assert ([c ⊢ p ≽ q0c]).
              eapply IHn with (j := x) (r := ⊥); eauto with zarith.
                apply (normJ_M ⊥ (normM_L ⊥ normL_bot)).  
                algMagic.
            left.
            split. auto. exists x. split. auto. auto.
            (* p is top *)
            right.
            eapply normJConfEq; eauto.
              apply (normJ_M ⊤ (normM_L ⊤ normL_top)).  
            setoid_rewrite <- isTop.
            algMagic.
            
        setoid_rewrite <- (conjIdent q→) in H5.
        setoid_rewrite <- botInteg in H5.
         assert([c ⊢ q0c ≽ q]
               /\ (exists n', [c ⊢ q0c→ ≽ q→ # n']
                              /\ n' <= n2) \/ q0c == ⊤).
          pose proof staticBasisDeriv _ _ _ _ _ _ H5 as Hbasis.
          destruct Hbasis as [|isTop]. 
            (* an x <= n1 exists *)
            do 3 destruct H17.
            destruct H18.
            destruct H19.
        
            assert ([c ⊢ q0c ≽ q]).
              eapply IHn with (j := x) (r := ⊥); eauto with zarith.
                apply (normJ_M ⊥ (normM_L ⊥ normL_bot)).  
                algMagic.
            left.
            split. auto. exists x. split. auto. auto.
            (* q0c is top *)
            right.
            eapply normJConfEq; eauto.
              apply (normJ_M ⊤ (normM_L ⊤ normL_top)).  
            setoid_rewrite <- isTop.
            algMagic.

        destruct H16, H17.
          (* neither is top *)
          do 2 destruct H16, H17.
          eexists.
          eapply AF_Trans with (q := q0c) ; eauto.
          (* q0c is top *)
          do 2 destruct H16.
          eexists.
          setoid_rewrite H17 in H16.
          eapply AF_Trans with (q := ⊤) ; eauto.
          eautoflam.
          (* p is top *)
          setoid_rewrite H16.
          eautoflam.
          (* both are top *)
          setoid_rewrite H16.
          eautoflam.

        (* AF_Fwd *)
        destruct (emptyUnion2 _ _ H8) as [HD1 HtransferD2].
        pose proof transferIsEmpty _ _ HtransferD2 as HD2.
        rewrite HD1, HD2, joinTopConf, meetTopInteg in *.

        assert ([rn ⊢ p ≽ q]).
          eapply IHn with (j := n2); eauto with zarith.
        destruct H18.

        assert (exists n', [c ⊢ p ≽ q # n'] /\ n' <= x)
            by (eapply staticCoerce ; eauto with zarith).
        do 2 destruct H19.

        eauto.

        (* AF_Weaken *)
        destruct (PH.emptyUnion2 _ _ H9) as [HH1 HH2].
        destruct (emptyUnion3 _ _ _ H10) as [HD1 [HD2 HD3]].
        rewrite HH1, HD1, HD2, HD3, (joinTopConf l') in *.

        assert (exists n', [c ⊢ ⊤→ ⊑ pc' # n'] /\ n' < n2)
            by (eapply staticRobust ; eauto with zarith).
        assert (exists n', [c ⊢ l' ⊑ ⊤← # n'] /\ n' < n3)
            by (eapply staticRobust ; eauto with zarith).
        do 2 destruct H20, H21.

        assert (pc' == ⊤→) as Hpc by (eapply flowsFromTopConf ; eauto).
        assert (l' == ⊤←) as Hl by (eapply flowsToTopInteg ; eauto).
        rewrite Hpc, Hl in *.

        eapply IHn with (j := n1) ; eauto with zarith.

  intros.
  eapply H with (r := ⊥); eauto with zarith.
    apply (normJ_M ⊥ (normM_L ⊥ normL_bot)).  
  eautoflam.
Qed.
(* end hide *)

Lemma staticVoice:
  forall c p q, [c ⊢ p ≽ q] -> [c ⊢ ∇(p) ≽ ∇(q)].
Proof.
  intros.
  destruct H.
  destruct (normForm p) as [pc [pi [pfpc [pfpi Hp]]]].
  destruct (normForm q) as [qc [qi [pfqc [pfqi Hq]]]].
  rewrite Hp, Hq in *.
  assert ([c ⊢ ∇(pc→) ≽ ∇(qc→)]).
    assert ([c ⊢ pc→ ≽ qc→]) as Hconf.
      eexists.
      setoid_replace (pc→) with ((pc→ ∧ pi←)→) by algMagic.
      setoid_replace (qc→) with ((qc→ ∧ qi←)→) by algMagic.
      eautoflam.
    destruct Hconf.

    repeat rewrite voiceOfNormJConf ; auto.
    assert ([c ⊢ pc ≽ qc]).
      eapply normJConf ; eauto.
    destruct H1.

    eexists.
    eapply AF_Proj_integ ; eauto using refl.

  assert ([c ⊢ ∇(pi←) ≽ ∇(qi←)]).
    repeat rewrite voiceOfInteg.
    setoid_replace (pi←) with ((pc→ ∧ pi←)←) by algMagic.
    setoid_replace (qi←) with ((qc→ ∧ qi←)←) by algMagic.
    eexists.
    eapply AF_Proj_integ ; eauto using refl ; algMagic.
  destruct H0, H1.

  setoid_rewrite Hp.
  setoid_rewrite Hq.
  repeat setoid_rewrite voiceOfConj.
  eautoflam.
Qed.

(**
 *** Weakening
 *)
Definition weakensTo (H : PH.t) (c : name) (pc0 l0 pc l : Label) D n : Prop :=
  exists D1 n1 D2 n2,
    [H;c;pc ⊔ l0;l ⊩ pc ⊑ pc0 # D1;n1]
      /\ [H;c;pc ⊔ l0;l ⊩ l0 ⊑ l # D2;n2]
      /\ D =D= union D1 D2
      /\ n = n1 + n2.

Instance weakensToProper:
  Proper
    (eq ==> eq ==> prinEq ==> prinEq ==> prinEq ==> prinEq
      ==> eq ==> eq ==> iff)
    weakensTo.
Proof.
  assert (forall H c pc0A l0A pcA lA pc0B l0B pcB lB D n,
           pc0A == pc0B
        -> l0A == l0B
        -> pcA == pcB
        -> lA == lB
        -> weakensTo H c pc0A l0A pcA lA D n
        -> weakensTo H c pc0B l0B pcB lB D n)
      as Hhelp.
    intros H c pc0A l0A pcA lA pc0B l0B pcB lB D n Hpc0 Hl0 Hpc Hl ?.
    do 4 destruct H0.
    do 4 eexists.
    setoid_rewrite <- Hpc0.
    setoid_rewrite <- Hl0.
    setoid_rewrite <- Hpc.
    setoid_rewrite <- Hl.
    eauto.

  intros HA HB HH.
  intros cA cB Hc.
  intros pc0A pc0B Hpc0.
  intros l0A l0B Hl0.
  intros pcA pcB Hpc.
  intros lA lB Hl.
  intros DA DB HD.
  intros nA nB Hn.
  split ; intro.
    rewrite <- HH, <- Hc, <- HD, <- Hn.
    eapply Hhelp ; eauto.

    rewrite HH, Hc, HD, Hn.
    eapply Hhelp ; eauto ; autoalg.
Qed.

Lemma weaken:
  forall (H : PH.t) (c : name) (pc0 l0 : Label) (pc l : Label) D1 n1,
    weakensTo H c pc0 l0 pc l D1 n1
      -> forall (p q : Prin) D2 n2 D n,
          [H;c;pc0;l0 ⊢ p ≽ q # D2;n2]
            -> D =D= union D1 D2
            -> n = n1 + n2 + 1
            -> [H;c;pc;l ⊢ p ≽ q # D;n].
Proof.
  intros.
  unfold weakensTo in H0.
  do 5 destruct H0.
  destruct H4.
  destruct H5.
  eapply AF_Weaken with (H2 := PH.empty) ; eauto.
    rewrite PH.unionComm.
    symmetry.
    autoflam.

    rewrite <- H5.
    rewrite unionComm.
    auto.

    omega.
Qed.

Hint Resolve weaken : flam.

Lemma staticWeaken:
  forall H c pc l p q n1 n,
    [c ⊢ p ≽ q # n1]
      -> n = n1+7
      -> [H;c;pc;l ⊢ p ≽ q # empty;n].
Proof.
  intros.
  eapply AF_Weaken with (H2 := H).
    eapply H0.
    eautoflamz.

    autoalg.
    eautoflamz.

    symmetry. autoflam.

    repeat rewrite emptyZero. auto.

    omega.
Qed.

Hint Resolve staticWeaken : flam.

Lemma weakenD:
  forall H c pc l p q D' D,
    [H;c;pc;l ⊢ p ≽ q | D']
      -> included D' D
      -> [H;c;pc;l ⊢ p ≽ q | D].
Proof.
  intros.
  do 3 destruct H0.
  eexists.
  split ; eauto.
  apply inclusionTrans with (y := D') ; auto.
Qed.

Lemma weakenH:
  forall H' H c pc l p q D,
    [H';c;pc;l ⊢ p ≽ q | D]
      -> PH.included H' H
      -> [H;c;pc;l ⊢ p ≽ q | D].
Proof.
 intros.
  do 3 destruct H0.
  eexists.
  split.
    eexists.
    eapply AF_Weaken
        with (pc' := pc) (l' := l) (H2 := H) ; eauto.
      eautoflam.
      eautoflam.
    rewrite PH.unionComm.
    rewrite PH.unionAbsorbs.
    autoflam. auto.
    repeat rewrite unionEmptyEmpty.
    rewrite unionComm.
    rewrite emptyZero ; autoflam.
Qed.

Lemma robWeakenD:
  forall H c pc l p q D' D,
    [H;c;pc;l ⊩ p ≽ q | D']
      -> included D' D
      -> [H;c;pc;l ⊩ p ≽ q | D].
Proof.
  intros.
  do 3 destruct H0.
  eexists.
  split ; eauto.
  apply inclusionTrans with (y := D') ; auto.
Qed.

(**
 *** Robust judgements
 *)

Lemma robStatic:
  forall H c pc l p q D,
    [c ⊢ p ≽ q] -> [H;c;pc;l ⊩ p ≽ q | D].
Proof.
  intros.
  destruct H0.
  eautoflam.
Qed.

Hint Resolve robStatic : flam.

Lemma robLift:
  forall H c pc l p q D,
    [H;c;pc; l ∧ ∇(q) ⊢ p ≽ q | D]
      -> [H;c;pc;l ⊩ pc ≽ ∇(q) | D]
      -> [H;c;pc;l ⊩ ∇(p →) ≽ ∇(q →) | D]
      -> [H;c;pc;l ⊩ p ≽ q | D].
Proof.
  intros.
  do 3 destruct H0, H1, H2.
  eexists.
  split ; eautoflam.
Qed.

Hint Resolve robLift : flam.

Lemma robLiftPC:
  forall H c pc l q D,
    [H;c;pc; l ∧ ∇(q) ⊢ pc ≽ ∇(q) | D]
      -> [H;c;pc;l ⊩ pc ≽ ∇(q) | D].
Proof.
  intros.
  do 3 destruct H0.
  eautoflam.
Qed.

Hint Resolve robLiftPC : flam.

Lemma robConjL1:
  forall H c pc l p1 p2 p D,
    [H;c;pc;l ⊩ p1 ≽ p | D] -> [H;c;pc;l ⊩ p1 ∧ p2 ≽ p | D].
Proof.
  assert (forall n H c pc l p1 p2 p D,
      (exists D', [H;c;pc;l ⊩ p1 ≽ p # D';n] /\ included D' D)
        -> [H;c;pc;l ⊩ p1 ∧ p2 ≽ p | D]).
  {
    induction n as [|n IHn] using strong_induction ; intros ; do 2 destruct H0.
    { inversion H0 ; omega. }
    {
      inversion H0.
      { eexists. eautoflam. }
      {
        assert ([H;c;pc;l ⊩ ∇((p1 ∧ p2)→) ≽ ∇(p→) | D3]).
        {
          setoid_replace (∇((p1 ∧ p2)→)) with (∇(p1→) ∧ ∇(p2→)) by algMagic.
          eapply IHn with (j := n3) ; eautoflamz.
        }
        do 3 destruct H16.

        eexists.
        split.
        { eautoflam. }
        {
          apply inclusionTrans with (y := x) ; rewrite H6.
          {
            apply unionMinimal ; autoflam.
            apply unionMinimal ; autoflam.
            { apply inclusionTrans with (y := D3) ; autoflam. }
          }
          {
            apply inclusionTrans with (y := x) ; autoflam.
            rewrite H6.
            autoflam.
          }
        }
      }
      {
        setoid_rewrite H3.
        setoid_rewrite H4.
        eexists.
        split.
        {
          eexists.
          eapply RAF_Lift.
          {
            rewrite voiceIdemp.
            eapply AF_Trans with (q := pc) ; eautoflam.
          }
          {
            setoid_rewrite voiceIdemp.
            eautoflam.
          }
          {
            destruct (voiceIsInteg q') as [y Hvoice].
            rewrite Hvoice.
            algMagic.
            eautoflam.
          }
          {
            repeat rewrite unionEmptyEmpty.
            rewrite (unionComm x _) , unionIdemp, emptyZero ; autoflam.
          }
          { auto. }
        }
        { apply inclusionTrans with (y := x) ; autoflam. }
      }
      {
        assert ([H;c;pc;l ⊩ p1 ∧ p2 ≽ q1 | D1]).
        { eapply IHn with (j := n1) ; eautoflamz. }
        do 3 destruct H16.

        assert ([H;c;pc;l ⊩ p1 ∧ p2 ≽ q2 | D2]).
        { eapply IHn with (j := n2) ; eautoflamz. }
        do 3 destruct H18.

        eexists.
        split.
        { eautoflam. }
        {
          apply inclusionTrans with (y := x) ; auto.
          rewrite H6.
          apply unionMinimal.
          { apply inclusionTrans with (y := D1) ; autoflam. }
          { apply inclusionTrans with (y := D2) ; autoflam. }
        }
      }
      {
        setoid_rewrite H3.
        assert ([H;c;pc;l ⊩ p3 ∧ p2 ≽ p | D1]).
        { eapply IHn with (j := n1) ; eautoflamz. }
        do 3 destruct H16.

        assert ([H;c;pc;l ⊩ p4 ∧ p2 ≽ p | D2]).
        { eapply IHn with (j := n2) ; eautoflamz. }
        do 3 destruct H18.

        eexists. split.
        { setoid_rewrite conjRDistDisj. eautoflam. }
        {
          apply inclusionTrans with (y := x) ; auto.
          rewrite H6.
          eautoflam.
        }
      }
      {
        assert ([H3;c;pc';l' ⊩ p1 ∧ p2 ≽ p | D1]).
        { eapply IHn with (j := n1) ; eautoflamz. }
        do 3 destruct H19.

        eexists. split.
        {
          eexists.
          eapply RAF_Weaken ; eauto.
        }
        {
          apply inclusionTrans with (y := x) ; auto.
          rewrite H9.
          apply unionMinimal ; autoflam.
        }
      }
      {
        assert ([H;c;pc;l ⊩ p1 ∧ p2 ≽ q | D1]).
        { eapply IHn with (j := n1) ; eautoflamz. }
        do 3 destruct H16.

        eexists.
        split.
        { eexists. eapply RAF_Trans with (q := q) ; eauto. }
        {
          apply inclusionTrans with (y := x) ; auto.
          rewrite H6.
          apply unionMinimal ; autoflam.
        }
      }
      {
        assert ([H;rn;pc ⊔ l ⊔ c0←;l ⊓ c0→ ⊩ p1 ∧ p2 ≽ p | D2]).
        { eapply IHn with (j := n2) ; eautoflamz. }
        do 3 destruct H17.

        eexists.
        split.
        {
          eexists.
          eapply RAF_Fwd with (rn := rn) ; eauto.
        }
        {
          apply inclusionTrans with (y := x) ; auto.
          rewrite H7.
          apply unionMinimal ; autoflam.
          apply inclusionTrans with (y := transfer D2 rn);
              autoflam.
        }
      }
    }
  }

  intros.
  do 3 destruct H1.
  eauto.
Qed.

Hint Resolve robConjL1 : flam.

Lemma robConjL2:
  forall H c pc l p1 p2 p D,
    [H;c;pc;l ⊩ p2 ≽ p | D] -> [H;c;pc;l ⊩ p1 ∧ p2 ≽ p | D].
Proof.
  intros.
  setoid_rewrite conjComm.
  eautoflam.
Qed.

Hint Resolve robConjL2 : flam.

Lemma robConjR:
  forall H c pc l p p1 p2 D,
    [H;c;pc;l ⊩ p ≽ p1 | D]
      -> [H;c;pc;l ⊩ p ≽ p2 | D]
      -> [H;c;pc;l ⊩ p ≽ p1 ∧ p2 | D].
Proof.
  intros.
  do 3 destruct H0, H1.
  eexists.
  eautoflam.
Qed.

Hint Resolve robConjR : flam.

Lemma robDisjL:
  forall H c pc l p1 p2 p D,
    [H;c;pc;l ⊩ p1 ≽ p | D]
      -> [H;c;pc;l ⊩ p2 ≽ p | D]
      -> [H;c;pc;l ⊩ p1 ∨ p2 ≽ p | D].
Proof.
  intros.
  do 3 destruct H0, H1.
  eexists.
  eautoflam.
Qed.

Hint Resolve robDisjL : flam.

Lemma robStaticWeaken:
  forall (H : PH.t) (c : name) (pc0 l0 pc l : Label) p q n1 n2 D,
    [c ⊢ pc ⊑ pc0 # n1]
      -> [c ⊢ l0 ⊑ l # n2]
      -> [H;c;pc0;l0 ⊩ p ≽ q | D]
      -> [H;c;pc;l ⊩ p ≽ q | D].
Proof.
  intros.
  do 3 destruct H2.
  eexists.
  split.
    eexists.
    eapply RAF_Weaken with (H2 := H) ; eautoflam.
    apply PH.unionIdemp.
    autoflam.
Qed.

Hint Resolve robStaticWeaken : flam.

Lemma staticWeakenContext:
  forall (H : PH.t) (c : name) (pc0 l0 pc l : Label) p q n1 n2 D,
    [c ⊢ pc ⊑ pc0 # n1]
      -> [c ⊢ l0 ⊑ l # n2]
      -> [H;c;pc0;l0 ⊢ p ≽ q | D]
      -> [H;c;pc;l ⊢ p ≽ q | D].
Proof.
  intros.
  do 3 destruct H2.
  eexists.
  split.
    eexists.
    eapply AF_Weaken with (H2 := H) ; eautoflam.
    apply PH.unionIdemp.
    autoflam.
Qed.

Lemma robWeakenH:
  forall H H1 H2 c pc l p q D,
    [H1;c;pc;l ⊩ p ≽ q | D]
      -> H =H= PH.union H1 H2
      -> [H;c;pc;l ⊩ p ≽ q | D].
Proof.
  intros.
  do 3 destruct H0.
  eexists.
  split.
    eexists.
    eapply RAF_Weaken
        with (pc' := pc) (l' := l) (H1 := H1)
          (D1 := x) (D2 := empty) (D3 := empty) ; eauto.
      eautoflam.
      eautoflam.
      rewrite emptyZero, unionComm ; autoflam.
Qed.

Definition robWeakensTo (H : PH.t) (c : name) (pc0 l0 pc l : Label) D :=
  [H;c;pc ⊔ l0;l ⊩ pc ⊑ pc0 | D]
    /\ [H;c;pc ⊔ l0;l ⊩ l0 ⊑ l | D].

Instance robWeakensToProper:
  Proper
    (eq ==> eq ==> prinEq ==> prinEq ==> prinEq ==> prinEq
      ==> eq ==> iff)
    robWeakensTo.
Proof.
  assert (forall H c pc0A l0A pcA lA pc0B l0B pcB lB D,
           pc0A == pc0B
        -> l0A == l0B
        -> pcA == pcB
        -> lA == lB
        -> robWeakensTo H c pc0A l0A pcA lA D
        -> robWeakensTo H c pc0B l0B pcB lB D)
      as Hhelp.
    intros H c pc0A l0A pcA lA pc0B l0B pcB lB D Hpc0 Hl0 Hpc Hl ?.
    destruct H0.
    split;
        try (setoid_rewrite <- Hpc0);
        setoid_rewrite <- Hl0;
        setoid_rewrite <- Hpc;
        setoid_rewrite <- Hl;
        eauto.

  intros HA HB HH.
  intros cA cB Hc.
  intros pc0A pc0B Hpc0.
  intros l0A l0B Hl0.
  intros pcA pcB Hpc.
  intros lA lB Hl.
  intros DA DB HD.
  split ; intro.
    rewrite <- HH, <- Hc, <- HD.
    eapply Hhelp ; eauto.

    rewrite HH, Hc, HD.
    eapply Hhelp ; eauto ; autoalg.
Qed.

Lemma robWeaken:
  forall (H : PH.t) (c : name) (pc0 l0 pc l : Label) (q : Prin) D,
    robWeakensTo H c pc0 l0 pc l D
      -> forall (p : Prin),
          [H;c;pc0;l0 ⊩ p ≽ q | D]
            -> [H;c;pc;l ⊩ p ≽ q | D].
Proof.
  intros.
  unfold robWeakensTo in H0.
  do 4 destruct H0.
  do 3 destruct H2.
  do 3 destruct H1.
  eexists.
  split.
    eexists.
    eapply RAF_Weaken with (H2 := PH.empty) (pc' := pc0) (l' := l0) ; eauto.
    rewrite PH.unionComm.
    rewrite PH.emptyZero.
    autoflam.

    autoflam.
Qed.

Lemma weaken2:
  forall (H : PH.t) (c : name) (pc0 l0 pc l : Label) (q : Prin) D,
    robWeakensTo H c pc0 l0 pc l D
      -> forall (p : Prin),
          [H;c;pc0;l0  ⊢ p ≽ q | D]
            -> [H;c;pc;l ⊢  p ≽ q | D].
Proof.
  intros.
  unfold robWeakensTo in H0.
  do 4 destruct H0.
  do 3 destruct H1, H2.
  eexists.
  split.
    eexists.
    eapply AF_Weaken with (H2 := PH.empty) (pc' := pc0) (l' := l0) ; eauto.
    rewrite PH.unionComm.
    symmetry.
    apply PH.emptyZero.

    autoflam.
Qed.

(* begin hide *)
Module robTransHelpers.
  Lemma robTransRule:
    forall H c pc l p q r D,
      [H;c;pc;l ⊩ p ≽ q | D]
        -> [H;c;pc;l ⊩ q ≽ r | D]
        -> [H;c;pc;l ⊩ pc ≽ ∇(r→) | D]
        -> [H;c;pc;l ⊩ p ≽ r | D].
  Proof.
    intros.
    do 3 destruct H0, H1, H2.
    eexists.
    split.
      eexists. eapply RAF_Trans with (q := q) ; eauto.
      autoflam.
  Qed.

  Lemma robTransStatic1:
    forall H c pc l p q r D,
           [c ⊢ p ≽ q]
        -> [H;c;pc;l ⊩ q ≽ r | D]
        -> [H;c;pc;l ⊩ p ≽ r | D].
  Proof.
    assert (forall n H c pc l p q r D,
             [c ⊢ p ≽ q]
          -> (exists D', [H;c;pc;l ⊩ q ≽ r # D';n] /\ included D' D)
          -> [H;c;pc;l ⊩ p ≽ r | D]).
    {
      induction n as [|n IHn] using strong_induction ; intros.
      {
        (* Base case. *)
        do 2 destruct H1.
        inversion H1 ; omega.
      }
      {
        (* Inductive case. *)
        destruct H0.
        do 2 destruct H1.
        inversion H1.
        {
          (* q ≽ r derived via RAF_Static. *)
          eexists.
          split.
          {
            eexists.
            eapply RAF_Static.
            { eapply AF_Trans with (q := q) ; eautoflam. }
            { autoflam. }
            { auto. }
          }
          { autoflam. }
        }
        {
          (* q ≽ r derived via RAF_Lift. *)
          assert ([H;c;pc;l ⊩ ∇(p→) ≽ ∇(r→) | D3]) as Hvoice.
          {
            eapply IHn with (j := n3) (q := ∇(q→)) ; eauto with zarith.
            { eapply staticVoice. eautoflam. }
            { eautoflam. }
          }
          destruct Hvoice as [? [[? ?] ?]].

          eexists.
          split.
          {
            eexists.
            eapply RAF_Lift.
            { eapply AF_Trans with (q := q) ; eautoflam. }
            { eauto. }
            { eauto. }
            { eautoflam. }
            { auto. }
          }
          {
            apply inclusionTrans with (y := x0) ; auto.
            rewrite H7.
            apply unionMinimal ; autoflam.
            apply unionMinimal ; autoflam.
            apply inclusionTrans with (y := D3) ; autoflam.
          }
        }
        {
          (* q ≽ r derived via RAF_LiftPC. *)
          assert ([H;c;q;l ⊩ q ≽ ∇(q') | D]) as Hweakened.
          {
            eapply robStaticWeaken with (pc0 := q) (l0 := l ∧ ∇(q')).
            { eautoflam. }
            { algMagic. eautoflam. }
            {
              eexists.
              split.
              {
                eexists.
                eapply RAF_LiftPC.
                { autoalg. }
                { eapply refl. }
                {
                  setoid_replace ((l ∧ ∇(q')) ∧ ∇(q')) with (l ∧ ∇(q'))
                      by algMagic.
                  setoid_rewrite H4.
                  eauto.
                }
                { auto. }
              }
              { eapply inclusionTrans ; eautoflam. }
            }
          }
          destruct Hweakened as [? [[? ?] ?]].

          rewrite H4, H5 in *.
          setoid_rewrite H5.
          eexists.
          split.
          {
            eexists.
            eapply RAF_Lift ; algMagic.
            { eapply AF_Trans with (q := pc) ; eautoflam. }
            { eauto. }
            { eautoflam. }
          }
          { autoflam. }
        }
        {
          (* q ≽ r derived via RAF_ConjR. *)
          assert ([H;c;pc;l ⊩ p ≽ q1 | D1]) as Hq1.
          { eapply IHn with (j := n1) (q := q) ; eautoflamz. }

          assert ([H;c;pc;l ⊩ p ≽ q2 | D2]) as Hq2.
          { eapply IHn with (j := n2) (q := q) ; eautoflamz. }

          destruct Hq1 as [? [[? ?] ?]].
          destruct Hq2 as [? [[? ?] ?]].

          eexists.
          split.
          { eautoflam. }
          {
            apply inclusionTrans with (y := x0) ; auto.
            rewrite H7.
            apply unionMinimal ; eapply inclusionTrans ; eautoflam.
          }
        }
        {
          (* q ≽ r derived via RAF_DisjL. *)
          assert (exists d1 d2, p == d1 ∨ d2 /\ [c ⊢ d1 ≽ p1] /\ [c ⊢ d2 ≽ p2])
              as Hdecomp.
          {
            eapply staticActsForDisj.
            setoid_rewrite <- H4.
            eautoflam.
          }
          destruct Hdecomp as [d1 [d2 [Hp [Hd1 Hd2]]]].

          setoid_rewrite Hp.
          eapply robDisjL.
          {
            eapply IHn with (j := n1) (q := p1) ; eauto with zarith.
            eexists.
            split ; eauto.
            apply inclusionTrans with (y := x0) ; auto.
            rewrite H7.
            autoflam.
          }
          {
            eapply IHn with (j := n2) (q := p2) ; eauto with zarith.
            eexists.
            split ; eauto.
            apply inclusionTrans with (y := x0) ; auto.
            rewrite H7.
            autoflam.
          }
        }
        {
          (* q ≽ r derived via RAF_Weaken. *)
          assert ([H4;c;pc';l' ⊩ p ≽ r | D1]) as Hstronger.
          { eapply IHn with (j := n1) (q := q) ; eautoflamz. }
          destruct Hstronger as [? [[? ?] ?]].

          eexists.
          split.
          { eexists. eapply RAF_Weaken with (H1 := H4) (H2 := H5) ; eautoflam. }
          {
            apply inclusionTrans with (y := x0) ; auto.
            rewrite H10.
            apply unionMinimal ; autoflam.
          }
        }
        {
          (* q ≽ r derived via RAF_Trans. *)
          assert ([H;c;pc;l ⊩ p ≽ q0 | D1]) as Hq0.
          {
            eapply IHn with (j := n1) (q := q) ; eauto with zarith.
            eexists.
            split ; eautoflam.
          }
          destruct Hq0 as [? [[? ?] ?]].

          eexists.
          split.
          {
            eexists.
            eapply RAF_Trans with (q := q0) ; eautoflam.
          }
          {
            apply inclusionTrans with (y := x0) ; auto.
            rewrite H7.
            apply unionMinimal ; autoflam.
          }
        }
        {
          (* q ≽ r derived via RAF_Fwd. *)
          assert ([H;rn;pc ⊔ l ⊔ c0←;l ⊓ c0→ ⊩ p ≽ r | D2]) as Hremote.
          {
            eapply IHn with (j := n2) (q := q) ; eautoflamz.
            assert (exists n, [rn ⊢ p ≽ q # n] /\ n <= x) as Hremote.
            { eapply staticCoerce ; eauto. }
            destruct Hremote as [? [? ?]].
            eauto.
          }
          destruct Hremote as [? [[? ?] ?]].

          eexists.
          split.
          { eexists. eapply RAF_Fwd with (rn := rn) ; eauto with zarith. }
          {
            apply inclusionTrans with (y := x0) ; auto.
            rewrite H8.
            apply unionMinimal ; autoflam.
            apply inclusionTrans with (y := transfer D2 rn) ; autoflam.
          }
        }
      }
    }

    intros.
    do 3 destruct H2.
    eauto.
  Qed.

  Lemma robTransStatic2:
    forall H c pc l p q r D,
           [H;c;pc;l ⊩ p ≽ q | D]
        -> [c ⊢ q ≽ r]
        -> [H;c;pc;l ⊩ p ≽ r | D].
  Proof.
    assert (forall n H c pc l p q r D,
             (exists D', [H;c;pc;l ⊩ p ≽ q # D';n] /\ included D' D)
          -> [c ⊢ q ≽ r]
          -> [H;c;pc;l ⊩ p ≽ r | D]).
      induction n as [|n IHn] using strong_induction ; intros.
        (* Base case. *)
        do 2 destruct H0.
        inversion H0 ; omega.

        (* Inductive case. *)
        assert ([c ⊢ ∇(q) ≽ ∇(r)]) as Hvoice.
          eapply staticVoice. eauto.
        destruct Hvoice.

        assert (exists n,
            weakensTo H c pc (l ∧ ∇(q)) pc (l ∧ ∇(r)) empty n) as Hweakens.
          do 5 eexists.
          split ; [|split].
            eautoflam.
            eapply RAF_Static. algMagic. eautoflam.
              autoflam.
              auto.
            eautoflam.
        destruct Hweakens as [? ?].

        destruct H1.
        do 2 destruct H0.
        inversion H0.
          (* p ≽ q derived via RAF_Static. *)
          eexists.
          split.
            eexists.
            eapply RAF_Static.
            eapply AF_Trans with (q := q) ; eautoflam.
            autoflam.
            auto.

            autoflam.

          (* p ≽ q derived via RAF_Lift. *)
          assert ([H;c;pc;l ⊩ pc ≽ ∇(r) | D2]) as Hpc.
            eapply IHn with (j := n2) (q := ∇(q)) ; eautoflamz.
          destruct Hpc as [? [[? ?] ?]].

          assert ([c ⊢ ∇(q→) ≽ ∇(r→)]) as Hvoice.
            eapply staticVoice ; eautoflam.
          destruct Hvoice.

          assert ([H;c;pc;l ⊩ ∇(p→) ≽ ∇(r→) | D3]) as Hvoice.
            eapply IHn with (j := n3) (q := ∇(q→)) ; eautoflamz.
          destruct Hvoice as [? [[? ?] ?]].

          eexists.
          split.
            eexists.
            eapply RAF_Lift ; eautoflam.
            eapply AF_Trans with (q := q) ; eautoflam.

            apply inclusionTrans with (y := x2) ; auto.
            rewrite H9.
            apply unionMinimal ; autoflam.
              apply unionMinimal ; autoflam.
              apply inclusionTrans with (y := D2) ; autoflam.
              apply inclusionTrans with (y := D3) ; autoflam.

          (* p ≽ q derived via RAF_LiftPC. *)
          pose proof voiceIsInteg q' as [vq' Hvq'].

          assert (r == r←) as Hr.
            rewrite Hvq', H7 in *.
            eapply staticActsForInteg ; eauto.
          rewrite Hr in *.

          assert (exists n,
              weakensTo H c p (l ∧ ∇(q')) p (l ∧ ∇(r←)) empty n) as Hweakens.
            do 5 eexists.
            split ; [|split].
              eautoflam.

              eapply RAF_Static.
                algMagic.
                rewrite <- H7 in *.
                eautoflam.

                autoflam.
                auto.

              autoflam.
          destruct Hweakens.

          eexists.
          split.
            eexists.
            rewrite H6, Hr.
            rewrite <- voiceOfInteg.
            eapply RAF_LiftPC ; try (eapply refl) ; auto.
            eapply AF_Trans with (q := ∇(q')).
              eapply weaken ; eauto.
              rewrite H7. algMagic. eauto.
              autoflam.

              rewrite H7 in H2.
              rewrite <- (voiceIdemp q').
              eautoflam.

              autoflam. auto.

            eautoflam.

          (* p ≽ q derived via RAF_ConjR. *)
          assert (exists r1 r2, r == r1 ∧ r2 /\ [c ⊢ q1 ≽ r1] /\ [c ⊢ q2 ≽ r2])
              as Hconj.
            eapply staticConjActsFor.
            setoid_rewrite <- H6.
            eauto.
          destruct Hconj as [? [? [? [? ?]]]].

          setoid_rewrite H19.
          eapply robConjR.
            eapply IHn with (j := n1) ; eauto with zarith.
            eexists.
            split.
              eauto.

              apply inclusionTrans with (y := x2) ; eauto.
              rewrite H9.
              autoflam.

            eapply IHn with (j := n2) ; eauto with zarith.
            eexists.
            split.
              eauto.
              apply inclusionTrans with (y := x2) ; eauto.
              rewrite H9.
              autoflam.

          (* p ≽ q derived via RAF_DisjL. *)
          assert ([H;c;pc;l ⊩ p1 ≽ r | D1]) as Hp1.
            eapply IHn with (j := n1) ; eautoflamz.
          destruct Hp1 as [? [[? ?] ?]].

          assert ([H;c;pc;l ⊩ p2 ≽ r | D2]) as Hp2.
            eapply IHn with (j := n2) ; eautoflamz.
          destruct Hp2 as [? [[? ?] ?]].

          eexists.
          split.
            eautoflam.

            apply inclusionTrans with (y := x2) ; eauto.
            rewrite H9.
            apply unionMinimal.
              apply inclusionTrans with (y := D1) ; autoflam.
              apply inclusionTrans with (y := D2) ; autoflam.

          (* p ≽ q derived via RAF_Weaken. *)
          assert (robWeakensTo H6 c pc' l' pc l D).
            split.
              eexists.
              split.
                eauto.
                apply inclusionTrans with (y := x2) ; auto.
                rewrite H12.
                autoflam.

              eexists.
              split.
                eauto.
                apply inclusionTrans with (y := x2) ; auto.
                rewrite H12.
                autoflam.

          eapply robWeakenH with (H1 := H6) (H2 := H7) ; autoflam.
          eapply robWeaken with (pc0 := pc') (l0 := l').
            auto.
            eapply IHn with (j := n1) ; eauto with zarith.
            eexists.
            split.
              eauto.
              apply inclusionTrans with (y := x2) ; auto.
              rewrite H12.
              autoflam.

          (* p ≽ q derived via RAF_Trans. *)
          eapply robTransRule with (q := q0).
            eexists.
            split.
              eauto.
              apply inclusionTrans with (y := x2) ; auto.
              rewrite H9.
              autoflam.

            assert ([H;c;pc;l ⊩ q0 ≽ r | D2]) as Hr
                by (eapply IHn with (j := n2) ; eautoflamz).
            destruct Hr as [? [[? ?] ?]].

            eexists.
            split.
              eauto.
              apply inclusionTrans with (y := x2) ; auto.
              rewrite H9.
              apply inclusionTrans with (y := D2) ; autoflam.

            assert ([c ⊢ q→ ≽ r→]) by eautoflam.
            assert ([c ⊢ ∇(q→) ≽ ∇(r→)]) by (eapply staticVoice ; eauto).
            assert ([H;c;pc;l ⊩ pc ≽ ∇(r→) | D3]) as Hpc.
                eapply IHn with (j := n3) ; eautoflamz.
            destruct Hpc as [? [[? ?] ?]].

            eexists.
            split.
              eauto.
              apply inclusionTrans with (y := x2) ; auto.
              rewrite H9.
              apply inclusionTrans with (y := D3) ; autoflam.

          (* p ≽ q derived via RAF_Fwd. *)
          assert ([H;c;pc;l ⊩ r0 ≽ (pc→ ∧ l) ∧ ∇(r) | D1]) as Hr0.
            eapply IHn with (j := n1) ; eautoflamz.
          destruct Hr0 as [? [[? ?] ?]].

          assert ([H;rn;pc ⊔ l ⊔ c0←;l ⊓ c0→ ⊩ p ≽ r | D2]) as Hp.
            eapply IHn with (j := n2) ; eautoflamz.
            assert (exists n, [rn ⊢ q ≽ r # n] /\ n <= x1) as Hq.
              eapply staticCoerce ; eauto.
            destruct Hq as [? [? ?]].
            eauto.
          destruct Hp as [? [[? ?] ?]].

          eexists.
          split.
            eexists.
            eapply RAF_Fwd with (rn := rn) ; eautoflam.

            apply inclusionTrans with (y := x2) ; auto.
            rewrite H10.
            apply unionMinimal.
              apply inclusionTrans with (y := D1) ; autoflam.
              apply inclusionTrans
                    with (y := transfer D2 rn);
                  autoflam.

    intros.
      do 3 destruct H1.
      eauto.
  Qed.

  Lemma robProj_integ:
    forall H c pc l p q D,
      [H;c;pc;l ⊩ p ≽ q | D]
        -> [H;c;pc;l ⊩ p← ≽ q← | D].
  Proof.
    assert (forall n H c pc l p q D,
      (exists D', [H;c;pc;l ⊩ p ≽ q # D';n] /\ included D' D)
        -> [H;c;pc;l ⊩ p← ≽ q← | D]).
      intro.
      induction n as [|n IHn] using strong_induction;
          intros;
          do 2 destruct H0.
        inversion H0 ; omega.

        assert (exists n, weakensTo H c
            pc (l ∧ ∇(q)) pc (l ∧ ∇(q←)) empty n).
          do 5 eexists.
          split ; [|split].
            eautoflam.
            eapply RAF_Static.
              algMagic.
              rewrite <- (confIntegBasis q) at 1.
              rewrite voiceOfConj.
              algMagic.
              rewrite conjPermute.
              eautoflam.
              autoflam.
              auto.

            eautoflam.
        destruct H2.

        inversion H0.
          eexists.
          split ; eautoflam.

          assert ([H; c; pc; l ⊩ pc ≽ ∇(q←) | D2]).
            algMagic.
            eapply robTransStatic2 with (q := ∇(q)) ; eautoflam.
          do 3 destruct H17.

          eexists.
          split.
            eexists.
            eapply RAF_Lift.
              eautoflam.
              eauto.
              algMagic. eautoflam.
              autoflam.
              auto.

            apply inclusionTrans with (y := x) ; auto.
            rewrite H7.
            apply unionMinimal ; autoflam.
            apply unionMinimal ; autoflam.
            apply inclusionTrans with (y := D2) ; autoflam.

          setoid_rewrite H4.
          setoid_rewrite H5.
          eexists.
          split.
            eexists.
            eapply RAF_Lift.
              eapply AF_Proj_integ ; try (eapply refl).
                algMagic.
                eautoflam.
                auto.

              algMagic. eautoflam.
              algMagic. eautoflam.
              autoflam.
              auto.

            autoflam.

          assert ([H;c;pc;l ⊩ p← ≽ q1← | D1]).
            eapply IHn with (j := n1) ; eautoflamz.
          do 3 destruct H17.

          assert ([H;c;pc;l ⊩ p← ≽ q2← | D2]).
            eapply IHn with (j := n2) ; eautoflamz.
          do 3 destruct H19.

          setoid_rewrite H4.
          eexists.
          split.
            eexists.
            algMagic.
            eautoflam.

            apply inclusionTrans with (y := x) ; auto.
            rewrite H7.
            apply unionMinimal;
                eapply inclusionTrans;
                eautoflam.

          assert ([H;c;pc;l ⊩ p1← ≽ q← | D1]).
            eapply IHn with (j := n1) ; eautoflamz.
          do 3 destruct H17.

          assert ([H;c;pc;l ⊩ p2← ≽ q← | D2]).
            eapply IHn with (j := n2) ; eautoflamz.
          do 3 destruct H19.

          setoid_rewrite H4.
          eexists.
          split.
            eexists.
            algMagic.
            eapply RAF_DisjL ; eautoflam.

            apply inclusionTrans with (y := x) ; auto.
            rewrite H7.
            apply unionMinimal;
                eapply inclusionTrans;
                eautoflam.

          assert ([H4;c;pc';l' ⊩ p← ≽ q← | D1]).
            eapply IHn with (j := n1) ; eauto with zarith.
            eexists.
            split ; eautoflam.
          do 3 destruct H20.

          eexists.
          split.
            eexists.
            eapply RAF_Weaken with (H1 := H4) (H2 := H5) ; eautoflam.

            apply inclusionTrans with (y := x) ; auto.
            rewrite H10.
            apply unionMinimal ; autoflam.

          assert ([H;c;pc;l ⊩ p← ≽ q0← | D]).
            eapply IHn with (j := n1) ; eauto with zarith.
            eexists.
            split ; eauto.
            apply inclusionTrans with (y := x) ; auto.
            rewrite H7.
            autoflam.

          assert ([H;c;pc;l ⊩ q0← ≽ q← | D]).
            eapply IHn with (j := n2) ; eauto with zarith.
            eexists.
            split ; eauto.
            apply inclusionTrans with (y := x) ; auto.
            rewrite H7.
            autoflam.
          do 3 destruct H17, H18.

          eexists.
          split.
            eexists. eapply RAF_Trans ; eauto.
            algMagic. eautoflam.
            autoflam.

            autoflam.

          assert ([H;c;pc;l ⊩ r ≽ (pc→ ∧ l) ∧ ∇(q←) | D1]).
            eapply robTransStatic2 with (q := (pc→ ∧ l) ∧ ∇(q)).
              eautoflam.
              algMagic.
              destruct (voiceActsForInteg c q).
              eexists.
              eapply AF_ConjR ; auto using refl.
                eautoflam.
                eapply AF_ConjL2 ; eauto.
                autoflam.

          assert ([H;rn;pc ⊔ l ⊔ c0←;l ⊓ c0→ ⊩ p← ≽ q← | D2]).
            eapply IHn with (j := n2) ; eauto with zarith.
            eexists.
            split ; eautoflam.
          do 3 destruct H18, H19.

          eexists.
          split.
            eexists.
            eapply RAF_Fwd with (rn := rn) ; eautoflam.

            eapply inclusionTrans with (y := x) ; auto.
            rewrite H8.
            apply unionMinimal.
              eapply inclusionTrans with (y := D1) ; autoflam.
              eapply inclusionTrans with (y := transfer D2 rn);
                  autoflam.

    intros.
    do 3 destruct H1.
    eauto.
  Qed.

  Lemma robFlow2Integ:
    forall H c pc l p q D,
      [H;c;pc;l ⊩ p ⊑ q | D]
        -> [H;c;pc;l ⊩ p← ≽ q← | D].
  Proof.
    intros.
    setoid_replace (p←) with ((q→ ∧ p←)←) by algMagic.
    setoid_replace (q←) with ((p→ ∧ q←)←) by algMagic.
    eapply robProj_integ.
    auto.
  Qed.

  Lemma robFwd:
    forall H c cn pc l p q r rn D1 D2 D,
      r = native rn
        -> c = native cn
        -> [H;cn;pc;l ⊩ r ≽ (pc→ ∧ l) ∧ ∇(q) | D1]
        -> [H;rn;pc ⊔ l ⊔ c←;l ⊓ c→ ⊩ p ≽ q | D2]
        -> included (union D1 (transfer D2 rn)) D
        -> [H;cn;pc;l ⊩ p ≽ q | D].
  Proof.
    intros.
    do 3 destruct H2, H3.
    eexists.
    split.
      eexists.
      eapply RAF_Fwd with (rn := rn) ; eautoflam.

      apply inclusionTrans
          with (y := union D1 (transfer D2 rn)) ; auto.
      apply unionMinimal.
        apply inclusionTrans with (y := D1) ; autoflam.
        apply inclusionTrans with (y := transfer D2 rn);
            autoflam.
  Qed.

  Lemma factorize:
    forall H c pc l p q D,
      [H;c;pc;l ⊩ p ≽ q | D]
        -> exists d s,
               q == s ∧ d
            /\ [c ⊢ p ≽ s]
            /\ [H;c;pc;l ⊩ pc ≽ ∇(d) | D]
            /\ [H;c;pc;l ⊩ p ≽ d | D].
  Proof.
    assert (forall n H c pc l p q D,
        (exists D', [H;c;pc;l ⊩ p ≽ q # D';n] /\ included D' D)
          -> exists d s,
                 q == s ∧ d
              /\ [c ⊢ p ≽ s]
              /\ [H;c;pc;l ⊩ pc ≽ ∇(d) | D]
              /\ [H;c;pc;l ⊩ p ≽ d | D]).
      induction n as [|n IHn] using strong_induction;
          intros;
          do 2 destruct H0.
        inversion H0 ; omega.

        inversion H0.
          exists ⊥, q.
          split ; [|split ; [|split]] ; algMagic ; eautoflam.

          exists q, ⊥.
          split ; [|split ; [|split]].
            algMagic.
            eautoflam.
            eexists. split.
              eauto.
              eapply inclusionTrans with (y := x); auto.
              rewrite H6. autoflam.
            eauto.

          exists ∇(q'), ⊥.
          split ; [|split ; [|split]] ; algMagic.
            eautoflam.
            eexists. split ; rewrite H3, H4 in * ; eauto.
            eexists. split ; rewrite H4 in * ; eauto.

          assert (n1 <= n) as Hn1 by omega.
          assert (exists D', [H;c;pc;l ⊩ p ≽ q1 # D';n1]
              /\ included D' D1) as Hq1.
            eexists D1.
            autoflam.
          pose proof IHn n1 Hn1 _ _ _ _ _ _ _ Hq1.
          clear Hn1. clear Hq1.

          assert (n2 <= n) as Hn2 by omega.
          assert (exists D', [H;c;pc;l ⊩ p ≽ q2 # D';n2]
              /\ included D' D2) as Hq2.
            eexists D2.
            autoflam.
          pose proof IHn n2 Hn2 _ _ _ _ _ _ _ Hq2.
          clear Hn2. clear Hq2.

          do 3 destruct H16.
          destruct H18.
          destruct H19.

          do 3 destruct H17.
          destruct H21.
          destruct H22.

          eexists (x0 ∧ x2), (x1 ∧ x3).
          split ; [|split ; [|split]].
            rewrite H3, H16, H17. algMagic.

            destruct H18, H21.
            eautoflam.

            algMagic.
            eapply robConjR.
              eapply robWeakenD with (D' := D1) ; auto.
              apply inclusionTrans with (y := x) ; auto.
              rewrite H6. autoflam.

              eapply robWeakenD with (D' := D2) ; auto.
              apply inclusionTrans with (y := x) ; auto.
              rewrite H6. autoflam.

            eapply robConjR.
              eapply robWeakenD with (D' := D1) ; auto.
              apply inclusionTrans with (y := x) ; auto.
              rewrite H6. autoflam.

              eapply robWeakenD with (D' := D2) ; auto.
              apply inclusionTrans with (y := x) ; auto.
              rewrite H6. autoflam.

          assert (n1 <= n) as Hn1 by omega.
          assert (exists D', [H;c;pc;l ⊩ p1 ≽ q # D';n1]
              /\ included D' D1) as Hp1.
            eexists D1.
            autoflam.
          pose proof IHn n1 Hn1 _ _ _ _ _ _ _ Hp1.
          clear Hn1. clear Hp1.

          assert (n2 <= n) as Hn2 by omega.
          assert (exists D', [H;c;pc;l ⊩ p2 ≽ q # D';n2]
              /\ included D' D2) as Hp2.
            eexists D2.
            autoflam.
          pose proof IHn n2 Hn2 _ _ _ _ _ _ _ Hp2.
          clear Hn2. clear Hp2.

          do 3 destruct H16.
          destruct H18.
          destruct H19.

          do 3 destruct H17.
          destruct H21.
          destruct H22.

          eexists ((x0 ∨ x2) ∧ (x1 ∨ x2) ∧ (x0 ∨ x3)), (x1 ∨ x3).
          split ; [|split ; [|split]].
            assert (q == (x1 ∧ x0) ∨ (x3 ∧ x2)) as Hrewrite.
              rewrite <- H16, <- H17. algMagic.
            rewrite Hrewrite. algMagic.

            destruct H18, H21.
            eautoflam.

            pose proof (voiceDisjMonotonic c x0 x2) as Hvoice1.
            pose proof (voiceDisjMonotonic c x2 x1) as Hvoice2.
            setoid_rewrite disjComm in Hvoice2.
            pose proof (voiceDisjMonotonic c x0 x3) as Hvoice3.

            algMagic.
            eapply robConjR ; [|eapply robConjR].
              eapply robWeakenD with (D' := D1) ; auto.
              eapply robTransStatic2 with (q := ∇(x0)) ; auto.
              apply inclusionTrans with (y := x) ; auto.
              rewrite H6. autoflam.

              eapply robWeakenD with (D' := D2) ; auto.
              eapply robTransStatic2 with (q := ∇(x2)) ; auto.
              apply inclusionTrans with (y := x) ; auto.
              rewrite H6. autoflam.

              eapply robWeakenD with (D' := D1) ; auto.
              eapply robTransStatic2 with (q := ∇(x0)) ; auto.
              apply inclusionTrans with (y := x) ; auto.
              rewrite H6. autoflam.

            setoid_rewrite H3.
            eapply robConjR ; [|eapply robConjR].
              eapply robDisjL.
                eapply robTransStatic2 with (q := x0).
                  eapply robWeakenD with (D' := D1) ; auto.
                  apply inclusionTrans with (y := x) ; auto.
                  rewrite H6. autoflam.

                  eautoflam.

                eapply robTransStatic2 with (q := x2).
                  eapply robWeakenD with (D' := D2) ; auto.
                  apply inclusionTrans with (y := x) ; auto.
                  rewrite H6. autoflam.

                  eautoflam.

              eapply robDisjL.
                destruct H18.
                eapply robStatic.
                eautoflam.

                eapply robTransStatic2 with (q := x2).
                  eapply robWeakenD with (D' := D2) ; auto.
                  apply inclusionTrans with (y := x) ; auto.
                  rewrite H6. autoflam.

                  eautoflam.

              eapply robDisjL.
                eapply robTransStatic2 with (q := x0).
                  eapply robWeakenD with (D' := D1) ; auto.
                  apply inclusionTrans with (y := x) ; auto.
                  rewrite H6. autoflam.

                  eautoflam.

                destruct H21.
                eapply robStatic.
                eautoflam.

          assert (n1 <= n) as Hn1 by omega.
          assert (exists D', [H3;c;pc';l' ⊩ p ≽ q # D';n1]
              /\ included D' D1) as Hp.
            eexists D1.
            autoflam.
          pose proof IHn n1 Hn1 _ _ _ _ _ _ _ Hp.
          clear Hn1. clear Hp.
          do 3 destruct H19.
          destruct H20.
          destruct H21.

          assert (n2 <= n) as Hn2 by omega.
          assert (exists D', [H3;c;pc ⊔ l';l ⊩ pc ⊑ pc' # D';n2]
              /\ included D' D2) as Hpc.
            eexists D2.
            autoflam.
          pose proof IHn n2 Hn2 _ _ _ _ _ _ _ Hpc.
          clear Hn2. clear Hpc.
          do 3 destruct H23.
          destruct H24.
          destruct H25.

          assert (robWeakensTo H3 c pc' l' pc l D).
            eexists.
              eapply robWeakenD with (D' := D2).
                eautoflam.
                apply inclusionTrans with (y := x) ; auto.
                rewrite H9. autoflam.

              eapply robWeakenD with (D' := D3).
                eautoflam.
                apply inclusionTrans with (y := x) ; auto.
                rewrite H9. autoflam.

          assert (robWeakensTo H3 c (pc ⊔ l') l pc l D).
            split; eapply robStatic ; eautoflam.

          eexists x0, x1.
          split ; [|split ; [|split]] ; auto.
            eapply robTransStatic1 with (q := pc←). eautoflam.
            eapply robWeakenH with (H1 := H3) (H2 := H4) ; auto.
            eapply robTransRule with (q := pc'←).
              eapply robFlow2Integ.
              eapply robWeaken with (pc0 := pc ⊔ l') (l0 := l) ; auto.
              eapply robWeakenD with (D' := D2).
                eautoflam.

                apply inclusionTrans with (y := x) ; auto.
                rewrite H9. autoflam.

              eapply robWeaken with (pc0 := pc') (l0 := l') ; auto.
              destruct (voiceIsInteg x0) as [y Hvoice].
              setoid_rewrite Hvoice.
              setoid_rewrite Hvoice in H21.
              setoid_rewrite <- (integIdemp y).
              eapply robProj_integ.
              eapply robWeakenD with (D' := D1) ; auto.
              apply inclusionTrans with (y := x) ; auto.
              rewrite H9. autoflam.

              algMagic. eautoflam.

            eapply robWeakenH with (H1 := H3) (H2 := H4) ; auto.
            eapply robWeaken with (pc0 := pc') (l0 := l') ; auto.
            eapply robWeakenD with (D' := D1) ; auto.
            apply inclusionTrans with (y := x) ; auto.
            rewrite H9. autoflam.

          assert (n1 <= n) as Hn1 by omega.
          assert (exists D', [H;c;pc;l ⊩ p ≽ q0 # D';n1]
              /\ included D' D1) as Hp.
            eexists D1.
            autoflam.
          pose proof IHn n1 Hn1 _ _ _ _ _ _ _ Hp.
          clear Hn1. clear Hp.
          do 3 destruct H16.
          destruct H17.
          destruct H18.

          assert (n2 <= n) as Hn2 by omega.
          assert (exists D', [H;c;pc;l ⊩ q0 ≽ q # D';n2]
              /\ included D' D2) as Hq0.
            eexists D2.
            autoflam.
          pose proof IHn n2 Hn2 _ _ _ _ _ _ _ Hq0.
          clear Hn2. clear Hq0.
          do 3 destruct H20.
          destruct H21.
          destruct H22.

          setoid_rewrite H16 in H21.
          pose proof (staticConjActsFor _ _ _ _ H21).
          do 3 destruct H24.
          destruct H25.

          assert (included D1 D).
            apply inclusionTrans with (y := x) ; auto.
            rewrite H6. autoflam.

          assert (included D2 D).
            apply inclusionTrans with (y := x) ; auto.
            rewrite H6. autoflam.

          exists (x5 ∧ x2), x4.
          split ; [|split ; [|split]].
            rewrite H20, H24. algMagic.

            destruct H17, H25.
            eexists.
            eapply AF_Trans with (q := x1) ; eautoflam.

            algMagic.
            eapply robConjR.
              eapply robTransStatic2 with (q := ∇(x0)).
                eapply robWeakenD with (D' := D1) ; auto.
                eapply staticVoice ; auto.

              eapply robWeakenD with (D' := D2) ; auto.

            eapply robConjR.
              eapply robTransStatic2 with (q := x0) ; auto.
              eapply robWeakenD with (D' := D1) ; auto.

              eapply robTransStatic2 with (q := q).
                eapply robWeakenD with (D' := x) ; eautoflam.
                setoid_rewrite H20. eautoflam.

          assert (n2 <= n) as Hn2 by omega.
          assert (exists D', [H;rn;pc ⊔ l ⊔ c0←;l ⊓ c0→ ⊩ p ≽ q # D';n2]
              /\ included D' D2) as Hp.
            eexists D2.
            autoflam.
          pose proof IHn n2 Hn2 _ _ _ _ _ _ _ Hp.
          clear Hn2. clear Hp.
          do 3 destruct H17.
          destruct H18.
          destruct H19.

          exists x0, x1.
          split ; [|split ; [|split]] ; auto.
            destruct H18.
            assert (exists n, [c ⊢ p ≽ x1 # n] /\ n <= x2).
              eapply staticCoerce with (c := rn) ; auto.
            do 2 destruct H21.
            eauto.

            eapply robFwd with (rn := rn) (D1 := D1) (D2 := D2) ; eauto.
              algMagic.
              eapply robTransStatic2 with (q := (pc→ ∧ l) ∧ ∇(q)).
                eautoflam.
                setoid_rewrite H17.
                algMagic. eexists. eapply AF_ConjR ; eautoflam.

              eapply robTransStatic1 with (q := (pc ⊔ l ⊔ c0←)←).
                algMagic. eautoflam.
                destruct (voiceIsInteg x0).
                setoid_rewrite H21.
                setoid_rewrite H21 in H19.
                setoid_rewrite <- (integIdemp x2).
                eapply robProj_integ.
                eauto.

              rewrite <- H7. auto.

            eapply robFwd with (rn := rn) (D1 := D1) (D2 := D2); eauto.
              eapply robTransStatic2 with (q := (pc→ ∧ l) ∧ ∇(q)).
                eautoflam.
                setoid_rewrite H17. algMagic.
                eexists. eapply AF_ConjR ; eautoflam.

              rewrite <- H7. auto.

    intros.
    do 3 destruct H1.
    eapply H.
    eauto.
  Qed.

End robTransHelpers.
(* end hide *)

Lemma robTrans:
  forall H c pc l p q r D,
    [H;c;pc;l ⊩ p ≽ q | D]
      -> [H;c;pc;l ⊩ q ≽ r | D]
      -> [H;c;pc;l ⊩ p ≽ r | D].
Proof.
  intros.
  destruct (robTransHelpers.factorize _ _ _ _ _ _ _ H1)
      as [d [s [Hfact1 [Hfact2 [Hfact3 Hfact4]]]]].
  setoid_rewrite Hfact1.
  eapply robConjR.
    eapply robTransHelpers.robTransStatic2 with (q := q) ; auto.
    eapply robTransHelpers.robTransRule with (q := q) ; auto.
    eapply robTransHelpers.robTransStatic2 with (q := ∇(d)) ; eautoflam.
Qed.

Lemma robProj_conf:
  forall H c pc l p q D,
    [H;c;pc;l ⊩ p ≽ q | D]
      -> [H;c;pc;l ⊩ p→ ≽ q→ | D].
Proof.
  assert (forall n H c pc l p q D,
    (exists D', [H;c;pc;l ⊩ p ≽ q # D';n] /\ included D' D)
      -> [H;c;pc;l ⊩ p→ ≽ q→ | D]).
    intro.
    induction n as [|n IHn] using strong_induction ; intros ; do 2 destruct H0.
      inversion H0 ; omega.

      assert (exists n, weakensTo H c
          pc (l ∧ ∇(q)) pc (l ∧ ∇(q→)) empty n).
        do 5 eexists.
        split ; [|split].
          eautoflam.

          eapply RAF_Static.
          rewrite <- (confIntegBasis q), voiceOfConj.
          algMagic.
          eautoflam.
          autoflam.
          auto.

          autoflam.
      destruct H2.

      inversion H0.
        eexists.
        eautoflam.

        assert ([H; c; pc; l ⊩ pc ≽ ∇(q →) | D2]).
          eapply robTrans with (q := ∇(q)) ; eautoflam.
        do 3 destruct H17.

        eexists.
        split.
          eexists.
          eapply RAF_Lift.
            eautoflam.
            eauto.
            algMagic. eauto.
            autoflam.
            auto.

          eapply inclusionTrans with (y := x) ; auto.
          rewrite H7.
          apply unionMinimal ; autoflam.
          apply unionMinimal ; autoflam.
          apply inclusionTrans with (y := D2) ; autoflam.

        setoid_rewrite H4.
        setoid_rewrite H5.
        algMagic.
        eautoflam.

        assert ([H;c;pc;l ⊩ p→ ≽ q1→ | D1]).
          eapply IHn with (j := n1) ; eautoflamz.
        do 3 destruct H17.

        assert ([H;c;pc;l ⊩ p→ ≽ q2→ | D2]).
          eapply IHn with (j := n2) ; eautoflamz.
        do 3 destruct H19.

        setoid_rewrite H4.
        eexists.
        split.
          eexists.
          algMagic.
          eautoflam.

          apply inclusionTrans with (y := x) ; auto.
          rewrite H7.
          apply unionMinimal;
              eapply inclusionTrans;
              eautoflam.

        assert ([H;c;pc;l ⊩ p1→ ≽ q→ | D1]).
          eapply IHn with (j := n1) ; eautoflamz.
        do 3 destruct H17.

        assert ([H;c;pc;l ⊩ p2→ ≽ q→ | D2]).
          eapply IHn with (j := n2) ; eautoflamz.
        do 3 destruct H19.

        setoid_rewrite H4.
        eexists.
        split.
          eexists.
          algMagic.
          eautoflam.

          apply inclusionTrans with (y := x) ; auto.
          rewrite H7.
          apply unionMinimal;
              eapply inclusionTrans;
              eautoflam.

        assert ([H4;c;pc';l' ⊩ p→ ≽ q→ | D1]).
          eapply IHn with (j := n1) ; eautoflamz.
        do 3 destruct H20.

        eexists.
        split.
          eexists.
          eapply RAF_Weaken with (H1 := H4) (H2 := H5) ; eautoflam.

          apply inclusionTrans with (y := x) ; auto.
          rewrite H10.
          eapply unionMinimal ; autoflam.

        assert ([H;c;pc;l ⊩ p→ ≽ q0→ | D]).
          eapply IHn with (j := n1) ; eauto with zarith.
          eexists.
          split ; eauto.
          apply inclusionTrans with (y := x) ; auto.
          rewrite H7.
          autoflam.

        assert ([H;c;pc;l ⊩ q0→ ≽ q→ | D]).
          eapply IHn with (j := n2) ; eauto with zarith.
          eexists.
          split ; eauto.
          apply inclusionTrans with (y := x) ; auto.
          rewrite H7.
          autoflam.
        do 3 destruct H17, H18.

        eexists.
        split.
          eexists.
          eapply RAF_Trans ; eautoflam.
          algMagic. eauto.

          apply unionMinimal ; autoflam.
          apply unionMinimal ; autoflam.
          apply inclusionTrans with (y := x) ; auto.
          rewrite H7.
          autoflam.

        assert ([H;c;pc;l ⊩ r ≽ (pc→ ∧ l) ∧ ∇(q→) | D1]).
          eapply robTrans with (q := (pc→ ∧ l) ∧ ∇(q)).
            eautoflam.
            eapply robStatic.
            setoid_rewrite <- (confIntegBasis q).
            setoid_rewrite voiceOfConj.
            algMagic.
            eautoflam.

        assert ([H;rn;pc ⊔ l ⊔ c0←;l ⊓ c0→ ⊩ p→ ≽ q→ | D2]).
          eapply IHn with (j := n2) ; eauto with zarith.
          eexists.
          split ; eautoflam.
        do 3 destruct H18, H19.

        eexists.
        split.
          eexists.
          eapply RAF_Fwd with (rn := rn) ; eautoflam.

          eapply inclusionTrans with (y := x) ; auto.
          rewrite H8.
          apply unionMinimal.
            eapply inclusionTrans with (y := D1) ; autoflam.
            eapply inclusionTrans with (y := transfer D2 rn);
                autoflam.

  intros.
  do 3 destruct H1.
  eauto.
Qed.

Hint Resolve robProj_conf : flam.

Lemma robProj_integ:
  forall H c pc l p q D,
    [H;c;pc;l ⊩ p ≽ q | D]
      -> [H;c;pc;l ⊩ p← ≽ q← | D].
Proof.
  apply robTransHelpers.robProj_integ.
Qed.

Hint Resolve robProj_integ : flam.

Lemma robDisjR1:
  forall H c pc l p p1 p2 D,
    [H;c;pc;l ⊩ p ≽ p1 | D]
      -> [H;c;pc;l ⊩ p ≽ p1 ∨ p2 | D].
Proof.
  intros.
  eapply robTrans with (q := p1) ; eautoflam.
Qed.

Hint Resolve robDisjR1 : flam.

Lemma robDisjR2:
  forall H c pc l p p1 p2 D,
    [H;c;pc;l ⊩ p ≽ p2 | D]
      -> [H;c;pc;l ⊩ p ≽ p1 ∨ p2 | D].
Proof.
  intros.
  setoid_rewrite disjComm.
  eautoflam.
Qed.

Hint Resolve robDisjR2 : flam.

Lemma robFwd:
  forall H c cn pc l p q r rn D1 D2 D,
    r = native rn
      -> c = native cn
      -> [H;cn;pc;l ⊩ r ≽ (pc→ ∧ l) ∧ ∇(q) | D1]
      -> [H;rn;pc ⊔ l ⊔ c←;l ⊓ c→ ⊩ p ≽ q | D2]
      -> included (union D1 (transfer D2 rn)) D
      -> [H;cn;pc;l ⊩ p ≽ q | D].
Proof.
  apply robTransHelpers.robFwd.
Qed.

Lemma rob2Nonrob:
  forall H c pc l p q D,
    [H;c;pc;l ⊩ p ≽ q | D] -> [H;c;pc;l ⊢ p ≽ q | D].
Proof.
  assert (forall n H c pc l p q D,
      (exists D', [H;c;pc;l ⊩ p ≽ q # D';n] /\ included D' D) ->
        [H;c;pc;l ⊢ p ≽ q | D]).
    induction n as [|n IHn] using strong_induction ; intros ; do 2 destruct H0.
      inversion H0 ; omega.

      inversion H0.
        eexists.
        split. eautoflam.
        apply emptyMinimal.

        assert (exists n, weakensTo H c pc (l ∧ ∇(q)) pc l empty n).
          do 5 eexists.
          split ; [|split].
            eautoflam.

            eapply RAF_Static.
            algMagic. eautoflam.
            autoflam.
            auto.

            autoflam.
        destruct H16.

        eexists.
        split.
          eautoflam.
          rewrite emptyZero.
            apply inclusionTrans with (y := x) ; eauto.
            rewrite H6.
            autoflam.

        assert (exists n, weakensTo H c pc (l ∧ ∇(q')) pc l empty n).
          unfold weakensTo.
          do 5 eexists.
          split ; [|split].
            eautoflam.

            eapply RAF_Static.
            algMagic. eautoflam.
            autoflam.
            auto.

            autoflam.
        destruct H15.

        eexists.
        split.
          setoid_rewrite H3.
          setoid_rewrite H4.
          eautoflam.

          rewrite emptyZero ; eautoflam.

        assert ([H;c;pc;l ⊢ p ≽ q1 | D1]).
          eapply IHn with (j := n1) ; eauto with zarith.
          eexists.
          split ; eautoflam.
        do 3 destruct H16.

        assert ([H;c;pc;l ⊢ p ≽ q2 | D2]).
          eapply IHn with (j := n2) ; eauto with zarith.
          eexists.
          split ; eauto.
          autoflam.
        do 3 destruct H18.

        eexists.
        split.
          eexists.
          eapply AF_ConjR ; eautoflam.

          apply inclusionTrans with (y := x) ; auto.
          rewrite H6.
          apply unionMinimal;
              eapply inclusionTrans;
              eautoflam.

        assert ([H;c;pc;l ⊢ p1 ≽ q | D1]).
          eapply IHn with (j := n1) ; eauto with zarith.
          eexists.
          split ; eautoflam.
        do 3 destruct H16.

        assert ([H;c;pc;l ⊢ p2 ≽ q | D2]).
          eapply IHn with (j := n2) ; eauto with zarith.
          eexists.
          split ; eauto.
          autoflam.
        do 3 destruct H18.

        eexists.
        split.
          eexists.
          eapply AF_DisjL ; eautoflam.

          apply inclusionTrans with (y := x) ; auto.
          rewrite H6.
          apply unionMinimal;
              eapply inclusionTrans;
              eautoflam.

        assert ([H3;c;pc';l' ⊢ p ≽ q | D1]).
          eapply IHn with (j := n1) ; eauto with zarith.
          eexists.
          split ; eautoflam.
        do 3 destruct H19.

        eexists.
        split.
          eexists.
          eapply AF_Weaken with (H1 := H3) (H2 := H4) (pc' := pc') (l' := l');
              eautoflam.

          apply inclusionTrans with (y := x) ; auto.
          rewrite H9.
          apply unionMinimal ; autoflam.

        assert ([H;c;pc;l ⊢ p ≽ q0 | D]).
          eapply IHn with (j := n1).
            omega.
            eexists.
            split ; eauto.
            apply inclusionTrans with (y := x) ; auto.
            rewrite H6.
            autoflam.
        assert ([H;c;pc;l ⊢ q0 ≽ q | D]).
          eapply IHn with (j := n2).
            omega.
            eexists.
            split ; eauto.
            apply inclusionTrans with (y := x) ; auto.
            rewrite H6.
            autoflam.
        do 3 destruct H16, H17.

        eexists.
        split.
          eexists.
          eapply AF_Trans with (q := q0) ; eautoflam.
          autoflam.

        assert ([H;rn;pc ⊔ l ⊔ c0←;l ⊓ c0→ ⊢ p ≽ q | D2]).
          eapply IHn with (j := n2).
            omega.
            eexists.
            split ; eauto.
            autoflam.
        do 3 destruct H17.

        assert ([H;c;pc;l ⊩ r ≽ pc→ ∧ l | D1]).
          eapply robTrans with (q := (pc→ ∧ l) ∧ ∇(q)).
            eautoflam.
            eapply robConjL1. eapply robStatic. eautoflam.
        do 3 destruct H19.

        eexists.
        split.
          eexists.
          eapply AF_Fwd with (rn := rn) ; eautoflam.

          apply inclusionTrans with (y := x) ; auto.
          rewrite H7.
          apply unionMinimal.
            apply inclusionTrans with (y := D1) ; autoflam.
            apply inclusionTrans with (y := transfer D2 rn);
                autoflam.

  intros.
  do 3 destruct H1.
  eauto.
Qed.

Hint Resolve rob2Nonrob : flam.

Lemma rob2NonrobStatic:
  forall c p q, [c ⊩ p ≽ q] -> [c ⊢ p ≽ q].
Proof.
  intros.

  assert ([PH.empty;c;⊤→;⊤← ⊢ p ≽ q | empty]).
  {
    eapply rob2Nonrob.
    eexists.
    split ; eauto.
    autoflam.
  }

  do 3 destruct H0.
  assert (x =D= empty).
  { apply lessThanEmpty. auto. }
  rewrite H2 in H0.
  eauto.
Qed.

Hint Resolve rob2NonrobStatic : flam.

(**
 **** Robust flows-to judgements
 *)

(*
 * Converts a robust flows-to relationship to a robust acts-for relationship
 * between confidentiality projections.
 *)
Lemma robFlow2Conf:
  forall H c pc l p q D,
    [H;c;pc;l ⊩ p ⊑ q | D]
      -> [H;c;pc;l ⊩ q→ ≽ p→ | D].
Proof.
  intros.
  setoid_replace (q→) with ((q→ ∧ p←)→) by algMagic.
  setoid_replace (p→) with ((p→ ∧ q←)→) by algMagic.
  eautoflam.
Qed.

Hint Resolve robFlow2Conf : flam.

Lemma robFlowProj_conf:
  forall H c pc l p q D,
    [H;c;pc;l ⊩ p ⊑ q | D]
      -> [H;c;pc;l ⊩ p→ ⊑ q→ | D].
Proof.
  intros.
  algMagic.
  autoflam.
Qed.

Hint Resolve robFlowProj_conf : flam.

(*
 * Converts a robust flows-to relationship to a robust acts-for relationship
 * between integrity projections.
 *)
Lemma robFlow2Integ:
  forall H c pc l p q D,
    [H;c;pc;l ⊩ p ⊑ q | D]
      -> [H;c;pc;l ⊩ p← ≽ q← | D].
Proof.
  apply robTransHelpers.robFlow2Integ.
Qed.

Hint Resolve robFlow2Integ : flam.

Lemma robFlowProj_integ:
  forall H c pc l p q D,
    [H;c;pc;l ⊩ p ⊑ q | D]
      -> [H;c;pc;l ⊩ p← ⊑ q← | D].
Proof.
  intros.
  algMagic.
  autoflam.
Qed.

Hint Resolve robFlowProj_integ : flam.

Lemma robFlowTrans:
  forall H c pc l p q r D,
    [H;c;pc;l ⊩ p ⊑ q | D]
      -> [H;c;pc;l ⊩ q ⊑ r | D]
      -> [H;c;pc;l ⊩ p ⊑ r | D].
Proof.
  intros.
  eapply robConjR.
    eapply robConjL1.
    eapply robTrans with (q := q→).
      eapply robFlow2Conf. auto.
      eapply robFlow2Conf. auto.

    eapply robConjL2.
    eapply robTrans with (q := q←).
      eapply robFlow2Integ. auto.
      eapply robFlow2Integ. auto.
Qed.
